#
# main module for SUGR, basically used for calling sub-modules
#
#
# $Docid: Oct. 24th 2001  13:37  Stephan Heuel $

package SUGR;

require 5.6.0;

use strict;
use warnings;
no warnings 'redefine';

# set the version:
our $VERSION = '0.1.0';

# get the version of PDL and strip a possible cvs part of it. Remember
# if this was a cvs version.
#use PDL;
# my $cvs = ((my $v = $PDL::VERSION) =~ s/cvs//);
# unless ($cvs) {
#   $cvs = ($v =~ s/snap[\d]+//);
# }
# my $vstring_version = pack("U*",split(/\./,$v));
# # die if the current PDL version
# die "*** Error:\n".
#     "    Current version of PDL \"$PDL::VERSION\" does not match required version.\n".
#     "    Get the latest PDL version, see http://pdl.perl.org\n".
#     "    (Note that 2.2.1 is not enough, at least get a cvs snapshot, see\n".
#     "    ftp://pdl.sourceforge.net/pub/pdl )\n"
#   unless ($vstring_version gt v2.2.2 ||
# 	  ($vstring_version eq v2.2.1 && $cvs));


sub SUGR::import {
  my $pkg = (caller())[0];

  eval <<"EOD";
package $pkg;

# Load the fundamental packages

#use PDL::Math;
use PDL::Matrix;

use SUGR::Point2d;
use SUGR::Line2d;
use SUGR::Point;
use SUGR::Line;
use SUGR::Plane;

use SUGR::Relation;
use SUGR::Estimation;

# old transformations
use SUGR::Transformation;
use SUGR::Projection;

# new transformations
use SUGR::Homography;
use SUGR::ProjectiveCamera;

use SUGR::Scene;

EOD

die $@ if $@;

}

1;

__END__

=head1 NAME

SUGR -- A Library for I<S>tatistically I<U>ncertain I<G>eometric I<R>easoning

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR->VERSION

=head1 SYNOPSIS

=over 4

=item - 

Use I<SUGR> on the command line with sugr(1)

=item -

Use I<SUGR> in the interactive Perl shell sugrshell(1)

=item -

Use I<SUGR> in Perl:

  use SUGR;

  # create a new 3D point, with an homogeneous 4-vector and
  # a homogeneous 4x4 covariance matrix
  $X = new SUGR::Point([0,1,1,1],             
		       [[0.01,0   ,0   ,0],   
			[0   ,0.01,0   ,0],   
			[0   ,0   ,0.01,0],
			[0   ,0   ,0   ,0]]);

  # create two other points with 
  # the same covariance matrix as the first point
  $Y = new SUGR::Point([0,2,3,1],$X->cov); 
  $Z = new SUGR::Point([2,2,0,1],$X->cov); 

  # construct a new 3D line using $X and $Y, 
  # including error propagation
  $L = new SUGR::Line($X,$Y);

  # and a plane with the third point $Z
  $A = new SUGR::Plane($L,$Z);

  # create a projective camera $P with given
  # projection center $X0 (as a SUGR point!),
  # rotation matrix $R and focal length $c with its
  # variance $c_var
  $X0 = new SUGR::Point([10,10,10,1],$X->cov); 
  $R  = new PDL::Matrix([[0,1,0],[1,0,0],[0,0,1]]);
  $P  = sugr_ProjectiveCamera({ "X0"    => $X0,
			 	"R"     => $R,
			 	"c"     => 1,
			 	"c_var" => 0.001,
			      });

  # construct a new 2D point as an image of $X with 
  # the projective camera $P
  $x = new SUGR::Point2d($X,$P);

=back

=head1 OVERVIEW

The I<SUGR> module provides constructors and operations for geometrical
reasoning in 2D and 3D based on a simple but effective uncertainty
model. The geometrical objects that can be dealt with are points,
lines and planes. Let's say you have three 3D points $X, $Y and $Z and
you want to construct a 3D plane out of them. Or you want to construct
a 3D line $L from $X and $Y and test whether the third point is on
this line. Or you want to check whether two planes are parallel, two
2D lines are identical or a line is orthogonal to a plane. These and
many other constructions relationships may be performed with this
library using the second moments as a description of the uncertainty
of the objects. So I<SUGR> basically performs error propagation and
hypothesis testing on imprecise geometrical data.

Furthermore one can apply projective transformations such as a
hompgraphy, projection with a pin-hole camera on the geometrical
objects. In fact, these transformations are represented as objects
themselves.

The parameters of the geometrical objects are represented in
homogeneous coordinates, the uncertainty is modelled with the second
moments of the homogeneous vector.

=head2 Background

Recently, projective geometry was extensively used in areas
such as Computer Vision and Computer Graphics to handle
geometrical structures and relations in images and 3D
space. Using the so-called Grassmann-Cayley algebra, one can
find nice algebraic bilinear expressions for seemingly ugly
things such as the intersection of a line and a plane in
3D. Ugly, because if you want to write a general solution
for the line-plane-intersection, you end up with non-linear,
hence ugly expressions. 

But it is not clear how to include notions of uncertainty into the
algebraic expressions for projective geometry tasks. This library
attempts to model uncertainty by the second moments of the homogeneous
parameter vectors. Usually we expect the user to know the covariance
vector of a point in 3D or 2D or maybe the covariance vector of a line
in 2D. For all subsequent constructions from these entities we can do
error propagation which is sufficently accurate for practical cases.

=head1 DESCRIPTION

We now want do give a short description on how to use the I<SUGR>
library. For a details please refer to the subsequent documentation
pages given in  or check the (upcoming) cookbook
L<SUGR::Cookbook|SUGR::Cookbook> (not written yet).

=head2 Objects

Objects in I<SUGR> can bei either points, lines or planes in 2d and 3D. 

=head2 Constructors 

We can distinguish between direct constructors and constructors by
estimation.

=head2 Relations

=head2 Overdetermined Constructions

=head1 FILES

=over 4

=item I<L<SUGR::Scene|SUGR::Scene>>

A class for a set of SUGR objects

=item I<L<SUGR::Scene::GXF|SUGR::Scene::GXF>>

Definition if a SUGR scene using an  ASCII file, which is in
the format GXF(sugr)

=item I<L<SUGR::Point2d|SUGR::Point2d>>

A class for 2D points.

=item I<L<SUGR::Line2d|SUGR::Line2d>>

A class for 2D lines.

=item I<L<SUGR::Point|SUGR::Point>>

A class for 3D points.

=item I<L<SUGR::Line|SUGR::Line>>

A class for 3D lines.

=item I<L<SUGR::Plane|SUGR::Plane>>

A class for 3D planes.

=item I<L<SUGR::ProjectiveCamera|SUGR::ProjectiveCamera>>

A class for a projective camera modelling the classic pin-hole camera
with 11 parameters.

=item I<L<SUGR::Relation|SUGR::Relation>>

A class providing hypotheses tests for relations between geometric
objects.

=item I<L<SUGR::Estimation|SUGR::Estimation>>

A module providing a generic estimation for a large set of geometric
problems.


=back

=head1 BUGS

Undoubtedly there are many bugs and especially inconsistencies in this
code. However, this code was thouroughly checked and used extensively
in a vision system for 3D reconstruction. Please mail to
stephan@heuel.org for any problems.

=head1 TODO

=over

=item -
  
lots of documentation, including sub manual pages and Cookbook

=item -

more tests and test scripts

=item -

Some modules are obsolete and will be removed from this distribution

=item -

lots of other things ...

=back

=head1 SEE ALSO

L<sugr(1)>, L<sugrshell(1)>, L<PDL(3)>

=head1 LITERATURE

=over

=item * 

I<W. FE<ouml>rstner, A. Brunn, and S. Heuel.> "Statistically
testing uncertain geometric relations". In G. Sommer, N. KrE<uuml>ger,
and Ch. Perwass, editors, I<Mustererkennung 2000>, pages 17-26. DAGM,
Springer, September 2000. 

=item * 

I<S. Heuel.> "Points, lines and planes and their optimal
estimation. In I<Pattern Recognition", 23rd DAGM Symposium>, number 2191
in LNCS, pages 92-99. Springer, September 2001. 

=item * 

I<S. Heuel and W. FE<ouml>rstner> "Matching, reconstructing and
grouping 3d lines from multiple views using uncertain projective
geometry." In I<CVPR '01>. IEEE, 2001. accepted for publication.


=back

=head1 AUTHOR(S)

Stephan Heuel (stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

