=head1 NAME

SUGR::Basic -- base class for uncertain geometrical elements

=head1 VERSION

  This document refers to version 0.01

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 Overview

=head2 Constructurs and initializaton

=head2 Class and object methods

=head2 Something else?!

=head1 DIAGNOSTICS

=over 4

=item "error messages?!"

=back

=head1 BUGS

=head1 TODO

- documentation
- check EXPORT and maybe change it to EXPORT_OK

=head1 FILES

=head1 SEE ALSO

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000 Stephan Heuel All rights reserved. There is no
warranty. You are allowed to redistribute this software /
documentation under certain conditions. For details, see the file
COPYING in the PDL distribution. If this file is separated from the
PDL distribution, the copyright notice should be included in the file.

=cut

package SUGR::Basic;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 0.01;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request

use PDL;
use PDL::Matrix;
use PDL::Math;
use PDL::Slatec;
use SUGR::Optimize qw/S_pp Pi_pp Pidual_pp Gamma_pp Gammadual_pp/;

####################################################################
#
# Config variables
#

# this is for testing relations if we don't have covariance matrices
our $eps = 0.0001; 

# this factor determines how many times larger the homogeneous
# proportion should be compared to the euclidean portion of a vector
our $homo_euclid_factor = 10; 

# do we want to compute the constraints for the objects?
our $apply_constraints = 0; 

########################################
#
# look up tables
#

# look-up-table for homogeneous vector dimensions of geometric entities
our %no_of_dimensions =  ("Point2d"  => 3,
			  "Line2d"   => 3,
			  "Point"    => 4,
			  "Line"     => 6,
			  "Plane"    => 4,
			  "ProjectiveCamera"    => 12,
			  "Homography"          => 16,
			  "FundamentalMatrix"    => 9,
			);

push @EXPORT_OK, "%no_of_dimensions"; 

# look-up-table for number of constraints for each vector
our %no_of_constraints = ("Point2d"  => 1,
			  "Line2d"   => 1,
			  "Point"    => 1,
			  "Line"     => 2,
			  "Plane"    => 1,
			  "ProjectiveCamera"    => 1,
			  "Homography"    => 1,
			  "FundamentalMatrix"    => 2,
			);
push @EXPORT_OK, "%no_of_constraints"; 

our %possible_test_types =  ( "Equal" => [ 
					  [ "Point2d", "Point2d" ],
					  [ "Line2d" , "Line2d"],
					  [ "Point", "Point"],
					  [ "Line", "Line" ],
					  [ "Plane", "Plane"],
					  [ "Homography", "Point", "Point"] 
					 ], 

			     "Incident" => [ [ "Point2d","Line2d" ],
					     [ "Point", "Line" ],
					     [ "Point","Plane"],
					     [ "Line","Line"],
					     [ "Line","Plane"],
					     [ "ProjectiveCamera", "Point2d", "Point" ], 
					     [ "ProjectiveCamera", "Line2d" , "Line" ] 
					   ],
			     
			     "Orthogonal" => [ [ "Line2d","Line2d"],
					       [ "Line","Line"],	 
					       [ "Line","Plane"],	 
					       [ "Plane","Plane"] ],

			     "Parallel" => [ [ "Line2d","Line2d"],	 
					     [ "Line","Line"],		 
					     [ "Line","Plane"],		 
					     [ "Plane","Plane"]]);	 
push @EXPORT_OK, "%possible_test_types"; 


# look up table for degrees of freedom
my %DegFreedom =  ( "EqualPoint2dPoint2d"	 => 2,
		    "EqualLine2dLine2d"		 => 2,
		    "EqualPointPoint"		 => 3,
		    "EqualLineLine"		 => 4,
		    "EqualPlanePlane"		 => 3,

		    "IncidentPoint2dLine2d"	 => 1,
		    "IncidentPointLine"		 => 2,
		    "IncidentPointPlane"	 => 1,
		    "IncidentLineLine"		 => 1,
		    "IncidentLinePlane"		 => 2,

		    "OrthogonalLine2dLine2d"	 => 1,
		    "OrthogonalLineLine"	 => 1,
		    "OrthogonalLinePlane"	 => 2,
		    "OrthogonalPlanePlane"	 => 1,

		    "ParallelLine2dLine2d"	 => 1,
		    "ParallelLineLine"		 => 2,
		    "ParallelLinePlane"		 => 1,
		    "ParallelPlanePlane"	 => 2,

		    "IncidentProjectiveCameraPoint2dPoint"	 => 2,

		    "IncidentProjectiveCameraLine2dLine"	 => 2,

		    "EqualHomographyPointPoint"	 => 3,

		    # elements
		    "Point2d"			 => 2,
		    "Line2d"			 => 2,
		    "Point"			 => 3,
		    "Line"			 => 4,
		    "Plane"			 => 3,

		    # duplicates
		    "IncidentLine2dPoint2d"	 => 1,
		    "IncidentLinePoint"		 => 2,
		    "IncidentPlanePoint"	 => 1,
		    "IncidentPlaneLine"		 => 2,


		    "IncidentProjectiveCameraPointPoint2d"	 => 2,
		    "IncidentPoint2dPointProjectiveCamera"	 => 2,
		    "IncidentPointPoint2dProjectiveCamera"	 => 2,
		    "IncidentPointProjectiveCameraPoint2d"	 => 2,

		    "IncidentProjectiveCameraLineLine2d"	 => 2,
		    "IncidentLine2dLineProjectiveCamera"	 => 2,
		    "IncidentLineLine2dProjectiveCamera"	 => 2,
		    "IncidentLineProjectiveCameraLine2d"	 => 2,

		    "EqualHomographyPointPoint"	 => 3,
		    "EqualPointHomographyPoint"	 => 3,
		    "EqualPointPointHomography"	 => 3,

		    "OrthogonalPlaneLine"	 => 2,

		    "ParallelPlaneLine"		 => 1,
		  );

############################################################
#
# global methods
#

# this returns or sets the epsilon value which is used for tests in
# case there is no covariance matrix given
sub eps {
  my ($value) = @_;
  $eps = $value || $eps;
}
push @EXPORT_OK,"eps";

# this returns or sets the scale factor, which determins how much
# larger the homogeneous part of an entity-vector should be compared
# to the euclidean part.
sub homo_euclid_factor {
  my ($value) = @_;
  $homo_euclid_factor = $value || $homo_euclid_factor;
}
push @EXPORT_OK,"homo_euclid_factor";

# set or return the flag for applying a constraint or not
sub apply_constraints {
  my ($value) = @_;
  $apply_constraints = defined($value) ? $value : $apply_constraints;
}
push @EXPORT_OK,"apply_constraints";


################################################################################
#
# Utility functions
#



# we may need to normalize the homogenous vectors that go into the
# tests, therefore we have to calculate the largest norm of the
# euclidean resp. homogeneous part from a list of input objects.
sub max_norm_of_euclidean_part {
  use SUGR::MyListUtil;
  SUGR::MyListUtil::max(map { $_->norm_of_euclidean_part } @_);
}
push @EXPORT_OK,"max_norm_of_euclidean_part";

sub max_norm_of_homogeneous_part {
  use SUGR::MyListUtil;
  SUGR::MyListUtil::max(map { $_->norm_of_homogeneous_part } @_);
}
push @EXPORT_OK,"max_norm_of_homogeneous_part";

# normalize sugr objects in the euclidean way, so that the euclidean
# part is larger than the homogeneous part. This will make copies from
# the given sugr_objects
sub normalize_sugr_objects {
  my (@object_list) = @_;

  my $max_norm_euclid = homo_euclid_factor() * max_norm_of_euclidean_part(@object_list);
  my $max_norm_hom = max_norm_of_homogeneous_part(@object_list);

  # don't do normalizing if our condition is already fulfilled
  return (1,@object_list) if ($max_norm_euclid <= $max_norm_hom);
  
  # TODO: how do I normalize when I have a camera as a projective
  # element?! I don't know right now, so I just don't do anything yet.
  my $object_types = join(",",map { ref($_) } @object_list);
  return (1,@object_list) if $object_types =~ /ProjectiveCamera/;
  
  my $factor = 1/$max_norm_euclid;

  my @return = ($factor);
  foreach my $object (@object_list) {
    push @return,$object->copy->scale_euclidean_part($factor);
  }
#  warn "normalize @object_list with $factor => @return!\n";
  return @return;
}
push @EXPORT,"normalize_sugr_objects";


#
# returns degree of freedom for certain tests
#
sub DegFreedom {
  my ($type,$E1,$E2) = @_;
  return $DegFreedom{$type} if !defined($E1);
  my ($s1,$s2) = ( ref($E1), ref($E2) );
  $s1 =~ s/^.*:://;
  $s2 =~ s/^.*:://;
  return $DegFreedom{"$type$E1$E2"};
}
push @EXPORT_OK,"DegFreedom";

sub Identity {
    my ($n) = @_;

    ($n,undef) = $n->mdims if ref($n) =~ /PDL::Matrix/;
    # return precomputed matrix if possible
    die "*** Error in SUGR::Identity: \$n==$n\n" if $n<2;
    my $Id = mzeroes($n,$n);
    (my $tmp = $Id->diagonal(0,1) ) .= ones($n);
    return $Id;
}
push @EXPORT_OK,"Identity";

# Impose constraint-matrix H on covariance matrix C
sub ApplyConstraints {
    my ($C,$H,$force_constraints) = @_;

    $force_constraints ||= $apply_constraints;
    return $C if !$force_constraints;

    my $dimC;
    ($_,$dimC) = $C->mdims;
    
    my $P = Identity($dimC) - ($H x matinv(~$H x $H) x ~$H);
    return $P x $C x $P;
}
push @EXPORT_OK,"ApplyConstraints";

# Impose constraint-matrix H on covariance matrix C
sub ApplyConstraintsInv {
    my ($C,$H,$H0) = @_;

    my $dimC;
    ($_,$dimC) = $C->mdims;

    my $P = Identity($dimC) - ($H0 x matinv(~$H0 x $H0) x ~$H0);
    my $Pinv = pseudoinverse($P,$H);
    return $Pinv x $C x $Pinv;
}

sub S {
    my ($self) = @_;
    my $v = (ref($self) =~ m/SUGR/) ? $self->vec : $self;
    my $M = zeroes(3,3);
    $v->S_pp($M);
    bless $M, "PDL::Matrix";
    return $M;
}
push @EXPORT_OK,"S";

sub S_bak {
    my ($self) = @_;
    my ($u,$v,$w) = (ref($self) =~ m/SUGR/) ? list($self->vec) : list($self);
    return mpdl [
		 [  0, -$w,  $v],
		 [ $w,   0, -$u],
		 [-$v,  $u,  0 ]
		];
}
push @EXPORT_OK,"S_bak";

# old version, not using PDL::PP
sub Pi_bak {
    my ($self) = @_;
    my @v = (ref($self) =~ m/SUGR/) ?  list($self->vec) : list($self);
    return mpdl [
		 [$v[3], 0, 0, -$v[0]],
		 [0, $v[3], 0, -$v[1]],
		 [0, 0, $v[3], -$v[2]],
		 [0, -$v[2], $v[1], 0],
		 [$v[2], 0, -$v[0], 0],
		 [-$v[1], $v[0], 0, 0]
		];
}
push @EXPORT_OK,"Pi_bak";

sub Pi {
    my ($self) = @_;
    my $v = (ref($self) =~ m/SUGR/) ?  $self->vec : $self;
    my $M = zeroes(4,6);
    $v->Pi_pp($M);
    bless $M, "PDL::Matrix";
    return $M;
}
push @EXPORT_OK,"Pi";

# old version, not using PDL::PP
sub Pidual_bak {
    my ($self) = @_;
    my @v = (ref($self) =~ m/SUGR/) ?  list($self->vec) : list($self);
    return mpdl[
	  	[0,         -$v[2], $v[1], 0        ],
		[ $v[2],         0,-$v[0], 0        ],
		[-$v[1], $v[0], 0        , 0        ],
		[ $v[3], 0        , 0        ,-$v[0]],
		[0,          $v[3], 0        ,-$v[1]],
		[0,          0,         $v[3],-$v[2]]
	       ];
}
push @EXPORT_OK,"Pidual_bak";

sub Pidual {
    my ($self) = @_;
    my $v = (ref($self) =~ m/SUGR/) ?  $self->vec : $self;
    my $M = zeroes(4,6);
    $v->Pidual_pp($M);
    bless $M, "PDL::Matrix";
    return $M;
}
push @EXPORT_OK,"Pidual";

sub Gamma {
    my ($self) = @_;
    my $v = (ref($self) =~ m/SUGR/) ?  $self->vec : $self;
    my $M = zeroes(4,4);
    $v->Gamma_pp($M);
    bless $M, "PDL::Matrix";
    return $M;
}
push @EXPORT_OK,"Gamma";


sub Gamma_bak {
    my ($self) = @_;
    my @v = (ref($self) =~ m/SUGR/) ?  list($self->vec) : list($self);
    return mpdl [
		 [     0, $v[2],-$v[1],-$v[3]],
		 [-$v[2],     0, $v[0],-$v[4]],
		 [ $v[1],-$v[0],     0,-$v[5]],
		 [ $v[3], $v[4], $v[5],     0]
		];
}
push @EXPORT_OK,"Gamma_bak";

sub Gammadual {
    my ($self) = @_;
    my $v = (ref($self) =~ m/SUGR/) ?  $self->vec : $self;
    my $M = zeroes(4,4);
    $v->Gammadual_pp($M);
    bless $M, "PDL::Matrix";
    return $M;
}
push @EXPORT_OK,"Gammadual";

sub Gammadual_bak {
    my ($self) = @_;
    my @v = (ref($self) =~ m/SUGR/) ?  list($self->vec) : list($self);
    return mpdl [
		 [0     , $v[5],-$v[4],-$v[0]],
		 [-$v[5],     0, $v[3],-$v[1]],
		 [ $v[4],-$v[3],     0,-$v[2]],
		 [ $v[0], $v[1], $v[2],     0]
		];
}
push @EXPORT_OK,"Gammadual_bak";

# Jacobian for transformation of point projection to line projection
# (i.e. from P to Q or from P to Ps or whatever notation you may want
# to use).
sub Jp {
  my ($self) = @_;
  
  my $Pvec = (ref($self) =~ m/SUGR/) ?  $self->vec : $self;

  my $r1 = $Pvec->slice("0:3");
  my $r2 = $Pvec->slice("4:7");
  my $r3 = $Pvec->slice("8:11");

  my $Jp = mzeroes(18,12);
  my @Pid = map { Pidual($_) } ($r1,$r2,$r3);
  unshift @Pid, "not in use"; # insert a dummy element, so that
                              # indexing starts with 1

  ($_ = $Jp->slice("0:5,4:7"))   .=  -$Pid[3]; 
  ($_ = $Jp->slice("0:5,8:11"))  .=   $Pid[2]; 
					      
  ($_ = $Jp->slice("6:11,0:3"))  .=   $Pid[3]; 
  ($_ = $Jp->slice("6:11,8:11")) .= - $Pid[1]; 
					      
  ($_ = $Jp->slice("12:17,0:3")) .= - $Pid[2]; 
  ($_ = $Jp->slice("12:17,4:7")) .=   $Pid[1]; 

  return 0.5*$Jp;
}
push @EXPORT_OK,"Jp";

# returns pseudoinverse of matrix M with nullspace H
sub pseudoinverse {
  my ($M,$H)=@_;
  
  return matinv($M) if !defined($H);

  my ($rowdimM,$coldimM) = $M->mdims;
  my ($rowdimH,$coldimH) = $H->mdims;
  
  my $P = mzeroes($rowdimM+$coldimH,$coldimM+$coldimH);
  
  ($_ = $P->slice("0:".($rowdimM-1).",0:".($coldimM-1))) .= $M;
  ($_ = $P->slice("0:".($rowdimM-1).",$coldimM:".($coldimM+$coldimH-1))) .= $H;
  ($_ = $P->slice("$rowdimM:".($coldimM+$coldimH-1).",0:".($coldimM-1))) .= ~$H;
  
  my $Pinv = matinv($P);
  
  return $Pinv->slice("0:".($rowdimM-1).",0:".($coldimM-1));
}
push @EXPORT_OK,"pseudoinverse";

# the Jacobians for differences
sub C {
  my ($u,$i) = @_; # $u must be a vector
  # check for different datatypes
  $i = (!defined($i)) ? ChooseIndex_for_C($u) : ChooseIndex_for_C($u,$i);
  # $i is the index with the highest absolute value
#  print $u->clump(-1),"-------> $i\n";
  my $r = ($u->mdims)[0];
  my $C = mzeroes($r,$r);
  ($_ =  $C->slice(",$i")) .= $u;
  ($_ =  $C->diagonal(0,1)) -= $u->at($i);
  return $C;
}
push @EXPORT_OK,"C";

sub ChooseIndex_for_C {
  my ($v1,$v2) = @_;
  my $val = defined($v2) ? abs($v1*$v2) : abs($v1);
  my $idx = maximum_ind(~$val)->at(0);
  $idx = maximum_ind(abs(~$v2)) if $val->at($idx) == 0;
  return $idx;
}
push @EXPORT_OK,"ChooseIndex_for_C";


# these two functions could substitute older ChooseGamma functions
sub ChooseRowIndices {
  my ($M,$no_of_rows) = @_;

  my $vec = $M->PDL::abs->maximum;
  my @keep; # here we collect all those rows which have to be preserved
  # TODO: the below can be optimized using qsorti
  for (1..$no_of_rows) {
    push @keep, int $vec->maximum_ind;
    set($vec,$keep[-1],0);
  }
  sort @keep;
}
sub ChooseColIndices {
  my ($M,$no_of_rows) = @_;
  return ChooseRowIndices(~$M,$no_of_rows);
}
sub ChooseRows {
  my ($M,@indices) = @_;
  my $nC = ($M->mdims())[1];
  return $M->dice([@indices],[0..($nC-1)]);
}
sub ChooseCols {
  my ($M,@indices) = @_;
  
  return ~ChooseRows(~$M,@indices);
}
push @EXPORT_OK,"ChooseRowIndices";
push @EXPORT_OK,"ChooseColIndices";
push @EXPORT_OK,"ChooseRows";
push @EXPORT_OK,"ChooseCols";

my %chisqr; # this look up table will be created on the fly
sub chisqr {
  my ($degfree, $alpha) = @_;

  croak("no alpha value given\n") if !defined($alpha);

  # maybe we have already computed this value, so we can return it
  # without computing
  return $chisqr{"$degfree,$alpha"} if defined($chisqr{"$degfree,$alpha"});

  my ($chi2, $t, $x_alpha);
  my $c0 = 2.515517;
  my $c1 = 0.802853;
  my $c2 = 0.010328;
  my $d1 = 1.432788;
  my $d2 = 0.189269;
  my $d3 = 0.001308;

  $t = sqrt(log(1.0 / ((1-$alpha) * (1-$alpha))));

  #
  # x_alpha : alpha-Fraktil der standardisierten Normalverteilung,
  # Approximation nach Seite 104, Gl. 241.8 , Koch. Gilt fuer
  # alpha zwischen 0.5 und 1 ====> fuer alpha<0.5 muss x_alpha negiert werden,
  # siehe Koch Gleichung (241.7), SH und A. Brunn 17.9.98
  $x_alpha = $t - (($c0 + $c1 * $t + $c2 * $t**2) /
		   (1 + $d1 * $t + $d2 * $t**2 + $d3 * $t**3));

  $x_alpha -= $x_alpha if ($alpha<0.5);
  #
  # Approximation fuer Chi2 Fraktilwerte: nach Gl. 261.11 , Seite 118,
  # Koch **
  #
  $chi2 = $degfree * ($x_alpha * sqrt(2.0 / (9.0 * $degfree)) + 1.0 - 2.0 / (9.0 * $degfree))**3;
  
  return $chi2;
}
push @EXPORT_OK, "chisqr";

# if we apply a vec operation on a matrix, we align the matrix
# elements row-wise to a vector. If we transpose the matrix, the
# aligned vector looks differently. This function returns a matrix
# which can be applied to an aligned vector to reflect the
# transponation of the original matrix
sub permutation_vec_transpose {
  my ($r,$c) = @_;

  my $T = mzeroes($r*$c,$r*$c);
  foreach my $i (0..($r-1)) {
    foreach my $j (0..($c-1)) {
      $T->set($j*$r+$i,$i*$c+$j,1);
    }
  }
  return $T;
}
push @EXPORT_OK, "permutation_vec_transpose";

%EXPORT_TAGS = (ALL => [@EXPORT,@EXPORT_OK]);

return 1 if caller;

################################################################################
#
# test code

permutation_vec_transpose(3,3);
my $M34 = permutation_vec_transpose(3,4);

my $P = msequence(3,4);

print $P;
print $M34 x $P->reshape(12,1);
print $P->transpose->reshape(12,1); 
