=head1 NAME

SUGR::Cuboid -- generate random cuboid

=head1 VERSION

  This document refers to version 0.01

=head1 SYNOPSIS
 
=head1 DESCRIPTION

=head2 Overview

=head2 Constructurs and initializaton

=head2 Class and object methods

=head2 Something else? !

=head2 Literature

=head1 DIAGNOSTICS

=over 4

=item "error messages?!"

=back

=head1 BUGS

=head1 TODO

=head1 FILES

=head1 SEE ALSO

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000 Stephan Heuel All rights reserved. There is no
warranty. You are allowed to redistribute this software /
documentation under certain conditions. For details, see the file
COPYING in the PDL distribution. If this file is separated from the
PDL distribution, the copyright notice should be included in the file.

=cut

package SUGR::Cuboid;

  
use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 0.01;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL;
use PDL::Matrix;
use PDL::Math;
use Math::Random qw(random_normal
		    random_uniform
		    random_multivariate_normal
		    random_set_seed_from_phrase
		    random_set_seed);
use SUGR::Point;
use SUGR::Line;
use SUGR::Plane;
use SUGR::Transformation;



sub generate_cuboid_simple {
  my %args = (
	      # default values
	      "mean" => [1,1,1],
	      "standard_deviation" => [0.01,0.01,0.01],
	      "covariances" => 1,
	      # argument pair list goes here
	      @_ 
	     );

  my %Objects = generate_cuboid(%args,
				"no_points" => 1,
				"no_lines" => 1,
				"no_planes" => 1);
  
  return ("Points"       => [ map { $_->[0] } @{$Objects{Points}} ],
	  "Lines"        => [ map { $_->[0] } @{$Objects{Lines}} ],
	  "Planes"       => [ map { $_->[0] } @{$Objects{Planes}} ],
	  "Ideal_Points" => [ map { $_->[0] } @{$Objects{Ideal_Points}} ],
	  "Ideal_Lines"  => [ map { $_->[0] } @{$Objects{Ideal_Lines}} ],
	  "Ideal_Planes" => [ map { $_->[0] } @{$Objects{Ideal_Planes}} ]
	 );
		  
}
push @EXPORT,"generate_cuboid_simple";

=head2 generate_cuboid

This function generates a set of noisy cuboid elements. the output is
a hash table.

=head3 Input parameters

=over

=item 

=back


=for example

=cut

sub generate_cuboid {
  my %args = (
	      # default values
	      "mean" => [1,1,1],
	      "standard_deviation" => [0.01,0.01,0.01],
	      "no_points" => 5,
	      "no_lines" => 5,
	      "no_planes" => 5,
	      "covariances" => 1,
	      "verbose" => 1,
	      # argument pair list goes here
	      @_ 
	     );

  my ($meanx,$meany,$meanz) = @{$args{mean}};
  my ($sdx,$sdy,$sdz) = @{$args{standard_deviation}};
  my $no_points = $args{no_points};
  my $no_lines = $args{no_lines};
  my $no_planes = $args{no_planes};
  my $generate_covariance = $args{covariances};
  my $verbose = $args{verbose};

  random_set_seed(@{$args{random_seed}}) if defined($args{random_seed});

  print "Generating cube data with $no_points points, $no_lines lines and $no_planes planes per cuboid element.\n" if $verbose;
  print " (standard deviations for a point are $sdx,$sdy,$sdz)\n"  if $verbose;
  my ($X_cube,$X_idealcube) = generate_cuboid_points($meanx,$meany,$meanz,
						     $sdx,$sdy,$sdz,
						     [map {$no_points} (1..8)],
						     $generate_covariance);
  my ($L_cube,$L_idealcube) = generate_cuboid_lines ($meanx,$meany,$meanz,
						     $sdx,$sdy,$sdz,
						     [map {$no_lines} (1..12)],
						     $generate_covariance);
  my ($A_cube,$A_idealcube) = generate_cuboid_planes($meanx,$meany,$meanz,
						     $sdx,$sdy,$sdz,
						     [map {$no_planes} (1..6)],
						     $generate_covariance);
  
  return ("Points" => $X_cube,
	  "Lines" => $L_cube,
	  "Planes" => $A_cube,
	  "Ideal_Points" => $X_idealcube,
	  "Ideal_Lines" => $L_idealcube, 
	  "Ideal_Planes" => $A_idealcube);
}
push @EXPORT,"generate_cuboid";

sub transform_cuboid {
  my %args = (
	      # default values
	      "rotation" => [0,0,0],
	      "translation" => [1.0,1.0,1.0],
	      "verbose" => 0,
	      # argument pair list goes here
	      @_ 
	     );

  my ($omega,$phi,$kappa) = @{$args{rotation}};
  my ($tx,$ty,$tz) = @{$args{translation}};
  my $X_cube = $args{Points} || [];
  my $L_cube = $args{Lines} || [];
  my $A_cube = $args{Planes} || [];
  my $X_idealcube = $args{Ideal_Points} || [];
  my $L_idealcube = $args{Ideal_Lines} || [];
  my $A_idealcube = $args{Ideal_Planes} || [];
  my $verbose = $args{verbose};

  # now transform all the points, lines and planes
  my $T = mzeroes(4,4);
  my $R = mpdl([
		[cos($omega),-sin($omega), 0 ],
		[sin($omega), cos($omega), 0 ],
		[           0,          0, 1 ]
	       ]) x
		 mpdl([
		       [ cos($phi) ,0, sin($phi)  ],
		       [ 0         ,1,         0  ],
		       [-sin($phi), 0, cos($phi) ]
		      ]) x
			mpdl([
			      [1,          0,          0 ],
			      [0,cos($omega),-sin($omega)],
			      [0,sin($omega), cos($omega)]
			     ]);

  ($_ = $T->slice("0:3,3")) .= vpdl([$tx,$ty,$tz,1]);
  ($_ = $T->slice("0:2,0:2")) .= $R;

  my $Transform = new SUGR::Transformation($T,"Point");
  
  print "Transform with (omega,phi,kappa) = ($omega,$phi,$kappa) [rad]\n" if $verbose;
  print "          with (tx,ty,tz)        = ($ty,$ty,$tz)\n" if $verbose;
  foreach my $list (@$X_cube,@$L_cube,@$A_cube,
		    @$X_idealcube,@$L_idealcube,@$A_idealcube) {
    foreach my $object (@$list) {
      $object = $object->transform($Transform);
    }
  }
  print "done\n" if $verbose;

  return ("Points" => $X_cube,
	  "Lines" => $L_cube,
	  "Planes" => $A_cube,
	  "Ideal_Points" => $X_idealcube,
	  "Ideal_Lines" => $L_idealcube, 
	  "Ideal_Planes" => $A_idealcube);
}
push @EXPORT,"transform_cuboid";

sub generate_cuboid_planes {
  my ($meanx,$meany,$meanz,
      $sdx,$sdy,$sdz,
      $no_of_planes_per_surface,
      $generate_variance) = @_;
  
  $generate_variance = 1 if !defined($generate_variance);


  $no_of_planes_per_surface ||= [1,1,1,1,1,1];

  my $total_no = 0; $total_no += $_ foreach @{$no_of_planes_per_surface};
  return undef if $total_no == 0;


  my $max_points = SUGR::MyListUtil::max(@$no_of_planes_per_surface);
  my $no_of_points_per_corner = [ map { $max_points } (0..7) ];

  my ($X_cube,$X_idealcube) = generate_cuboid_points($meanx,$meany,$meanz,
						     $sdx,$sdy,$sdz,$no_of_points_per_corner,
						     $generate_variance);

  my @Cube = ([0,1,2], # first plane, defined by point indices.
	      [0,1,5], # ...
	      [1,2,6],
	      [2,3,7],
	      [3,0,4],
	      [4,5,6]
	     );

  my (@A,@A_ideal);
  for my $surface (0..5) {
    push @{$A_ideal[$surface]}, 
      new SUGR::Plane($X_idealcube->[$Cube[$surface]->[0]]->[0],
		 $X_idealcube->[$Cube[$surface]->[1]]->[0],
		 $X_idealcube->[$Cube[$surface]->[2]]->[0],
		);
    for my $i (1..$no_of_planes_per_surface->[$surface]) {
      push @{$A[$surface]}, 
  	new SUGR::Plane($X_cube->[$Cube[$surface]->[0]]->[$i-1],
		   $X_cube->[$Cube[$surface]->[1]]->[$i-1],
		   $X_cube->[$Cube[$surface]->[2]]->[$i-1],
		  );
    }
  }
  return ([@A],[@A_ideal]);
}

sub generate_cuboid_lines {
  my ($meanx,$meany,$meanz,
      $sdx,$sdy,$sdz,
      $no_of_lines_per_edge,
      $generate_variance) = @_;
  
  $generate_variance = 1 if !defined($generate_variance);

  $no_of_lines_per_edge ||= [1,1,1,1,
			     1,1,1,1,
			     1,1,1,1];

  my $total_no = 0; $total_no += $_ foreach @{$no_of_lines_per_edge};
  return undef if $total_no == 0;

  my $max_points = SUGR::MyListUtil::max(@$no_of_lines_per_edge);
  my $no_of_points_per_corner = [ map { $max_points } (0..7) ];

  my ($X_cube,$X_idealcube) = generate_cuboid_points($meanx,$meany,$meanz,
						     $sdx,$sdy,$sdz,$no_of_points_per_corner,
						     $generate_variance);

  my @Cube = ([0,1],[1,2],[2,3],[3,0], # lines 0..3, defined by
                                       # indices of points
	      [0,4],[1,5],[2,6],[3,7], # lines 4..7
	      [4,5],[5,6],[6,7],[7,4], # lines 8..11
	     );

  my (@L,@L_ideal);
  for my $edge (0..11) {
    push @{$L_ideal[$edge]}, new SUGR::Line($X_idealcube->[$Cube[$edge]->[0]]->[0],
				       $X_idealcube->[$Cube[$edge]->[1]]->[0]);
    
    for my $i (1..$no_of_lines_per_edge->[$edge]) {
      push @{$L[$edge]}, 
  	new SUGR::Line($X_cube->[$Cube[$edge]->[0]]->[$i-1],
  		  $X_cube->[$Cube[$edge]->[1]]->[$i-1]);
    }
  }
  return ([@L],[@L_ideal]);
}

sub generate_cuboid_points {
  my ($meanx,$meany,$meanz,
      $sdx,$sdy,$sdz,
      $no_of_points_per_corner,
      $generate_variance) = @_;
  
  $generate_variance = 1 if !defined($generate_variance);
  
  $no_of_points_per_corner ||= [1,1,1,1,
				1,1,1,1];

  my $total_no = 0; $total_no += $_ foreach @{$no_of_points_per_corner};
  return undef if $total_no == 0;

#   my @Cube = ([0,0,0], # coordinate of point
# 	      [1,0,0],
# 	      [1,1,0],
# 	      [0,1,0],
# 	      [0,0,1],
# 	      [1,0,1],
# 	      [1,1,1],
# 	      [0,1,1]);

  my @Cube = ([-0.5,-0.5,-0.5], # coordinate of point
	      [ 0.5,-0.5,-0.5],
	      [ 0.5, 0.5,-0.5],
	      [-0.5, 0.5,-0.5],
	      [-0.5,-0.5, 0.5],
	      [ 0.5,-0.5, 0.5],
	      [ 0.5, 0.5, 0.5],
	      [-0.5, 0.5, 0.5]);

  my (@X,@X_ideal);
  for my $corner (0..7) {
    # ideal point
    my $Cov = $generate_variance ? mpdl([[$sdx**2,0,0,0],
					 [0,$sdy**2,0,0],
					 [0,0,$sdz**2,0],
					 [0,0,0,   0]]) : undef;
    push @{$X_ideal[$corner]}, new SUGR::Point( [ 
					    $meanx*$Cube[$corner]->[0] ,
					    $meany*$Cube[$corner]->[1] ,
					    $meanz*$Cube[$corner]->[2] , 
					    1 ],$Cov);
  
    for my $i (1..$no_of_points_per_corner->[$corner]) {
#      print join(",",Math::Random::random_get_seed());
#      print " -> ",random_normal(1,0,$sdx),"\n";
      my $x = random_normal(1,0,$sdx);
      my $y = random_normal(1,0,$sdy);
      my $z = random_normal(1,0,$sdz);
#      print ($meanx*$Cube[$corner]->[0] + $x),"\n";
      push @{$X[$corner]}, 
	new SUGR::Point( [ $meanx*$Cube[$corner]->[0] + $x,
			   $meany*$Cube[$corner]->[1] + $y,
			   $meanz*$Cube[$corner]->[2] + $z, 
			   1 ],$Cov);
    }
  }
  
  return ([@X],[@X_ideal]);
}

return 1 if caller;

############################################################
#
# test code

my %Objects = generate_cuboid(covariances => 0);
%Objects = transform_cuboid(translation => [1,2,2],
			    rotation    => [0.5,0.7,1.3],
			    %Objects);
foreach my $Type (qw/Ideal_Points/) { #(keys %Objects) {
  print "-----------> $Type\n";
  map { print } @$_ foreach @{$Objects{$Type}};
}


%Objects = generate_cuboid_simple(covariances => 0,verbose=>0);
foreach my $Type (qw/Points/) {
  print "-----------> $Type\n";
  print  $_ foreach @{$Objects{$Type}};
}
