=head1 NAME

SUGR::Element -- base class for uncertain geometrical elements

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Element->VERSION

=head1 SYNOPSIS

  # not to be used directly

=head1 DESCRIPTION

This is an abstract base class for the geometrical objects such as
points, lines and planes, but also for transformations such as
homographies, projections etc.

=head1 METHODS

Common methods for all subclasses are listed below, for constructors
see the manual page of the actual subclasses, L<SEE ALSO>.

=over 4

=cut


package SUGR::Element;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = SUGR->VERSION;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request

use SUGR::Basic  qw/:ALL/;
use PDL;
use PDL::Matrix;
use PDL::Math;
use PDL::Slatec;
use Storable;

####################################################################
#
# Config variables
#

# counter for object ids
our $object_cnt=1;


############################################################
#
# Constructors
#

# this is a constructor hack for overloading parameters
sub new {
  my ($protoverse, @arg) = @_;
  my $class = ref($protoverse) || $protoverse;
  my ($s2,$s3) = ("",""); # initialize second and first argument string
  # now this is the overloading trick, it extracts the types of the
  # arguments and calls the according function.
  # types can be either PDL::Matrix or SUGR::(Point|Line|Plane)
  # or ARRAY
  (my $s1 = ref($arg[0])) =~ s/^.*:://; # strip leading subclasses
  ($s2 = ref($arg[1])) =~ s/^.*::// if defined($arg[1]);
  ($s3 = ref($arg[2])) =~ s/^.*::// if defined($arg[2]);
  $s3 =~ s/HASH//; # maybe we want to do some hash handling ? pass it on !
  my $code = "ConstructFrom$s1$s2$s3";
  # doing the following dispatching by hand may not be nice, but at the
  # moment I do not want to change the whole OO code
  my $coderef = $class->can($code); 
  croak("don't know method $code\n") if !defined($coderef);
  my $self = &$coderef(@arg); warn $@ if $@;

  $self->{id} = $class."_".($object_cnt++);

#  print $self->{id},"\n";

  bless $self,$class;
}

# constructor helpers from PDL matrices or from perl arrays
sub ConstructFromARRAY {
  my (@arg) = @_;
  ConstructFromMatrix(vpdl($arg[0]));
}
sub ConstructFromARRAYARRAY {
  my (@arg) = @_;
  ConstructFromMatrixMatrix(vpdl($arg[0]),mpdl($arg[1]));
}
sub ConstructFromARRAYMatrix {
  my (@arg) = @_;
  ConstructFromMatrixMatrix(vpdl($arg[0]),$arg[1]);
}
sub ConstructFromMatrixArray {
  my (@arg) = @_;
  ConstructFromMatrixMatrix($arg[0],mpdl($arg[1]));
}
sub ConstructFromMatrix {
  my (@arg) = @_;
  croak("no parameter vector given") if !defined($arg[0]);
  my ($r,$c) = $arg[0]->mdims;
  croak("first argument should be a (n,1) vector, is $arg[0]") if $c!=1;
  my $retval =  { vec => $arg[0], cov => undef };
  return $retval;
}

# sub ConstructFromHASH {
#   my ($arghash) = @_;
  
#   my $result = Estimate("Line",$arghash);

#   return $result->{Line};
# }


# this is the actual constructor !
sub ConstructFromMatrixMatrix {
  my ($vec,$cov) = @_;
  croak("no parameter vector given") if !defined($vec);
  my ($r,$c) = $vec->mdims;
  my $homo_nullspace = $vec;

  croak("parameter vector is no PDL::Matrix") if ref($vec) ne "PDL::Matrix";
  croak("first argument should be a (n,1) vector, is $vec") if $c!=1;
  croak("covariance matrix is no PDL::Matrix") if ref($cov) ne "PDL::Matrix";
  croak("homogeneous nullspace is no PDL::Matrix") if ref($homo_nullspace) ne "PDL::Matrix";

  my $retval = { vec => $vec, cov => 0 };

  # first we have to project the given covariance matrix on the
  # orthogonal space of the euclidean nullspace
  $retval->{cov} = ApplyConstraints($cov,$homo_nullspace);
  return $retval;
}

# how to copy objects
sub copy {
  my ($self,$enlarge) = @_;
  my ($vec,$cov);
  my ($origdim) = $self->vec->mdims;
  if (!defined($enlarge) || $enlarge == $origdim) {
    ($vec,$cov) = ($self->vec->copy, 
		   defined($self->cov) ? $self->cov->copy : undef );
  }
  else {
    # maybe we want to enlarge the vector and the matrix, this is a
    # hack for the estimation procedure
    $origdim--;
    ($vec,$cov) = (vzeroes($enlarge),mzeroes($enlarge,$enlarge));
    ($_ = $vec->slice("0:$origdim")) .= $self->vec->copy;
    ($_ = $cov->slice("0:$origdim,0:$origdim")) .= $self->cov->copy;
  }
  bless { vec => $vec, cov => $cov, id => ref($self)."_".($object_cnt++) }, ref($self);
}

# this is for Storable
# this is a Storable-hook to freeze value of a piddle by using
# get_dataref and additionally storing the dimenions of the piddle
sub STORABLE_freeze_bak {
  my ($self, $cloning) = @_;
  return if $cloning;         # Regular default serialization
  return $self->serialize;
} 

# hook for Storable to restore SUGR value with new id
sub STORABLE_thaw_bak {
  my ($obj, $cloning, $serialized) = @_;
  die "__PACKAGE__, __LINE__ no cloning please" if $cloning;
  return if $cloning;         # Regular default serialization

  $serialized =~ /^([^#]+)#([^#]+)#([^#]+)/;
  my $tmpobj = eval "new $1($2,$3)";
  $obj->{id} = "$1_".($object_cnt++);
  $obj->{vec} = $tmpobj->vec;	
  $obj->{cov} = $tmpobj->cov;	
  bless $obj, $1; # need this so that we now the correct subclass
}

# this is a hacky serialization method, used to be for Storable, now
# it stands on its own (see also perlify)
sub serialize {
  my ($self) = @_;
  my $s = sprintf "%s#%s",~$self->vec,(defined($self->cov) ? $self->cov : "undef");
  $s =~ s/([^\t\n \[\]]+)([ ]+)/$1,$2/g;
  $s =~ s/\[\n/\[/g;
  $s =~ s/(.+)\]\n/$1\],/g;
  $s =~ s/(.+),\]/$1\]/g;
  $s =~ s/[\s\n]//g;
  $s = "SUGR::".$self->type."#$s";
  return $s;
}

# this is for the Storable(3) module
#
# this is a Storable-hook to freeze value of a piddle by using
# get_dataref and additionally storing the dimenions of the piddle
sub STORABLE_freeze {
  my ($self, $cloning) = @_;
  return if $cloning;         # Regular default serialization
  my (undef,@sugrobject) = $self->perlify; 
  return Storable::freeze(\@sugrobject);
} 

# hook for Storable to restore SUGR value with new id
sub STORABLE_thaw {
  my ($obj, $cloning, $serialized) = @_;
  die "__PACKAGE__, __LINE__ no cloning please" if $cloning;
  return if $cloning;         # Regular default serialization
  
  my $class = ref($obj);
  my ($vec,$cov) = @{ Storable::thaw $serialized };
  $obj->{id}  = $class."_".($object_cnt++);
  $obj->{vec} = vpdl($vec);
  $obj->{cov} = $cov==0 ? undef : mpdl($cov);
#  bless $obj, $class; # need this so that we now the correct subclass
}

# this method converts the data structure of SUGR object into a simple
# representation of pure basic Perl data types. This is needed for the
# Storable, so that it can handle SUGR objects, This method returns a
# a 3-element array:
#
# - the first element is a string with the classname, 
#
# - the second is a ref to an array of numbers corresponding to the
# vector
#
# - the third is a ref to an array of arrays of numbers corresponding
# to the covariance matrix
sub perlify {
  my ($self) = @_;
  my $vec = [ $self->vec->list ];
  my $cov = defined($self->cov) ? [ map { [ $_->list ] } $self->cov->dog ] : 0;
  return ("SUGR::".$self->type,$vec,$cov);
}

############################################################
#
# methods
#

=item -

vec()

Returns the parameter vector of the element. The vector is an instance
of the <PDL::Matrix(3)|PDL::Matrix(3)> class, see PDL(3) for more
details on the Perl Data Language.

=cut 

sub vec {
  $_[0]->{vec};
}

=item -

matrix()

Returns the matrix representation of the object. The matrix is an instance
of the <PDL::Matrix(3)|PDL::Matrix(3)> class, see PDL(3) for more
details on the Perl Data Language.

=cut 

sub matrix {
  warn "call to abstract method matrix in SUGR::Element!";
  return undef;
}

=item -

cov()

Returns the covariance matrix of the parameter vector. The covariance
matrix is an instance of the <PDL::Matrix(3)|PDL::Matrix(3)> class,
see PDL(3) for more details on the Perl Data Language.

=cut 

sub cov {
  $_[0]->{cov};
}

=item -

setcov($cov)

This sets the covariance matrix of the parameter vector, taking a
PDL::Matrix object $cov as input. This is dangerous since you have to
take care that the operation that you are doing is valid.

=cut 

sub setcov {
  my ($self,$cov) = @_;
  $self->{cov} = ApplyConstraints($cov,$self->vec);
}


=item -

setvec($vec)

This sets the parameter vector, taking a PDL::Matrix object $cov as
input. This is dangerous since you have to take care that the
covariance matrix is compatible to your parameter vector.

=cut 

sub setvec {
  $_[0]->{vec} = $_[1];
}

=item -

id()

Returns a unique alphanumeric identifier of the element.

=cut 

sub id :lvalue {
  $_[0]->{id};
}

=item -

is_transform()

Returns true if the object is a transformation.

=cut 

sub is_transform {
  return 0;
}

# OLD: transform an element, provided we have a transformation
sub transform {
  my ($self,$Transformation) = @_;

  croak("argument is not a SUGR transformation") 
    if ref($Transformation) ne "SUGR::Transformation";

  return $Transformation->transform($self);
}

sub stringify { 
  my ($self) = shift;

  my $euclid_scale = $self->euclid_scale;
  (my $s = $self->id()." ") =~ s/SUGR:://;

  my $self_print = $euclid_scale != 0 ? $self->copy->euclidize : $self;
  
  $s .= ($self_print->vec->slice(",(0)"));
  ($euclid_scale != 0) ? 
    ($s .= " (scaled with ".sprintf("%4.3g",1/$euclid_scale).")") : 
      ($s .= "unscaled");


  $s .= defined($self_print->cov) ? $self_print->cov : " (no covariance)\n";
  return $s;
}

sub print {
  my ($self) = shift;
  print stringify($self);
}

sub stringifymaple {
  my ($self,@args) = @_;
  my $s = $args[0] || "";
  $s .= ":=" unless $args[0] eq "";
  $s .= "[".$self->vec->stringifymaple.(defined($self->cov) ? ",".$self->cov->stringifymaple : "")."]";
  return $s;
}
sub printmaple {
  print stringifymaple(@_).";\n";
}


#
# overloading operators
#
use overload "!="  => \&SUGR::Relation::op_Orthogonal,  # not agood idea, is it?!
             "<=>" => \&SUGR::Relation::op_Parallel,    # not agood idea, is it?!
             "=="  => \&SUGR::Relation::op_Equal,       # ok
             "<"   => \&SUGR::Relation::op_Incident,    # ok
             ">"   => \&SUGR::Relation::op_Incident,    # ok
             q{""} => \&stringify,
             "*"   => \&scale;


# scale projectively a SUGR element (i.e. the whole vector)
# (normal scale for vector, squared scale for covariance matrix)
sub scale {
  my ($self,$scale) = @_;
  ($self,$scale) = ($scale,$self) if ref($self) !~ "SUGR";
  $self->{vec} *= $scale;
  $self->{cov} *= $scale**2 if defined($self->cov);
  return $self;
}


# this function returns the new euclidean interpretation of a SUGR
# Element.

# TODO !?!

# euclidize shouldn't make a new Element. The calling function may
# copy the SUGR object before calling euclidize if wanted.

# BUT: maybe copying is correct since euclidization of covariance
# matrix implies an additional error (since we are normalizing the
# vector which is a non-linear function).
sub euclidize {
  my ($self) = shift;

  my $Cov_euclid = undef;
  if (defined($self->cov)) {
    my $J = $self->Jacobian_norm_euclid;
    $Cov_euclid = $J x $self->cov x ~$J;
  }
  my $tmp = $SUGR::Element::apply_constraints;
  $SUGR::Element::apply_constraints = 0;
  no strict "refs";
  # why copy?!?! see TODO
  my $Element = ref($self)->new($self->vec * (1/$self->euclid_scale),$Cov_euclid);
  $SUGR::Element::apply_constraints = $tmp;
  return $Element;
}


%EXPORT_TAGS = (ALL => [@EXPORT,@EXPORT_OK]);

return 1 if caller;

=back

=head1 STORABLE

For the <Storable(3)|Storable> module, we provide hooks such that a
SUGR object can be stored using the fancy persistence module,
available at a CPAN Archive near you.

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR::Point2d(3), SUGR::Line2d(3), SUGR::Point(3), SUGR::Line(3),
SUGR::Plane(3), SUGR::Homography2d(3), SUGR::Homography(3),
SUGR::ProjectiveCamera(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut
