=head1 NAME

SUGR::Estimation -- estimate entities based on relations

(old stuff ....)

=head1 VERSION

  This document refers to version 0.01

=head1 SYNOPSIS
 
=head1 DESCRIPTION

=head2 Overview

=head2 Constructurs and initializaton

=head2 Class and object methods

=head2 Something else? !

=head1 Literature

=head1 DIAGNOSTICS

=over 4

=item "error messages?!"

=back

=head1 BUGS

=head1 TODO

=head1 FILES

=head1 SEE ALSO

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000 Stephan Heuel All rights reserved. There is no
warranty. You are allowed to redistribute this software /
documentation under certain conditions. For details, see the file
COPYING in the PDL distribution. If this file is separated from the
PDL distribution, the copyright notice should be included in the file.

=cut

package SUGR::Estimation;
  
use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 0.01;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL::Matrix;
use PDL::Math;
use Statistics::GaussHelmert;
use Statistics::AlgebraicMinimization;
use SUGR::Basic qw/:ALL/;
use SUGR::Relation qw/:ALL/;

our %Config = ( 
	       # this chooses between several methods of estimation.
	       #
	       # == 0 means, that there are no constraints on the observations
	       #
	       # == 1 means that the constraints on the observations will be imposed
	       #      in a second step inside each iteration of the Gauss Helmert
	       #      model 
	       "constraints_on_observations" => 0,

	       # type of constraint which should be applied to the observations
	       "type_of_constraints" => "Euclidean", #"Sphere",

	       # if one means that we want algebraic minimization instead of
	       # GaussHelmert least squares
	       "algebraic" => 0,
	       
	       # do we want to scale euclidean parts so that
	       # homogeneous factor is larger and therefore no
	       # linearity problems occur for the covariance matrices 
	       "scale_euclidean_parts" => 1,

	       # verbose output ?
	       "verbose" => 0,
	      );

# store types of observation into an array
my @observation_types = 
  qw/EqualPoint2d EqualLine2d EqualLine EqualPoint EqualPlane
     IncidentPoint2d IncidentLine2d  IncidentPoint IncidentLine IncidentPlane  
     ParallelLine2d ParallelPlane ParallelLine 
     OrthogonalLine2d OrthogonalLine OrthogonalPlane

     IncidentPointPoint2d IncidentPoint2dPoint
     IncidentLineLine2d IncidentLine2dLine
  
     EqualPointPoint
  /;
#
#
# Estimate( ... )
#
# Estimate an arbitrary SUGR object by given SUGR objects and their
# constraints.  First argument specifies type of unknown
# object. 
#
# $unknown_type = "Point2d"
#
# Second argument is a reference to a hash
#
# %constraints=(
#    "EqualPoint2d"     => [ $X_1, ... , $X_n  ]
#    "IncidentLine2d"   => [ $L_1, ... , $L_m  ]
#              )
#
# where $X_i and $L_i are SUGR objects, which are
# (incident|equal|parallel|orthogonal) to the unknown object.
#
sub Estimate {
  my ($unknown_type,$observationhash,$argumenthash) = @_;

  #
  # process arguments

  # filter constraints out of hash
  my %constraints;
  foreach (@observation_types) {
    if (defined($observationhash->{$_})) {
      $constraints{$_} = delete $observationhash->{$_};
      # check if there is anything in the constraint
      if (defined(@{$constraints{$_}})) {
	print " ... $_ \n" if $Config{verbose};
      }
      else {
	delete $constraints{$_} 
      }
    }
  }
  # warn for unknown options in argument hash
  foreach (keys %$observationhash) {
    warn "*** Warning: Unknown observation type to Estimate: $_\n";
  }

  # do we want an algebraic minimization only ?
  my $algebraic_minimization = delete $argumenthash->{algebraic};
  $algebraic_minimization = $Config{algebraic} if !defined($algebraic_minimization);

  # do we want to impose constraints on the observations?
  my $obs_constraints = delete $argumenthash->{constraints_on_observations};
  $obs_constraints = $Config{constraints_on_observations} if !defined($obs_constraints);

  # do we want to have euclidean constraints or spherical constraints
  # on both the observations and the unknowns
  my $constraints_type = delete $argumenthash->{type_of_constraints};
  $constraints_type ||= $Config{type_of_constraints};

  my $verbose = delete $argumenthash->{verbose};
  $verbose = $Config{verbose} if !defined($verbose);

  my $scale_euclidean_parts = delete $argumenthash->{scale_euclidean_parts};
  $scale_euclidean_parts = $Config{scale_euclidean_parts} if !defined($scale_euclidean_parts); 

  # warn for unknown options in argument hash
  foreach (keys %$argumenthash) {
    warn "*** Warning: Unknown option to Estimate: $_\n";
  }
  # done with argument processing
  #
  ##############################################

  print "Start estimating $unknown_type object with observations ...\n" if $verbose;

  # number of unknowns 
  my $nU = $no_of_dimensions{$unknown_type};

  #
  # this determines which Jacobian is the reference for the selection
  # of linearly independent rows
  my $fixedJacobian = ($unknown_type =~ /Line$/) ? 0 : 1;
  $fixedJacobian = 2 if ($unknown_type =~ /ProjectiveCamera/ || 
			 $unknown_type =~ /Homography/);

  # maybe we have a trilinear constraint, because a projection matrix
  # is involved. Then we need to know this, actually for each
  # observation group. Therefore this is an array
  my @trilinear;

  #
  # build observation vector, covariance matrix, also
  # count number of observations
  #
  # @y_e this is an array of arrays. Each subarray contains the list of
  # constraint element vectors. 
  #
  # @y_e = ( [ X_1->vec, ... , X_n->vec ], 
  #          [ L_1->vec, ... , L_m->vec ] ) 
  #  
  # where X_1 could be incident points and L_1 could be incident
  # lines. Similar things hold for @Sigma_yy_e. @y_types stores the
  # types of constraints 
  my (@y_e,@y_length_squared,@y_types,@y_obs_types,@Sigma_yy_e);
  my $no_of_contradictions=0;
  my $no_of_contradictions_obs=0;
  my ($max_norm_euclid,$max_norm_hom,$euclidean_scale_factor) = (0,0,1);

  # first determine largest euclidean factor, for which we want to
  # scale the euclidean parts.
  if ($scale_euclidean_parts) {
    while ( my ($constraint_type,$elementlist) = each %constraints) {
      push @trilinear,0; # default value for trilinear constraint
                         # flag. 
      # we do not yet know how to scale consistently the euclidean
      # parts of an image point, a space point and a projection
      # matrix, so we skip this one
      if ($constraint_type =~ /ProjectiveCamera/ ||
	  $unknown_type =~ /ProjectiveCamera/ ||
	  $unknown_type =~ /Homography/ ) {
	$trilinear[-1] = 1;
	next;
      }
      my $max = homo_euclid_factor() * 
	max_norm_of_euclidean_part(@$elementlist);
      $max_norm_euclid = $max if ($max_norm_euclid<$max);
      $max = max_norm_of_homogeneous_part(@$elementlist);
      $max_norm_hom = $max if ($max_norm_hom<$max);
    }
    $euclidean_scale_factor = (1/$max_norm_euclid) if ($max_norm_euclid > $max_norm_hom);
    print "Scale euclidean part of observations by $euclidean_scale_factor...\n"   if $verbose;
  } 

  print "Building observation vectors and covariances ... \n"   if $verbose;
  while ( my ($constraint_type,$elementlist) = each %constraints) {
    # if wanted by the user, we scale the euclidean parts of the
    # homogeneous vectors so that the norm of the homogeneous part is
    # larger than the euclidean part (this will guarantee that the
    # linearization of the covariances of the constraints are good enough)
    if ($scale_euclidean_parts && $euclidean_scale_factor !=1) {
      @$elementlist = map { $_->copy->scale_euclidean_part($euclidean_scale_factor) } @$elementlist;
    }

    # exception handling for constraints involving trilinear
    # constraints, we need to cat together the observation vectors and
    # covariance matrices
    if ($constraint_type =~ /ProjectiveCamera/ ||
	$unknown_type =~ /ProjectiveCamera/ ||
	$unknown_type =~ /Homography/) {

      push @y_e, [ map { vpdl(PDL::append($elementlist->[$_*2]->vec->clump(-1),
					  $elementlist->[$_*2+1]->vec->clump(-1))) } (0..(@$elementlist/2 -1)) ];
      push @y_length_squared, [ map {  (PDL::sum($elementlist->[$_*2]->vec **2) +
 					PDL::sum($elementlist->[$_*2 +1]->vec **2))   } (0..(@$elementlist/2 -1)) ];
#      push @y_length_squared, [ map { PDL::sum($_ **2) } (@{$y_e[-1]}) ];
      
#      print "$_\n" foreach @{$y_length_squared[-1]};
#      die;
      # number of contradictions for _all_ observations in this group
      $no_of_contradictions += @$elementlist/2 * DegFreedom("$constraint_type$unknown_type");

      my @Covlist;
      foreach my $i (0..(@$elementlist/2 -1)) {
	my ($Cov1,$Cov2);
	if ($constraints_type =~ /Sphere/) {
	  $Cov1 = ApplyConstraints($elementlist->[$i*2]->cov,
				   $elementlist->[$i*2]->vec,1);
	  $Cov2 = ApplyConstraints($elementlist->[$i*2+1]->cov,
				   $elementlist->[$i*2+1]->vec,1);
	}
	else {
	  $Cov1 = $elementlist->[$i*2]->cov; 
	  $Cov2 = $elementlist->[$i*2+1]->cov; 
	}
	my ($Cov1r,undef,$Cov2r) = ($Cov1->mdims,$Cov2->mdims);
	my $Cov12r = $Cov1r+$Cov2r;
	push @Covlist, mzeroes($Cov12r,$Cov12r);
	($_ = $Covlist[-1]->slice("0:".($Cov1r-1).",0:".($Cov1r-1))) .= $Cov1;
	($_ = $Covlist[-1]->slice("$Cov1r:".($Cov12r-1).",$Cov1r:".($Cov12r-1))) .= $Cov2;
      } 
      push @Sigma_yy_e , [@Covlist];

#       print "$_\n" foreach @{$y_e[-1]};
#       print "$_\n" foreach @{$Sigma_yy_e[-1]};die;
    }
    else {
      # only points, lines or planes are involved
      push @y_e, [ map { $_->vec } (@$elementlist) ];
      push @y_length_squared, [ map { PDL::sum($_->vec **2) } (@$elementlist) ];

      # number of contradictions for _all_ observations in this group
      $no_of_contradictions += @$elementlist * DegFreedom("$constraint_type$unknown_type");

      if ($constraints_type =~ /Sphere/) {
	push @Sigma_yy_e , [ map { ApplyConstraints($elementlist->[$_]->cov,
						    $elementlist->[$_]->vec,1) } (0..$#{@$elementlist}) ];
      }
      else {
	push @Sigma_yy_e , [ map { defined($elementlist->[$_]->cov) ?
				   $elementlist->[$_]->cov : Identity($elementlist->[$_]->vec) 
				 } (0..$#{@$elementlist}) ];
      }
    }

    push @y_types, $constraint_type;
    (my $observation_type = $constraint_type) =~ s/(Incident|Equal|Orthogonal|Parallel)//;
    push @y_obs_types, $observation_type;

    # maybe we want to consider the constraints on the observation.
    if ($obs_constraints==1) {
      $no_of_contradictions_obs += @$elementlist *
	$no_of_constraints{$observation_type}; 
    }
    
    print "- type: $constraint_type, " if $verbose;
    print "Constraint on observations: $constraints_type\n" if $verbose;
  }
  
  print "done\n" if $verbose;
  print "number of contradictions $no_of_contradictions\n"   if $verbose;

  if ($no_of_contradictions + $no_of_constraints{$unknown_type} < $nU ) {
    warn "*** Warning: ",
      __FILE__," line ",__LINE__,"\n",
	"    # of unknowns is larger than # of obs. + # of constraints\n";
    return undef;
  } 

  # turn previously constructed array of arrays into an array of
  # multi-piddles using cat. This way we can do the computation in the
  # Gauss Hellmert estimation more efficiently
  my @y		   = map { PDL::cat(@{$_}) } (@y_e);
  my @Sigma_yy	   = map { PDL::cat(@{$_}) } (@Sigma_yy_e);
  
  # here we have the Jacobian of the constraints $w$ w.r.t. to unknown
  # vector $\beta$. This has to be implemented as a function, since
  # the Jacobian depends on $y$ resp. $\beta$.
  my $func_A = sub {
    my ($y,$b,$select_rows_flag) = @_; # input is a reference to the observation array
                                       # and $b is the unknown vector

#    print "start A\n";
    # array of multi-mpiddles where the Jacobians for each observation
    # group are defined
    my @Agroup;
    
    # build Jacobians for each group, @Agroup is an array of arrays,
    # where the subarray contain the actual Jacobian for a geometric
    # constraint with one object
    foreach my $groupi (0..$#{@$y}) {
      my @ygroup = PDL::dog( $y->[$groupi] ); # dog is the opposite of cat, converts
                                              # a multipiddle to an array
      push @Agroup,
	[ 
	 map {
	   my $i = $_;
	   my @yi = ($ygroup[$i]); # save the current element, also improving
                                   # readability

	   my $jacobian_index = 1;

	   # some corrections needed in case of trilinear constraints
	   if ($trilinear[$groupi]) {
	     push @yi, PDL->new(0);
	     $jacobian_index++;
	   }

	   # this gives strings like "EqualPoint2DPoint2d"
	   my $constraint_type = "$y_types[$groupi]$unknown_type";

	   # determine function name for computing Jacobian
	   my $Jacobian_func = "SUGR::Relation::Jacobian_$constraint_type";
	   
	   # we may want to select to linear independent rows,
	   # selection must be the same for Jacobian A and B !
	   my $degree_of_freedoms = DegFreedom("$constraint_type");

	   no strict "refs";

	   # the user can determine whether she wants only the
	   # linearly independent rows or all rows of the Jacobian,
	   # this can be done with the $select_rows_flag
	   my $J;
	   if (!defined($select_rows_flag)) {
#	     print "Here we go! $jacobian_index\n";
#	     die "*** Error in building A: $fixedJacobian != $jacobian_index\n" unless $fixedJacobian == $jacobian_index;
	     my @linear_indep_rows = ChooseRowIndices(&$Jacobian_func(@yi,$b,$fixedJacobian),
						      $degree_of_freedoms);
#	     print "Here we go again with ",join(",",@linear_indep_rows),"\n";
#	     print "Jacobian : <",&$Jacobian_func(@yi,$b,$jacobian_index),"> indep rows: ",join(",",@linear_indep_rows)," b: <$b>\n";
	     $J = ChooseRows(&$Jacobian_func(@yi,$b,$jacobian_index),@linear_indep_rows); 
#	     print "done\n"; 
	   }
	   else {
#	     print "**********)$Jacobian_func\n",join(",",@yi,$b,$fixedJacobian),"(***********\n";
	     $J = &$Jacobian_func(@yi,$b,$jacobian_index);
	   }

	   $J; # that's it
	 } (0..$#ygroup)
	];
    }
#    print "A",(map { PDL::cat(@$_) } @Agroup);
#    print "end A\n";
    return map { PDL::cat(@$_) } @Agroup ;
  };
  
  # here we have the Jacobian of the constraints $w$ w.r.t. to the observations $y$.
  my $func_B = sub {
    my ($y,$b) = @_; # input is a reference to the observation array
                     # and $b is the unknown vector

#    print "start B\n";
    # array of multi-mpiddles where the Jacobians for each observation
    # group are defined
    my @Bgroup;
    
    # build Jacobians for each group, @Bgroup is an array of arrays,
    # where the subarray contain the actual Jacobian for a geometric
    # constraint with one object
    foreach my $groupi (0..$#{@$y}) {
      my @ygroup = PDL::dog( $y->[$groupi] ); # dog is the opposite of cat, converts
					 # a multipiddle to an array
      push @Bgroup,
	[ 
	 map {
	   my $i=$_;
	   my @yi = ($ygroup[$i]); # save the current element, also improving
                                   # readability

	   push @yi, PDL->new(0) if $trilinear[$groupi];

	   my $jacobian_index = 0; # the first variable is the
                                   # observation, see definition of $Jacobian_func

	   # this gives strings like "EqualPoint2DPoint2d" depending
	   # on the observation group and the type of the unknown object
	   my $constraint_type = "$y_types[$groupi]$unknown_type";
	   
	   # determine function for computing Jacobian
	   my $Jacobian_func = "SUGR::Relation::Jacobian_$constraint_type";
	   
	   # we may want to select to linearly independent rows,
	   # selection must be the same for Jacobian A and B !
	   my $degree_of_freedoms = DegFreedom("$constraint_type");
	   no strict "refs";
	   
	   # select the linearly independent rows according to
	   # $func_A, so that the Jacobian $A$ and $B$ are
	   # consistent. This is why there is the "0" as the third
	   # argument of &$Jacobian_func
	   my @linear_indep_rows = ChooseRowIndices(&$Jacobian_func(@yi,$b,$fixedJacobian),
						      $degree_of_freedoms);
	   
	   my $J = ChooseRows(&$Jacobian_func(@yi,$b,$jacobian_index),@linear_indep_rows);

	   $J; # that's it
	 } (0..$#ygroup)
	];
    }
    
#    print "B",(map { PDL::cat(@$_) } @Bgroup),"\n";
#    print "end B\n";
    return map { PDL::cat(@$_) } @Bgroup ;
  };

  # now the function for the contradictions
  my $func_w = sub {
    my ($y,$b) = @_;

#    print "start w\n";
    return $no_of_contradictions if !defined($y);   # return number of constraints
						    # if called with
						    # no argument 
    my @w;
    # for each group of observations
    foreach my $groupi (0..$#{@$y}) {
      my @ygroup = PDL::dog( $y->[$groupi] ); # dog is the opposite of cat, converts
                                         # a multipiddle to an array
      my @wgroup; 
      # for each element in an observation group
      for my $i (0..$#ygroup) {
	my @yi = ($ygroup[$i]); # save the current  element, also improving
	                        # readability
	
	push @yi, PDL->new(0) if $trilinear[$groupi];
	
	my $jacobian_index = 0; 

	# this gives strings like "EqualPoint2DPoint2d" depending
	# on the observation group and the type of the unknown object
	my $constraint_type = "$y_types[$groupi]$unknown_type";
	my $observation_type = $y_obs_types[$groupi];
	
	# determine function for computing Jacobian
	my $Jacobian_func = "SUGR::Relation::Jacobian_$constraint_type";
	
	# we want to select linearly independent rows,
	# selection must be the same for Jacobian A and B !
	my $degree_of_freedoms = DegFreedom("$constraint_type");
	no strict "refs";
	
	# select the linearly independent rows according to
	# $func_A, so that it is consistent with  Jacobian $A$
	# and $B$. This is why there is $fixedJacobian as
	# the third argument of &$Jacobian_func
	my @linear_indep_rows = ChooseRowIndices(&$Jacobian_func(@yi,$b,$fixedJacobian),
						 $degree_of_freedoms);
	my $J = ChooseRows(&$Jacobian_func(@yi,$b,$jacobian_index),@linear_indep_rows);
	
	my $w = $J x $ygroup[$i];
	push @wgroup,$w; # that's it
      }
      push @w, PDL::cat(@wgroup);
    }
#    print "end w\n";

    return @w;
  };
  
  # constraint on unknowns 
  my $func_h = sub {
    # return no of constraints if called with no argument
    return $no_of_constraints{$unknown_type} if !defined($_[0]);
    # otherwise just call the appropriate function and we are done
    my $func = "SUGR::$unknown_type"."::Constraint_$constraints_type";
    no strict "refs";
    my $ret = &$func(@_);
    return $ret;
  };
  # and the Jacobian of the constraints $h$.
  my $func_H = sub { 
    # just call the appropriate function and we are done
    my $func = "SUGR::$unknown_type"."::Jacobian_Constraint_$constraints_type";
    no strict "refs";
    my $ret = &$func(@_);
    return $ret;
  };

  # constraints only on observations
  my $func_g = ($obs_constraints != 1) ? undef : sub {
    my ($y) = @_;
    return $no_of_contradictions_obs if !defined($y);   # return number of constraints
						        # if called with
						        # no argument 
    my @g;
    # for each group of observations
    foreach my $groupi (0..$#{@$y}) {
      my @ygroup = PDL::dog( $y->[$groupi] ); # dog is the opposite of cat, converts
                                         # a multipiddle to an array
      my @ggroup; 
      # for each element in an observation group
      for my $i (0..$#ygroup) {
	my @yi = ($ygroup[$i]); # save the current  element, also improving
	                        # readability
	
	# this gives strings like "Point2d" depending
	# on the observation group 
	my $observation_type = $y_obs_types[$groupi];
	
	# determine function for computing the constraint
	my $func_constraint_obs = "SUGR::".$observation_type."::Constraint_$constraints_type";
	
	# compute the constraint
	no strict "refs";
	my $g = &$func_constraint_obs(@yi,$y_length_squared[$groupi]->[$i]);

	push @ggroup,$g; # that's it
      }
      push @g, PDL::cat(@ggroup);
    }
    return @g;
  };
  
  # and the Jacobian of the constraints $h$.
  my $func_G =  ($obs_constraints != 1) ? undef : sub { 
    my ($y) = @_;

    my @Ggroup;

    # for each group of observations
    foreach my $groupi (0..$#{@$y}) {
      my @ygroup = PDL::dog( $y->[$groupi] ); # dog is the opposite of cat, converts
      # a multipiddle to an array
      push @Ggroup,
	[ 
	 map {
	   my $yi = $_; # save the current  element, also improving
	   # readability
	   my $observation_type = $y_obs_types[$groupi];
	   my $Jacobian_func = "SUGR::".$observation_type."::Jacobian_Constraint_$constraints_type";
	   no strict "refs";	     
	   my $J = &$Jacobian_func($yi);
	   $J; # that's it
	 } @ygroup
	];
    }
    return map { PDL::cat(@$_) } @Ggroup ;
  };

  #
  # calculate approximate values
  #
  my $GtG = mzeroes($nU,$nU);
  my @A = &$func_A([@y],PDL->new(0),1);
  foreach my $i (0..$#y) {
    $GtG += msumover(~$A[$i] x $A[$i]);
  }  

  # compute smallest eigenvector 
  my ($evector,$evalue) = eigens($GtG);
#  print "eigenproblem for approximate value: ",$GtG,$evector,$evalue,"\n";
  my $smallest_index = (qsorti $evalue)->at(0); # ... of eigenvalue vector
  
  # eigenvector for smallest eigenvalue is the approximate value
  my $b0 = vpdl( $evector->slice(",($smallest_index)" ));

  if ($verbose) {
    my $b0_sugr = eval "SUGR\:\:$unknown_type\:\:sugr_$unknown_type(\$b0)"; die $@ if $@;
    if ($scale_euclidean_parts && $euclidean_scale_factor !=1) {
      $b0_sugr->scale_euclidean_part(1/$euclidean_scale_factor);
    }
    print "Approximate value = $b0_sugr\n";
  }


  #
  # now do estimation according to the Gauss-Helmert model
  # 
  # setup arguments
  my $args = {"observations" => [@y],
	      "initial_guess" => $b0,
	      "covariance_observations" => [@Sigma_yy],
	      "observation_equations" => $func_w,
	      "Jacobian_unknowns" => $func_A,
	      "Jacobian_observations" => $func_B,
	      "constraints" => $func_h,
	      "Jacobian_constraints" => $func_H,
	      "constraints_observations" => $func_g,
	      "Jacobian_constraints_observations" => $func_G
	     };

  # choose between GaussHelmert model or algebraic minimization
  my $result = ($algebraic_minimization) ? 
    AlgebraicMinimizationBlocks($args) : GaussHelmertBlocks($args);

  my $b_est = delete $result->{estimated_unknown};
  my $sigma0_squared = delete $result->{sigma0_squared};
  my $Sigma_bb_est = delete $result->{covariance_unknown};
  my $y_est = delete $result->{estimated_observations};
  my $b_est_i = delete $result->{estimated_unknowns_iterations};
  my $no_iter = delete $result->{number_of_iterations};

  return undef if !defined($b_est);

  print "computed in $no_iter iterations.\n"  if ($verbose);
  print "Estimated value = ",$b_est->slice(",(0)"),"\n" if ($verbose);
  print "sigma_0^2 = $sigma0_squared\n"  if ($verbose);
  
  # generate SUGR objects for estimated unknown
  my $Sigma_bb_est_sigma0 = $Sigma_bb_est * $sigma0_squared;
  my $func_constructor = "SUGR::".$unknown_type."::sugr_$unknown_type";
  no strict "refs";
  my $b_est_sugr = &$func_constructor($b_est ,$Sigma_bb_est_sigma0);
  my $b0_sugr = &$func_constructor($b0);

  my @b_est_i_sugr; # temporary ?!
  for my $i (0..($no_iter-1)) {
    push @b_est_i_sugr,
      eval "SUGR\:\:$unknown_type\:\:sugr_$unknown_type(\$b_est_i->[$i])";
  }

  # generate SUGR objects for estimated observations
  my %y_est_sugr;
  if (!$algebraic_minimization) {
    for my $groupid (0..$#y) {
      next if $trilinear[$groupid];
      my @ygroup = PDL::dog($y_est->[$groupid]);
      
      for my $i (0..$#ygroup) {
	my $observation_type = $y_obs_types[$groupid];
	my $constraint_type = $y_types[$groupid];
	my $y_est_i = $ygroup[$i];
	
	push @{$y_est_sugr{$constraint_type}} , 
	  eval "SUGR\:\:$observation_type\:\:".
	    "sugr_$observation_type(\$y_est_i ,\$Sigma_yy_e[$groupid]->[$i])";
	die $@ if $@;
      }
    }
  }

  if ($scale_euclidean_parts && $euclidean_scale_factor !=1) {
    $b_est_sugr->scale_euclidean_part(1/$euclidean_scale_factor);
    @b_est_i_sugr = map {  $_->scale_euclidean_part(1/$euclidean_scale_factor) } @b_est_i_sugr;
    foreach my $y_est_list (values %y_est_sugr) {
      @$y_est_list = map {  $_->scale_euclidean_part(1/$euclidean_scale_factor) } @$y_est_list ;
    }
  }

  return { "$unknown_type" => $b_est_sugr,
	   "initial_$unknown_type" => $b0_sugr,
	   "sigma0_squared" => $sigma0_squared,
	   "Sigma_unknown" => $Sigma_bb_est,
	   "estimated_observations" => \%y_est_sugr,
	   "estimated_unkowns_iterations" => \@b_est_i_sugr,
	   "number_of_iterations" => $no_iter
	 }
}
push @EXPORT,"Estimate";

return 1 if caller;

################################################################################
#
# test code

# as a test, generate elements of a cuboid and estimate a point,
# line and a plane
use SUGR::Cuboid;

my $random_seed = [4711,222712];
my $sd_points = 0.01;
my %Objects = generate_cuboid(no_points => 5,
			      no_planes => 5,
			      no_lines  => 5,
			      standard_deviation => [$sd_points,$sd_points,$sd_points],
			      random_seed => $random_seed);
# %Objects = transform_cuboid(translation => [1,2,2],
# 			    rotation    => [0.5,0.7,1.3],
# 			    %Objects);

my $X_cube        = $Objects{Points};
my $L_cube        = $Objects{Lines};
my $A_cube        = $Objects{Planes};
my $X_cube_ideal  = $Objects{Ideal_Points};
my $L_cube_ideal  = $Objects{Ideal_Lines};
my $A_cube_ideal  = $Objects{Ideal_Planes};

################################################################################
#
# Estimation of Points, Lines and Planes
#
# estimate point  of corner 0
my %constraints_point = (
			 "EqualPoint"      => [@{$X_cube->[0]}],
			 
			 "IncidentPlane"   => [@{$A_cube->[0]},
					       @{$A_cube->[1]},
					       @{$A_cube->[4]}],
			 
			 "IncidentLine"    => [@{$L_cube->[0]},
					       @{$L_cube->[3]},
					       @{$L_cube->[4]}
					      ]
			);

# estimate plane of surface 0
my %constraints_plane = ("EqualPlane"      => [@{$A_cube->[0]}],
			 
			 "OrthogonalPlane" => [@{$A_cube->[1]},
					       @{$A_cube->[2]},
					       @{$A_cube->[3]},
					       @{$A_cube->[4]}],
			 
			 "ParallelPlane"   => [@{$A_cube->[5]}],
			 
			 "IncidentPoint"   => [@{$X_cube->[0]},
					       @{$X_cube->[1]},
					       @{$X_cube->[2]},
					       @{$X_cube->[3]}],
			 
			 "IncidentLine"    => [@{$L_cube->[0]},
					       @{$L_cube->[1]},
					       @{$L_cube->[2]},
					       @{$L_cube->[3]}
					      ],
			 
			 "ParallelLine"    => [@{$L_cube->[8 ]},
					       @{$L_cube->[9 ]},
					       @{$L_cube->[10]},
					       @{$L_cube->[11]}],
			 
			 "OrthogonalLine"  => [@{$L_cube->[4]},
					       @{$L_cube->[5]},
					       @{$L_cube->[6]},
					       @{$L_cube->[7]}]
			);

# estimate line of edge 0
my %constraints_line = (
			"EqualLine"      => [@{$L_cube->[0]}],
			
			"IncidentPlane"   => [@{$A_cube->[0]},
					      @{$A_cube->[1]}],
			
			"OrthogonalPlane" => [@{$A_cube->[2]},
					      @{$A_cube->[4]}],
			
			"ParallelPlane"   => [@{$A_cube->[3]},
					      @{$A_cube->[5]}],
			
			"IncidentPoint"   => [@{$X_cube->[0]},
					      @{$X_cube->[1]}],
			
			"IncidentLine"    => [@{$L_cube->[1]},
					      @{$L_cube->[3]},
					      @{$L_cube->[4]},
					      @{$L_cube->[5]}],
			
			"ParallelLine"    => [@{$L_cube->[2]},
					      @{$L_cube->[10]},
					      @{$L_cube->[8]}],
			
			"OrthogonalLine"  => [@{$L_cube->[1]},
					      @{$L_cube->[3]},
					      @{$L_cube->[4]},
					      @{$L_cube->[5]},
					      @{$L_cube->[6]},
					      @{$L_cube->[7]},
					      @{$L_cube->[9]},
					      @{$L_cube->[11]}]
		       );
$Config{verbose}=1;

# Estimate("Point",\%constraints_point);
# Estimate("Line", \%constraints_line );
# Estimate("Plane",\%constraints_plane);

# exit;

############################################################
#
# Estimation of an Homography
#
use SUGR::Homography;

#my $H = sugr_Homography(msequence(4,4) x ~msequence(4,4) ) ; #mpdl([[1,2,3,-5],[4,5,6,-3],[7,8,9,-7],[0,0,0,1]]));
my $H = sugr_Homography(mpdl [[2,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]]);

my @Xs;
foreach my $i (0..7) {
  push @Xs, new SUGR::Point($H->transform($X_cube->[$i]->[0])->vec,
			    mpdl([[$sd_points**2,0,0,0],[0,$sd_points**2,0,0],[0,0,$sd_points**2,0],[0,0,0,0]]));
  print $Xs[-1];
}
my $result_H = Estimate("Homography",
		      {
		       "EqualPointPoint" => 
		       [ map { my $i = $_; ($X_cube_ideal->[$i]->[0],$Xs[$i])
			     } (0..7) 
		       ] 
		      },
		      {
		       "type_of_constraints" => "Sphere",
		       "verbose" => 1
		      });

print $result_H->{Homography}->vec->reshape(4,4) / $result_H->{Homography}->vec->at(15);
print "Differences between projected point and ideal point\n";
foreach my $i (0..7) {
  my $Diff = 
    $result_H->{Homography}->transform($X_cube_ideal->[$i]->[0])->euclidize->vec -
      $Xs[$i]->euclidize->vec;
  print PDL::sum($Diff->slice("0:2")**2);
  print "\n";
}

die;

############################################################
#
# Estimation of Projection Matrices
#

#use SUGR::ProjectiveCamera;
#use SUGR::Point;
#use SUGR::Line;
#use SUGR::Point2d;
#use SUGR::Line2d;
use Math::Random qw(random_normal
		    random_set_seed);
use SUGR::MyListUtil qw/max min/;

my $X0 = new SUGR::Point([1.5,1.5,10,1],
		    0.01**2*mpdl([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,0]]));
my $R  = mpdl([[1,0,0],
	       [0,1,0],
	       [0,0,1]]);
my $P = new SUGR::ProjectiveCamera({ 
				       "X0"    => $X0,
				       "R"     => $R,
				       "c"     => 1,
				       "c_var" => 0.001,
				      });
my (@xs,@ls);

print "Projecting cube points ...\n";
foreach my $X_cubepoint (map { $_->[0] } (@$X_cube_ideal)) {
  push @xs, new SUGR::Point2d($X_cubepoint,$P)->euclidize; 
  print $xs[-1]->vec->slice("0:1")->clump(-1),"\n";
}

my $xmin = min map { $_->vec->at(0) } @xs;
my $ymin = min map { $_->vec->at(1) } @xs;
my $xmax = max map { $_->vec->at(0) } @xs;
my $ymax = max map { $_->vec->at(1) } @xs;
my $sd = 0.01*max(abs($xmax-$xmin),abs($ymax-$ymin));
print "pertubing projecting points with sdev $sd\n";
random_set_seed(@$random_seed);
foreach my $xs (@xs) {
  ($_ = $xs->vec->slice("0:1")) += vpdl([1/$xs->vec->at(2) * random_normal(2,0,$sd)]);
  print $xs->vec->slice("0:1")->clump(-1),"\n";
}


print "Projecting cube lines ...\n";
foreach my $L_cubeline (map { $_->[0] } (@$L_cube_ideal)) {
  push @ls, new SUGR::Line2d($L_cubeline,$P)->euclidize; 
  print $ls[-1]->vec->slice("0:2")->clump(-1),"\n";
}

my $csphimin = min map { $_->vec->at(0),$_->vec->at(1) } @ls;
my $csphimax = max map { $_->vec->at(0),$_->vec->at(1) } @ls;
my $dmin = max map { $_->vec->at(2) } @ls;
my $dmax = max map { $_->vec->at(2) } @ls;
my $sd_d = 0.01*abs($dmax-$dmin);
my $sd_csphi = 0.01*abs($csphimax-$csphimin);
print "pertubing projecting lines with sdev for distance: $sd and for cos/sin(phi): $sd_csphi\n";
random_set_seed(@$random_seed);
foreach my $ls (@ls) {
  ($_ = $ls->vec->slice("0:1")) += vpdl([random_normal(2,0,$sd_csphi)]);
  ($_ = $ls->vec->slice("2:2")) += vpdl([random_normal(1,0,$sd_d)]);
  print $ls->vec->slice("0:2")->clump(-1),"\n";
}

print "P(Original) :$P";


my $result_x = Estimate("ProjectiveCamera",
		      {
		       "IncidentPoint2dPoint" => 
		       [ map { $xs[$_], $X_cube->[$_]->[0],   } (0..7) ]  #$#{@$X_cube}) ]
		      },
		      {
		       "type_of_constraints" => "Sphere",
		       "verbose" => 1
		      });

my $result_l = Estimate("ProjectiveCamera",
			{
			 "IncidentLine2dLine" => 
			 [ map { $ls[$_], $L_cube->[$_]->[0],   } (0..11) ]  
			},
			{
			 "type_of_constraints" => "Sphere",
			 "verbose" => 1
			});

my $result = Estimate("ProjectiveCamera",
		      {
		       "IncidentLine2dLine" => 
		       [ map { $ls[$_], $L_cube->[$_]->[0],   } (0..11) ],
		       "IncidentPoint2dPoint" => 
		       [ map { $xs[$_], $X_cube->[$_]->[0],   } (0..7) ]  
		      },
		      {
		       "type_of_constraints" => "Sphere",
		       "verbose" => 1
		      });

my %results = ( "FromPoints        " => $result_x,
		"FromLines         " => $result_l,
		"FromPointsAndLines" => $result);

# here we store the total sum of all errors
my %total_sum = ();
foreach my $key (keys %results) {
  $total_sum{$key} = [0,0];
}
    
my (@xs_sum1,@xs_sum2,@xs_sum3);
foreach my $i (0..7) {
  my $xs_calc1 = $P->matrix x $X_cube->[$i]->[0]->vec;
  push @xs_sum1, sqrt(PDL::sum((SUGR::Basic::S($xs_calc1) x $xs[$i]->vec)**2));

  print "Reprojecting Error for Point $i (sqrt of squared sum):\n";
  print "x$i' x (P(Original) x X$i)     $xs_sum1[-1]\n";
  foreach my $key (keys %results) {
    my $xs_calc2 = $results{$key}->{initial_ProjectiveCamera}->matrix x $X_cube->[$i]->[0]->vec;
    my $xs_calc3 = $results{$key}->{ProjectiveCamera}->matrix x $X_cube->[$i]->[0]->vec;
    
    
    push @xs_sum2, sqrt(PDL::sum((SUGR::Basic::S($xs_calc2) x $xs[$i]->vec)**2));
    push @xs_sum3, sqrt(PDL::sum((SUGR::Basic::S($xs_calc3) x $xs[$i]->vec)**2));
    
    $total_sum{$key}->[0] += $xs_sum2[-1];
    $total_sum{$key}->[1] += $xs_sum3[-1];

    print "x$i' x (P(Approximated $key) x X$i) $xs_sum2[-1]\n";
    print "x$i' x (P(Estimated    $key) x X$i) $xs_sum3[-1]\n";
  }
  print "\n";
}			


my (@ls_sum1,@ls_sum2,@ls_sum3);
foreach my $i (0..11) {
  my $ls_calc1 = new SUGR::Line2d($P,$L_cube->[$i]->[0]);
  print "Reprojecting Error for Line $i (sqrt of squared sum):\n";
  push @ls_sum1, sqrt(PDL::sum((SUGR::Basic::S($ls_calc1) x $ls[$i]->vec)**2));
  print "l$i' x (Q(Original) x L$i)     $ls_sum1[-1]\n";
  
  foreach my $key (keys %results) {
    my $ls_calc2 = new SUGR::Line2d($results{$key}->{initial_ProjectiveCamera},
			       $L_cube->[$i]->[0]);
    my $ls_calc3 = new SUGR::Line2d($results{$key}->{ProjectiveCamera},
			       $L_cube->[$i]->[0]);
    
    push @ls_sum2, sqrt(PDL::sum((SUGR::Basic::S($ls_calc3) x $ls[$i]->vec)**2));
    push @ls_sum3, sqrt(PDL::sum((SUGR::Basic::S($ls_calc2) x $ls[$i]->vec)**2));
    
    $total_sum{$key}->[0] += $ls_sum2[-1];
    $total_sum{$key}->[1] += $ls_sum3[-1];

    print "l$i' x (Q(Approximated-$key) x L$i) $ls_sum2[-1]\n";
    print "l$i' x (Q(Estimated-$key) x L$i)    $ls_sum3[-1]\n";
  }
  print "\n";
}

foreach my $key (keys %total_sum) {
  print "$key: Total sum approximated solution: $total_sum{$key}->[0]\n";
  print "$key: Total sum estimated solution   : $total_sum{$key}->[1]\n";
}

#####################################################################
#
# TODO: 

# my $result = Estimate("Point",
# 		      {
# 		       "IncidentPoint2dProjection" => 
# 		       [ # ... 
# 		       ]
# 		       "IncidentPlane" => 
# 		       [ # ...
# 		       ]  
# 		      },
# 		      {
# 		       "type_of_constraints" => "Sphere",
# 		       "verbose" => 1
# 		      });

__END__
