=head1 NAME

SUGR::Homography -- class for  3D->3D Homography

=head1 NOT DOCUMENTED YET

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Element(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

package SUGR::Homography;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::TransformationNew;
use PDL::Matrix;

$VERSION = SUGR->VERSION;
@ISA = qw(Exporter SUGR::TransformationNew);

@EXPORT = qw(sugr_Homography); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use SUGR::Basic qw/:ALL/;

# this is a constructor that can be called like it is
sub sugr_Homography { 
  SUGR::Homography->new(@_);
}

# currently only translation and rotation are supported
sub ConstructFromHASH {
  my ($arghash) = @_;

  my %args = %$arghash;

  my $R = SUGR::Basic::Identity(4);
  my $T = SUGR::Basic::Identity(4);
  if (defined $args{rotation}) {
    my ($omega,$phi,$kappa) = @{$args{rotation}};
    my $mR = mpdl([
		 [cos($omega),-sin($omega), 0 ],
		 [sin($omega), cos($omega), 0 ],
		 [           0,          0, 1 ]
		]) x
		  mpdl([
			[ cos($phi) ,0, sin($phi)  ],
			[ 0         ,1,         0  ],
			[-sin($phi), 0, cos($phi) ]
		       ]) x
			 mpdl([
			       [1,          0,          0 ],
			       [0,cos($omega),-sin($omega)],
			       [0,sin($omega), cos($omega)]
			      ]);
    ($_ = $R->slice("0:2,0:2")) .= $mR;
  }
  if (defined $args{translation}) {
    my ($tx,$ty,$tz) = @{$args{translation}} if defined $args{translation};
    ($_ = $T->slice("0:2,3")) .= vpdl([$tx,$ty,$tz]);
  }

  return new SUGR::Homography($R x $T);
}

# the following two constructor helper functions may be rewritten less
# redundantly, since they actually only check dimensions, but what the
# heck!

# constructor with no covariance matrix
sub ConstructFromMatrix { 
  ConstructFromMatrixMatrix(@_);
}

# this is the actual constructor !
sub ConstructFromMatrixMatrix {
  my ($matrix,$cov) = @_;
  croak("no transformation matrix given") if !defined($matrix);

  # check dimensions
  my ($r,$c) = $matrix->mdims;
  if ($r*$c==16) {
    $matrix = $matrix->reshape(4,4);
  }
  else {
    croak("first argument should be a (4,4) matrix, is $matrix");
  }

  my $self = defined($cov) ?
    SUGR::TransformationNew::ConstructFromMatrixMatrix($matrix,$cov) :
	SUGR::TransformationNew::ConstructFromMatrix($matrix);

  # this is so that I don't have to compute the projection matrix for
  # lines over and over again.
  ($self->{vec_line},$self->{cov_line}) = PointToLineTransform($self->{vec}->reshape(4,4),$self->{cov});

  return $self;
}

sub type {
  "Homography";
}

sub result_type {
  my ($self,$type) = @_;
  $type ||= "SUGR::Point"; # default value

  return $type; # input type is output type for an homography
}

my $Q = mpdl [ [0,0,0,1,0,0],
	       [0,0,0,0,1,0],
	       [0,0,0,0,0,1],
	       [1,0,0,0,0,0],
	       [0,1,0,0,0,0],
	       [0,0,1,0,0,0] ];

sub PointToLineTransform {
  my ($M,$Cov) = @_ ; # M is the point transform 
  my $N = mzeroes(6,6);

  my @ind1 = (1,2,0,3,3,3);
  my @ind2 = (2,0,1,0,1,2);
  for (my $i=0;$i<6;$i++) {
    my $L = Pidual(~$M->slice($ind1[$i].",")) x ~$M->slice($ind2[$i].",");
    ($_ = $N->slice("$i,")) .= ~$L->rotate(3); # store the dual
  }
  # TODO: need to figure out covariance matrix for transformation!
  return ($Q x $N,undef);
#  return ~SL x  kroneckerproduct($M,$M) x $SL;
}

my @euclid_indices = [0..14];
my @homo_indices = [15];

sub euclid_scale {
  my ($self) = shift;
  return sqrt(PDL::inner($self->vec->dice(@euclid_indices)->clump(-1),
			 $self->vec->dice(@euclid_indices)->clump(-1)));
}

# needed for euclidize
sub Jacobian_norm_euclid {
  Jacobia_norm_homogeneous(@_);
}

sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian
  my $M = SUGR::Element::Identity(12);
  ($_ = $M->diagonal(0,1)->dice(@euclid_indices)) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_euclidean_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->dice(@euclid_indices)**2));
}

sub norm_of_homogeneous_part {
  my $self = shift;
  PDL::sum(abs($self->vec->dice(@homo_indices)));
}

sub Constraint_Sphere {
  my ($P,$squared_radius) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  $squared_radius ||=1;
  return vpdl([ 0.5*PDL::sum($P**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($P) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  ~$P;
}

return 1 if caller;

#################################################################
# Test code
use SUGR::Point;
use SUGR::Cuboid;
use SUGR::Point2d;
use SUGR::Line2d;
use SUGR::MyListUtil qw/max min/;
#use PDL::Graphics::PGPLOT;

my %Objects = generate_cuboid(no_points   => 1,
			      no_planes   => 1,
			      no_lines    => 1,
			      standard_deviation => [0.1,0.1,0.1],
			      random_seed => [4711,222712]);

#print $Objects{Ideal_Points}->[$_]->[0] foreach (0..7);

my $Pi = 2*1.57079632679489661922;
%Objects = transform_cuboid(translation => [0,0,0],
			    rotation    => [-0.25*$Pi,0.75*$Pi,0],
			    %Objects);

#print "Transformed:\n------------------\n";
#print $Objects{Ideal_Points}->[$_]->[0] foreach (0..7); die;

my $X_cube  = $Objects{Points};
my $L_cube  = $Objects{Lines};
my $A_cube  = $Objects{Planes};
my $X_cube_ideal  = $Objects{Ideal_Points};
my $L_cube_ideal  = $Objects{Ideal_Lines};
my $A_cube_ideal  = $Objects{Ideal_Planes};


my $H = sugr_Homography(msequence(4,4)+1,0.01*mones(16,16));

my @HX_cube;
foreach my $corner (0..$#{@$X_cube_ideal}) {
  foreach my $X_cubepoint (@{$X_cube_ideal->[$corner]})  {
    push @{$HX_cube[$corner]}, $H->transform($X_cubepoint);
  }
}

print $H;

foreach my $corner (0..$#HX_cube) {
  foreach my $i (0..$#{$HX_cube[$corner]})  {
    foreach my $X ($X_cube_ideal->[$corner]->[$i],
		   $HX_cube[$corner]->[$i]) {
      print $X->euclidize->vec->slice("0:2")->clump(-1)," ";
    }
  }
  print "\n";
}

print $HX_cube[$_]->[0], $X_cube->[$_]->[0],"\n" foreach (0..7);



my $Hparam = sugr_Homography({translation => [2,2,-3],
			      rotation    => [0.2,0.2,0],
			     }
			    );

print $Hparam;

__END__
