=head1 NAME

SUGR::InverseProjectiveCamera -- class for projective transformation from a 2D plane to a 3D plane

=head1 NOT DOCUMENTED YET

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Element(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut


package SUGR::InverseProjectiveCamera;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::TransformationNew;

$VERSION = SUGR->VERSION
@ISA = qw(Exporter SUGR::TransformationNew);

@EXPORT = qw(sugr_InverseProjectiveCamera); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use SUGR::Basic qw/:ALL/;

# this is a constructor that can be called like it is
sub sugr_InverseProjectiveCamera { 
  my @args = @_;
  SUGR::InverseProjectiveCamera->new(@args);
}

# the following two constructor helper functions may be rewritten less
# redundantly, since they actually only check dimensions, but what the
# heck!

# constructor with no covariance matrix
sub ConstructFromMatrix { 
  ConstructFromMatrixMatrix(@_);
}

# this is the actual constructor !
sub ConstructFromMatrixMatrix {
  my ($matrix,$cov) = @_;
  croak("no transformation matrix given") if !defined($matrix);

  # check dimensions
  my ($r,$c) = $matrix->mdims;
  if ($r*$c==12) {
    $matrix = $matrix->reshape(4,3);
  }
  else {
    croak("first argument should be a (4,3) matrix, is $matrix");
  }

  my $self = defined($cov) ?
    SUGR::TransformationNew::ConstructFromMatrixMatrix($matrix,$cov) :
	SUGR::TransformationNew::ConstructFromMatrix($matrix);

  # this is so that I don't have to compute the projection matrix for
  # lines over and over again.
#  ($self->{vec_line},$self->{cov_line}) = LineProjection($self->{vec},$self->{cov});

  return $self;
}

# this is the actual constructor !
sub ConstructFromProjectiveCameraPlane {
  my ($P,$A) = @_;

  my $P_inv = $A->Pi->transpose x  ~$P->matrix("line");

  # Jacobian with respect to the projection
  my $J1 = kroneckerproduct(Identity(3),$A->Pi->transpose); # x $P->vec("line");

  # Jacobian with respect to the plane 

  # (probably the same trick as with computing covariance in
  # ProjectiveCamera constructor with euclidean parameters: just take
  # the transpose of the actual computation
#  my $J2 = kroneckerproduct(Identity(3),~$A->Pi) x Jp($P); 

  # TODO: take uncertainty of plane $A into account
  my $Cov = $J1 x $P->cov("line") x ~$J1 
;
# + $J2 x $P->cov("line") x ~$J2;

  my $self = ConstructFromMatrixMatrix($P_inv,$Cov);

  $self->{plane} = $A->copy;

  return $self;
}

sub type {
  "InverseProjectiveCamera";
}

sub result_type {
  my ($self,$type) = @_;
  $type ||= "point2d"; # default value
  $type = lc($type);
  
  # TODO: no error checking yet!
  return ($type =~ /point2d/) ? "SUGR::Point" : "SUGR::Line";
}

my @euclid_indices = [0..2];
my @homo_indices = [3..11];

# needed for euclidize
sub Jacobian_norm_euclid {
  Jacobia_norm_homogeneous(@_);
}

sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian
  my $M = SUGR::Element::Identity(12);
  ($_ = $M->diagonal(0,1)->dice(@euclid_indices)) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_euclidean_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->dice(@euclid_indices)**2));
}

sub norm_of_homogeneous_part {
  my $self = shift;
  PDL::sum(abs($self->vec->dice(@homo_indices)));
}

sub Constraint_Sphere {
  my ($P,$squared_radius) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  $squared_radius ||=1;
  return vpdl([ 0.5*PDL::sum($P**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($P) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  ~$P;
}

return 1 if caller;

#################################################################
# Test code
use PDL::Matrix;
use SUGR::Point;
use SUGR::Cuboid;
use SUGR::Point2d;
use SUGR::Line2d;
use SUGR::ProjectiveCamera;
use SUGR::Relation;
#use PDL::Graphics::PGPLOT;
use SUGR::MyListUtil qw/max min/;

my %Objects = generate_cuboid(no_points   => 1,
			      no_planes   => 1,
			      no_lines    => 1,
			      standard_deviation => [0.01,0.01,0.01],
			      random_seed => [4711,222712]);

# %Objects = transform_cuboid(translation => [1,2,2],
# 			    rotation    => [0.5,0.7,1.3],
# 			    %Objects);


my $X_cube  = $Objects{Points};
my $L_cube  = $Objects{Lines};
my $A_cube  = $Objects{Planes};
my $X_cube_ideal  = $Objects{Ideal_Points};
my $L_cube_ideal  = $Objects{Ideal_Lines};
my $A_cube_ideal  = $Objects{Ideal_Planes};


my $X0 = sugr_Point([1.5,1.5,10,1],
		    0.01**2*mpdl([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,0]]));


my $R  = mpdl([[1,0,0],
	       [0,1,0],
	       [0,0,1]]);
my $P = sugr_ProjectiveCamera({ 
			       "X0"    => $X0,
			       "R"     => $R,
			       "c"     => 1,
			       "c_var" => 0.001,
			      });
my $P_inv = sugr_InverseProjectiveCamera($P,$A_cube_ideal->[0]->[0]);

print $P;
print $P_inv;


my @xs;
print "Projecting noisy cube points into the image...\n";
foreach my $X_cubepoint (map { $_->[0] } (@$X_cube[0..3])) {
  push @xs, sugr_Point2d($X_cubepoint,$P)->euclidize; 
  print $xs[-1]->vec->slice("0:1")->clump(-1),"\n";
}

my @X;
print "Reprojecting image points points into the first ideal plane...\n";
foreach my $i (0..3) {
  print "Test ",($i+1),"th reprojected 3D point with ideal point: ";
  push @X, sugr_Point($P_inv,$xs[$i]);
  my %result = Equal($X_cube_ideal->[$i]->[0],$X[-1]);
  print "$result{testvalue} => ";
  print $result{bool} ? "succeeded!\n" : "failed!\n";
}



__END__
