=head1 NAME

SUGR::Line -- class for uncertain lines in 3D

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Line->VERSION

=head1 SYNOPSIS

  use SUGR;

  $X = new SUGR::Line([$LC1,$LC2,$LC3,$LC4,$LC5,$LC6],

		      [[0.01,0,0,0]
		       [0,0.01,0,0]
		       [0,0,0.01,0],
		       [0,0,0   ,0]]);

=head1 DESCRIPTION

=head2 Parameter vector

The parameter vector is a homogeneous 6-vector. Since a 3D line
requires a minimum of 4 parameters, there is an additional constraint
on this vector besides the homogeneity; the so-called Plcker
constraint, stating that the dot product of the first 3-vector with
the second 3-vector is zero.

The first 3-vector may be interpreted as the direction of the
line. The second 3-vector is orthogonal to the plane that is defined
by the line itself and the origin. If we normalize the first 3-vector,
i.e. i.e. ($LC1**2 + $LC2**2 + $LC3**2) == 1, the length of the
orthogonal vector is equal to the smallest distance of the line to the
origin.

=head2 Constructor new()

The C<new> constructor is quite flexible and allows several different
creation methods. Note that the arguments can be swapped.

=over 4

=item - 

Construction by specifying parameter vector and covariance matrix. 

The parameter vector can be specified either as a PDL::Matrix object
or as a reference to a Perl array (the covariance matrix then is a
reference to Perl array of arrays, see perldoc perldsc for more
details).

You may omit the covariance matrix, putting up with the lack of error
propagation for this element.

=item -

Construct from two planes $A1, $A2

  # ref($A1) eq "SUGR::Plane"
  # ref($A2) eq "SUGR::Plane"
  $L = new SUGR::Line($A1,$A2,$A3);

=item -

Construct from two points $X1, $X2

  # ref($X1) eq "SUGR::Point"
  # ref($X2) eq "SUGR::Point"
  $L = new SUGR::Line($X1,$X2);

=item -

Construct from a 2D point $x and a pin-hole camera $P 
  
  # ref($x) eq "SUGR::Point2d"
  # ref($P) eq "SUGR::ProjectiveCamera"
  $L = new SUGR::Line($x,$P);

=item - 

Estimation of a 3D line from sets of observations. The observations
may be incident 2D lines or equal 2D points. The observations will be
specified in a hash. See also <SUGR::Estimation(3)|SUGR::Estimation>
for more details on the input and the technique.

  $L = new SUGR::Line({ "EqualLine"       => [ $L1, $L2 ],
			"IncidentLine"    => [ $L3, $L4, $L5, $L6 ],
			"ParallelLine"    => [ $L7, $L8, $L9],
			"OrthogonalLine"  => [ $L10 ]

			"IncidentPlane"   => [ $A1, $A2, $A3 ],
			"OrthogonalPlane" => [ $A4, $A5, $A6],
			"ParallelPlane"   => [ $A7 ],
			
			"IncidentPoint"   => [ $X1, $X2, $X3 ],
		       });

=cut

package SUGR::Line;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::Element;
use SUGR::Estimation ();
$VERSION = SUGR->VERSION;
@ISA = qw(Exporter SUGR::Element);

@EXPORT = qw(sugr_Line); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL::Matrix;
use SUGR::Basic qw/Gammadual/;
################################################################################
#
# global variables;
#

# this helps with dualizing lines
my $Q = mpdl [ [0,0,0,1,0,0], 
	       [0,0,0,0,1,0],
	       [0,0,0,0,0,1],
	       [1,0,0,0,0,0],
	       [0,1,0,0,0,0],
	       [0,0,1,0,0,0] ];

#
# Constructors
#

# this is a constructor that can be called like it is
sub sugr_Line { SUGR::Line->new(@_) }

sub ConstructFromHASH {
  my ($arghash) = @_;

  print "Here we go!\n";
  my $result = SUGR::Estimation::Estimate("Line",$arghash);

  return $result->{Line};
}

# returns sugr_Line-object from two sugr_Point-objects
sub ConstructFromPointPoint {
  my ($X,$Y) = @_;
  my $L   = $X->Pi x $Y->vec;
  my $Cov = (defined($X->cov) && defined($Y->cov)) ?
    ($X->Pi x $Y->cov x ~$X->Pi +
     $Y->Pi x $X->cov x ~$Y->Pi) :
       undef;
  return sugr_Line($L,$Cov);
}

# returns sugr_Line-object from two sugr_Plane-objects
sub ConstructFromPlanePlane {
  my ($P,$Q) = @_;
  my $L =   $P->Pidual x $Q->vec ;
  my $Cov = (defined($P->cov) && defined($Q->cov)) ?
    ($P->Pidual x $Q->cov x ~$P->Pidual +
     $Q->Pidual x $P->cov x ~$Q->Pidual ) : 
       undef;
  return sugr_Line($L,$Cov);	
}

# compute 3d line from image point and projection matrix
sub ConstructFromPoint2dMatrix {
  my ($p,$Ps) = @_;
  
  my $L =   ~$Ps x $p->vec;
  my $Cov = defined($p->cov) ?
    (~$Ps x $p->cov x $Ps) :
      undef;
  
  return sugr_Line($L,$Cov);
}
sub ConstructFromMatrixPoint2d { ConstructFromPoint2dMatrix(@_[1,0]) }

# compute 3d line from image point and projection matrix
sub ConstructFromPoint2dProjection {
  my ($x,$P) = @_;

  my $L = ~$P->matrix("line") x $x->vec;

  my $Cov = defined($x->cov) ?
    (~$P->matrix("line") x $x->cov x $P->matrix("line")) :
      undef;
  
  if (defined($P->cov("line"))) {
    my $J = mzeroes(6,18);
    ($_ = $J->slice("0:5,0:5")->diagonal(0,1)) .= $x->vec->at(0);
    ($_ = $J->slice("0:5,6:11")->diagonal(0,1)) .= $x->vec->at(1);
    ($_ = $J->slice("0:5,12:17")->diagonal(0,1)) .= $x->vec->at(2);

#    print "test projecting line:",$L->clump(2)," - ",clump($J x $P->vec("line"),2),"\n";

    $Cov += ( $J x $P->cov("line") x ~$J); # this is also works if
                                           # $Cov was not defined before!
  } 
  return sugr_Line($L,$Cov);
}

sub ConstructFromProjectionPoint2d { ConstructFromPoint2dProjection(@_[1,0]) }

sub ConstructFromPointProjectiveCamera {
  ConstructFromPointProjection(@_)
}
sub ConstructFromProjectiveCameraPoint {
  ConstructFromPointProjection(@_[1,0])
}

# methods

# matrix representation of the entity
sub matrix {
  Gammadual($_[0]);
}

sub dual {
  my ($self) = @_;
  my $L = $self->vec->rotate(3);
  my $Cov = defined($self->cov) ?
    ($Q x $self->cov x ~$Q) : # this may be optimized since this is only reordering
      undef;
  return sugr_Line($L,$Cov);
}

sub type {
  "Line";
}

=pod

=back

=head2 Class and object methods

See the L<SUGR::Element(3)|SUGR::Element> manual page for inherited
methods.

=cut

sub vec1 {
  my ($self) = @_;
  my $L = $self->vec;
  return vpdl [ $L->at(0), $L->at(1), $L->at(2) ];
}

sub vec2 {
  my ($self) = @_;
  my $L = $self->vec;
  return vpdl [ $L->at(3), $L->at(4), $L->at(5),];
}

# distance to origin
sub distorigin {
  my ($self) = @_;
  my $pdlvec1 = $self->vec1->slice(",(0)");
  my $pdlvec2 = $self->vec2->slice(",(0)");
  my $denominator = PDL::inner($pdlvec1,$pdlvec1);
  $denominator != 0 ? (sqrt(PDL::inner($pdlvec2,$pdlvec2)/$denominator)) : ("infinity");
}

# point closest to origin
sub point_closest_to_origin {
  my ($self) = @_;
  my $L  = $self->vec1;
  my $L0 = $self->vec2;
  my $X = ~PDL::append( -PDL::Matrix::slice(SUGR::Element::S($L0) x $L,",(0)"),
			PDL::inner($L->slice(",(0)"),$L->slice(",(0)")));

  # TODO: computation of covariance matrix !
  return new SUGR::Point($X);
}

sub distance {
  my ($self,$X) = @_;
  
  my $Lh = $self->vec->slice("0:2");
  my $L0 = $self->vec->slice("3:5");
  my $X0 = $X->vec->slice("0:2");
  my $Xh = $X->vec->at(3);
  
  return PDL::sum(sqrt((SUGR::Element::S($Lh) x $X0 + $Xh * $L0)**2)) /
    PDL::sum(sqrt(($Xh * $Lh)**2));
}

# point closest to a given point
sub point_closest_to_point {
  my ($self,$X) = @_;

  $X = $X->copy->euclidize;
  my $T   = SUGR::Transformation::Translation(-$X->vec->slice("0:2"));
  my $selfT  = $self->transform($T);
  my $LT     = $selfT->vec1;
  my $LT0    = $selfT->vec2;
  
  my $D = -(SUGR::Element::S($LT0) x $LT) / PDL::sum(~$LT x $LT);
  my $XC = $X->vec;
  ($_ = $XC->slice("0:2")) += $D;
  
  return new SUGR::Point($XC);
}

sub euclid_scale {
  my ($self) = shift;
  return sqrt(PDL::inner($self->vec->slice("0:2,(0)"),$self->vec->slice("0:2,(0)")));
}

# needed for euclidize
sub Jacobian_norm_euclid {
  my ($self) = @_;
  my @L = $self->vec->list;
  my $L1L2L3 = $L[0]**2+$L[1]**2+$L[2]**2;
  die $self." is not a euclidean element." if $L1L2L3==0;
  my $L1L2L3_32 = (($L1L2L3)**(3/2));

  mpdl([[-$L[0]**2/($L1L2L3_32)+1/(sqrt($L1L2L3)),
	 -$L[0]*$L[1]/($L1L2L3_32), 
	 -$L[0]*$L[2]/($L1L2L3_32), 
	 0, 
	 0, 
	 0], 
	[-$L[0]*$L[1]/($L1L2L3_32),
	 -$L[1]**2/($L1L2L3_32)+1/(sqrt($L1L2L3)), 
	 -$L[1]*$L[2]/($L1L2L3_32), 
	 0, 
	 0,
	 0], 
	[-$L[0]*$L[2]/($L1L2L3_32), 
	 -$L[1]*$L[2]/($L1L2L3_32), 
	 -$L[2]**2/($L1L2L3_32)+1/(sqrt($L1L2L3)), 
	 0, 
	 0, 
	 0], 
	[-$L[3]*$L[0]/($L1L2L3_32), 
	 -$L[3]*$L[1]/($L1L2L3_32), 
	 -$L[3]*$L[2]/($L1L2L3_32), 
	 1/(sqrt($L1L2L3)), 
	 0,
	 0],
	[-$L[4]*$L[0]/($L1L2L3_32), 
	 -$L[4]*$L[1]/($L1L2L3_32), 
	 -$L[4]*$L[2]/($L1L2L3_32), 
	 0, 
	 1/(sqrt($L1L2L3)), 
	 0], 
	[-$L[5]*$L[0]/($L1L2L3_32), 
	 -$L[5]*$L[1]/($L1L2L3_32), 
	 -$L[5]*$L[2]/($L1L2L3_32), 
	 0,
	 0,
	 1/(sqrt($L1L2L3))]]);
}

sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian for euclidean scaling
  my $M = SUGR::Element::Identity(6);
  ($_ = $M->diagonal(0,1)->PDL::slice("3:5")) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_homogeneous_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->slice("0:2")**2));
}

sub norm_of_euclidean_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->slice("3:5")**2));
}

sub Constraint_Euclidean {
  my ($L) = @_;
  $L = (ref($L) =~ /SUGR/) ? $L->vec : $L;
  return vpdl [ 0.5*PDL::sum($L->slice("0:2")**2) - 0.5 ,
		0.5*PDL::sum($L * $L->rotate(3))
	      ];
}
sub Jacobian_Constraint_Euclidean {
  my ($L) = @_;
  $L = (ref($L) =~ /SUGR/) ? $L->vec : $L;
  my $HJ = mzeroes(6,2);
  ($_ = $HJ->slice("0:2,0")) .= $L->slice("0:2");
  ($_ = $HJ->slice("0:5,1")) .= $L->rotate(3);
  return(~$HJ);
}

sub Constraint_Sphere {
  my ($L,$squared_radius) = @_;
  $L = (ref($L) =~ /SUGR/) ? $L->vec : $L;
  $squared_radius ||=1;
  return vpdl [ 0.5*PDL::sum($L**2) - 0.5*$squared_radius ,
		0.5*PDL::sum($L * $L->rotate(3))
	      ];
}
sub Jacobian_Constraint_Sphere {
  my ($L) = @_;
  $L = (ref($L) =~ /SUGR/) ? $L->vec : $L;
  my $HJ = mzeroes(6,2);
  ($_ = $HJ->slice("0:5,0")) .= $L->slice("0:5");
  ($_ = $HJ->slice("0:5,1")) .= $L->rotate(3);
  return(~$HJ);
}


=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Estimation(3), SUGR::Element(3), PDL::Matrix(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut


return 1 if caller;

################################################################################
#
# Test code
  
my $L = sugr_Line(vsequence(6),msequence(6,6));
  
print $L;
print $L->matrix;
print $L->Jacobian_Constraint_Euclidean();
print $L->Jacobian_Constraint_Sphere();
$L->scale_euclidean_part(0.1);
print $L->Jacobian_Constraint_Euclidean();


__END__

# this is the old stringify...
sub stringify {
  my ($self) = shift;
  my $pdlvec1 = $self->vec1->slice(",(0)");
  my $pdlvec2 = $self->vec2->slice(",(0)");
  my $pdlvec1scalar = sqrt(PDL::inner($pdlvec1,$pdlvec1));
  my $pdlvec2scalar = sqrt(PDL::inner($pdlvec2,$pdlvec2));
  my $s = "\nLine Direction: ";
  $pdlvec1scalar != 0 ? 
    ($s .= ($pdlvec1/$pdlvec1scalar)." (scaled with ".(1/$pdlvec1scalar).")") : ($s .= "undef");
    $s .= "\nLine Normal   : ";
  $pdlvec2scalar != 0 ? 
    ($s .= ($pdlvec2/$pdlvec2scalar)." (scaled with ".(1/$pdlvec2scalar).")") : ($s .= "undef");
  $s .= "\nDistance to origin: ".$self->distorigin."\n";
  $s .= (1/$pdlvec1scalar)**2 * $self->cov if defined($self->cov);
  return $s;
}
