=head1 NAME

SUGR::Line2d -- class for uncertain 2D lines

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Line2d->VERSION

=head1 SYNOPSIS

  use SUGR;

  $x = new SUGR::Line2d([$cos_a,$sin_a,-$d],[[0.01,0,0]
					     [0,0.0 ,0]
					     [0,0,0.01]]);

=head1 DESCRIPTION

=head2 Parameter vector

The parameter vector is a homogeneous 3-vector. The first two coordinates
refer to the direction of the normal of the line. If the first two
coordinates are normalized , ($cos_a**2 + $sin_a**2) == 1, the last
coordinate refers to the distance of the line to the origin of the
coordinate system.

=head2 Constructor new()

The C<new> constructor is quite flexible and allows several different
creation methods. Note that the arguments may be swapped.

=over 4

=item - 

Construction by specifying parameter vector and covariance matrix. 

The parameter vector can be specified either as a PDL::Matrix object
or as a reference to a Perl array (the covariance matrix then is a
reference to Perl array of arrays, see perldoc perldsc for more
details).

You may omit the covariance matrix, putting up with the lack of error
propagation for this element.

=item -

Construct from two 2D points

  # ref($x1) eq "SUGR::Point2d"
  # ref($x2) eq "SUGR::Point2d"
  $l = new SUGR::Line2d($x1,$x2);

=item -

Construct from 3D line and pin-hole camera

  # ref($L) eq "SUGR::Line"
  # ref($P) eq "SUGR::ProjectiveCamera"
  $l = new SUGR::Line2d($L,$P);

=item - 

Estimation of a 2D line from sets of observations. The observations may
be equal, orthogonal or parallel 2D lines or incident 2D points. The
observations will be specified in a hash. See also
<SUGR::Estimation(3)|SUGR::Estimation> for more details on the input
and the technique.

  $x = new SUGR::Line2d({ "IncidentPoint2d" => [ $x1, $x2, $x3 ],
			  "EqualLine2d" => [ $l1, $l2 ],
			  "ParallelLine2d" => [ $l3, $l4, $l5 ],
			  "OrthogonalLine2d" => [ $l6 ],
			});

=cut

package SUGR::Line2d;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::Element;
use SUGR::Estimation ();
$VERSION = SUGR->VERSION;
@ISA = qw(Exporter SUGR::Element);

@EXPORT = qw(sugr_Line2d); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL ();
use PDL::Matrix;
use SUGR::Basic qw/:ALL/;

################################################################################
#
# global variables;
#
my $CLine = mpdl [ [0,0,0,1,0,0],
		   [0,0,0,0,1,0],
		   [0,0,0,0,0,1],
		   [1,0,0,0,0,0],
		   [0,1,0,0,0,0],
		   [0,0,1,0,0,0] ];

#
# Constructors
#

# this is a constructor that can be called like it is
sub sugr_Line2d { SUGR::Line2d->new(@_) }

sub ConstructFromHASH {
  my ($arghash) = @_;
  
  my $result = SUGR::Estimation::Estimate("Line2d",$arghash);

  return $result->{Line2d};
}

# returns sugr_Line2d-object from two sugr_Point-objects
sub ConstructFromPoint2dPoint2d {
  my ($x,$y) = @_;
  my $l   = S($x) x $y->vec;
  my $cov = (defined($x->cov) && defined($y->cov)) ?
    ( S($x) x $y->cov x ~S($x) +
      S($y) x $x->cov x ~S($y)  ) :
	undef;
  return sugr_Line2d($l,$cov);
}

# project 3D line into image
sub ConstructFromLineMatrix {
    my ($L,$Ps) = @_;
    my $Ldual = $L->dual;

    my $l = $Ps x $Ldual->vec;
    my $cov =  defined($L->cov) ? 
      ($Ps x $Ldual->cov x ~$Ps) :
	undef;

    return sugr_Line2d($l,$cov);
}
sub ConstructFromMatrixLine { ConstructFromLineMatrix(@_[1,0]) }


sub ConstructFromLineProjection {
  my ($L,$P) = @_;

  my $Ld = $L->dual;

  my $l = $P->matrix("line") x $Ld->vec;
  my $Cov = defined($Ld->cov) ?
    ($P->matrix("line") x $Ld->cov x ~$P->matrix("line")) :
      undef;

  if (defined($P->cov("line"))) {
    my $J = mzeroes(3,18);
    ($_ = $J->slice("0,0:5"))   .= ~$Ld->vec;
    ($_ = $J->slice("1,6:11"))  .= ~$Ld->vec;
    ($_ = $J->slice("2,12:17")) .= ~$Ld->vec;

#    print "test projection of line :",$l->clump(2)," - ",clump($J x $P->vec("line"),2),"\n";

    $Cov += ( $J x $P->cov("line") x ~$J); # this is also works if
                                           # $Cov was not defined before!
  } 
  
  return sugr_Line2d($l,$Cov);
}
sub ConstructFromProjectionLine { ConstructFromLineProjection(@_[1,0]) }

sub ConstructFromLineProjectiveCamera {
  ConstructFromLineProjection(@_)
}
sub ConstructFromProjectiveCameraLine {
  ConstructFromLineProjection(@_[1,0])
}

# methods

sub _euclidean_nullspace {
  my $vector = shift;
  return $vector->copy->set(2,0);
}

=pod

=back

=head2 Class and object methods

See the L<SUGR::Element(3)|SUGR::Element> manual page for inherited
methods.

=over 4

=cut

sub matrix {
  S($_[0]);
}

sub dual {
  my ($self) = @_;
  my $p = $self->copy;
  bless $p, "SUGR::Point2d";
}

sub type {
  "Line2d";
}


sub euclid_scale {
  my ($self) = shift;
  return sqrt($self->vec->at(0)**2+$self->vec->at(1)**2);
}

sub Jacobian_norm_euclid {
  my ($self) = @_;
  my @l = $self->vec->list;
  my $l1l2 = $l[0]**2+$l[1]**2;
  die $self." is not a euclidean element." if $l1l2==0;
  my $l1l2_32 = (($l1l2)**(3/2));
  mpdl([[-$l[0]**2/$l1l2_32+1/(sqrt($l1l2)), 
	 -$l[0]*$l[1]/$l1l2_32, 
	 0], 
	[-$l[0]*$l[1]/$l1l2_32, 
	 -$l[1]**2/$l1l2_32+1/(sqrt($l1l2)), 
	 0], 
	[-$l[2]*$l[0]/$l1l2_32, 
	 -$l[2]*$l[1]/$l1l2_32, 
	 1/(sqrt($l1l2))]]);
}

sub Jacobian_norm_point {
  my ($self,$index) = @_;
  my @l = $self->vec->list;
  if ($index==0) {
    return mpdl([[0      , 0      ,                 0 ],
		 [-$l[1]/($l[0]**2), 1/$l[0],       0 ], 
		 [-$l[2]/($l[0]**2), 0      , , 1/$l[0]]
		]);
  }
  elsif ($index==1){
    return mpdl([[1/$l[1], -$l[0]/($l[1]**2),      0 ], 
		 [0      , 0      ,                0 ],
		 [0      , -$l[2]/($l[1]**2), 1/$l[1]] 
		]);
  }
  else { # $index == 2
    ($index == 2) or die;  # assert
    return mpdl([[1/$l[2], 0      , -$l[0]/($l[2]**2)], 
		 [0      , 1/$l[2], -$l[1]/($l[2]**2)], 
		 [0      , 0      ,                0 ]]);
  }
}


sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian for euclidean scaling
  my $M = SUGR::Element::Identity(3);
  ($_ = $M->diagonal(0,1)->PDL::slice("2")) .= $scalar;


  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_homogeneous_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->slice("0:1")**2));
}

sub norm_of_euclidean_part {
  my $self = shift;
  abs($self->vec->at(2));
}


sub Constraint_Euclidean {
  my ($l) = @_;
  my $v = (ref($l) =~ /SUGR/) ? $l->vec : $l;
  return vpdl([ 0.5*PDL::sum($v->slice("0:1")**2)  - 0.5 ]);
}

sub Jacobian_Constraint_Euclidean {
  my ($l) = @_;
  my $v = (ref($l) =~ /SUGR/) ? $l->vec : $l;
  my $J = vzeroes(3);
  ($_ = $J->slice("0:1")) .= $v->slice("0:1");
  return ~$J;
}

sub Constraint_Sphere {
  my ($l,$squared_radius) = @_;
  my $v = (ref($l) =~ /SUGR/) ? $l->vec : $l;
  $squared_radius ||=1;
  return vpdl([ 0.5*PDL::sum($v**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($l) = @_;
  (ref($l) =~ /SUGR/) ? ~$l->vec : ~$l;
}

=pod

=item euclidean_representation()

returns a 5-tupel consisting of:

(i) the center of gravity of the line according to the covariance
matrix and the angle of the line. These three parameters define a
translation and rotation such that the origin of the new coordinate
system is the center of gravity and the line is on the first
coordinate axis.

(ii) the standard deviations of the angle and the distance. Angle and
distance are independent from each other after applying the above
translation and rotation.

=cut

sub euclidean_representation {
  my $l  = $_[0]->euclidize;

  my $x0   =  -$l->cov->at(0,2)/$l->cov->at(2,2);
  my $y0   =  -$l->cov->at(1,2)/$l->cov->at(2,2);
  my $phi  =   atan2($l->vec->at(1),$l->vec->at(0));
  my $T    =   mpdl [[1,0,$x0],[0,1,$y0],[0,0,1]];
  my $R    =   mpdl [[$l->vec->at(0),$l->vec->at(1),0],[-$l->vec->at(1),$l->vec->at(0),0],[0,0,1]];
  my $Cov  =   $R x $T x $l->cov x ~$T x ~$R;
  my $sphi =   sqrt($Cov->at(1,1));
  my $sd   =   sqrt($Cov->at(2,2));
  
  return ($x0,$y0,$phi,$sphi,$sd,$Cov);
}

sub euclidean_representation2 {
  my $l  = $_[0]->euclidize;

  my $R     =  	mpdl [[$l->vec->at(0),$l->vec->at(1),0],[-$l->vec->at(1),$l->vec->at(0),0],[0,0,1]];
  my $Cov1  =  	$R x  $l->cov x  ~$R;
  print "Cov1 = $Cov1";
  my $phi   =  	atan2($l->vec->at(1),$l->vec->at(0));
  my $u0    =  -$l->cov->at(1,2)/$l->cov->at(1,1);
  my $x     =   ~$R x vpdl([$u0,0,1]);
  my $x0    =   $l->vec->at(2)*$l->vec->at(0) - $u0*$l->vec->at(1);
  my $y0    =   $l->vec->at(2)*$l->vec->at(1) + $u0*$l->vec->at(0);
  my $T     =  	mpdl [[1,0,0],[0,1,0],[0,$u0,1]];
  my $Cov2  =  	$T x  $Cov1 x  ~$T;
  
  my $sphi =  sqrt($Cov2->at(1,1));
  my $sd   =  sqrt($Cov2->at(2,2));
  
  return ($x0,$y0,$phi,$sphi,$sd,$u0,$x,$Cov1,$Cov2);
}

return 1 if caller;

=pod

=back

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Estimation(3), SUGR::Element(3), PDL::Matrix(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

################################################################################
# test code

my $l = new SUGR::Line2d([1,2,3],mones(3,3));

print $l;
print $l->matrix;
print $l->Jacobian_Constraint_Euclidean();
print $l->Jacobian_Constraint_Sphere();
$l->scale_euclidean_part(0.1);
print $l->Jacobian_Constraint_Euclidean();

use SUGR::Point2d;

# points and lines
my $x1 = sugr_Point2d(vpdl([-1 ,0,1]),mones(3,3));
my $x2 = sugr_Point2d(vpdl([ 1 ,0,1]),mones(3,3));
$l = sugr_Line2d($x1,$x2);

print $l;
