=head1 NAME

SUGR::Plane -- class for uncertain planes in 3D

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Plane->VERSION

=head1 SYNOPSIS

  use SUGR;

  $x = new SUGR::Plane([$nx,$ny,$nz,-$d],[[0.01,0,0,0]
					  [0,0.01,0,0]
					  [0,0,0.0 ,0],
					  [0,0,0,0.01]]);;

=head1 DESCRIPTION

=head2 Parameter vector

The parameter vector is a homogeneous 4-vector. The first three
coordinates refer to the direction of the normal of the plane. If the
first three coordinates are normalized , ($nx**2 + $ny**2 + $nz**2) ==
1, the last coordinate refers to the distance of the plane to the
origin of the coordinate system.

=head2 Constructor new()

The C<new> constructor is quite flexible and allows several different
creation methods. Note that the arguments may be swapped.

=over 4

=item -

Construction by specifying parameter vector and covariance matrix. 

The parameter vector can be specified either as a PDL::Matrix object
or as a reference to a Perl array (the covariance matrix then is a
reference to Perl array of arrays, see perldoc perldsc for more
details).

You may omit the covariance matrix, putting up with the lack of error
propagation for this element.

=item -

Construct from three 3D points $X1, $X2, $X3

  # ref($X1) eq "SUGR::Point"
  # ref($X2) eq "SUGR::Point"
  # ref($X3) eq "SUGR::Point"
  $A = new SUGR::Plane($X1,$X2,$X3);

=item -

Construct from 3D line $L and 3D point $X

  # ref($L) eq "SUGR::Line"
  # ref($X) eq "SUGR::Point"
  $A = new SUGR::Plane($L,$X);

=item -

Construct from a 2D line $l and a pin-hole camera $P 
  
  # ref($l) eq "SUGR::Line2d"
  # ref($P) eq "SUGR::ProjectiveCamera"
  $A = new SUGR::Line($l,$P);

=item - 

Estimation of a 3D plane from sets of observations. The observations
may be equal, orthogonal or parallel planes, incident 3D points or
incident, parallell or orthogonal 3D lines or any combination of
those. The observations will be specified in a hash. See also
<SUGR::Estimation(3)|SUGR::Estimation> for more details on the input
and the technique.

Example:

  $A = new SUGR::Plane({ "EqualPlane"      => [$A1, $A2],
			 "OrthogonalPlane" => [$A3, $A4, $A5],
			 "ParallelPlane"   => [$A6, $A7, $A8],
			 
			 "IncidentPoint"   => [$X1, $X2, $X3],
			 
			 "IncidentLine"    => [$L1, $L2],
			 
			 "ParallelLine"    => [$L3, $L4, $L5],
			 
			 "OrthogonalLine"  => [$L6, $L7, $L8]
			});


=cut

package SUGR::Plane;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::Element;
use SUGR::Estimation ();
$VERSION = 0.01;
@ISA = qw(Exporter SUGR::Element);

@EXPORT = qw(sugr_Plane); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL::Matrix;
use SUGR::Basic qw/Pidual/;

# this is a constructor that can be called like it is
sub sugr_Plane { SUGR::Plane->new(@_) }

sub ConstructFromHASH {
  my ($arghash) = @_;
  
  my $result = SUGR::Estimation::Estimate("Plane",$arghash);

  return $result->{Plane};
}

# returns sugr_Plane-object from sugr_Point $X and sugr_Line $L
sub ConstructFromPointLine {
  my ($X,$L) = @_;
  my $P   = ~$X->Pidual x $L->vec;
  my $Cov = (defined($X->cov) && defined($L->cov)) ?
    (~$X->Pidual x $L->cov x  $X->Pidual +
     $L->Gamma  x $X->cov x ~$L->Gamma) :
       undef;
  return sugr_Plane($P,$Cov);
}
sub ConstructFromLinePoint { ConstructFromPointLine(@_[1,0]) }

sub ConstructFromPointPointPoint {
  my ($X,$Y,$Z) = @_;
  my $L = new SUGR::Line($X,$Y);
  return ConstructFromPointLine($Z,$L);
}

# compute 3d plane from image line and projection matrix
sub ConstructFromLine2dMatrix {
    my ($l,$P) = @_;

    my $A = ~$P x $l->vec;
    my $Cov = defined($l->cov) ?
      (~$P x $l->cov x $P) :
	undef;

    return sugr_Plane($A,$Cov);
}
sub ConstructFromMatrixLine2d { ConstructFromLine2dMatrix(@_[1,0]) }


# compute 3d plane from image line and projection matrix
sub ConstructFromLine2dProjection {
  my ($l,$P) = @_;
  
  my $A = ~$P->matrix x $l->vec;
  my $Cov = defined($l->cov) ?
    (~$P->matrix x $l->cov x $P->matrix) :
      undef;
  
  if (defined($P->cov("point"))) {
    my $J = mzeroes(4,12);
    ($_ = $J->slice("0:3,0:3")->diagonal(0,1))  .= $l->vec->at(0);
    ($_ = $J->slice("0:3,4:7")->diagonal(0,1))  .= $l->vec->at(1);
    ($_ = $J->slice("0:3,8:11")->diagonal(0,1)) .= $l->vec->at(2);

#    print "test projecting plane :",$A->clump(2)," - ",clump($J x $P->vec("point"),2),"\n";

    $Cov += ( $J x $P->cov("point") x ~$J); # this is also works if
                                            # $Cov was not defined before!
  } 

  return new SUGR::Plane($A,$Cov);
}
sub ConstructFromProjectionLine2d { ConstructFromLine2dProjection(@_[1,0]) }

sub ConstructFromPointProjectiveCamera {
  ConstructFromPointProjection(@_)
}
sub ConstructFromProjectiveCameraPoint {
  ConstructFromPointProjection(@_[1,0])
}

# matrix representation of the entity
sub matrix {
  ~Pidual($_[0]);
}

# a dual of a plane is a point
sub dual {
  my ($self) = @_;
  my $copy = $self->copy();
  bless $copy,"SUGR::Point";
}

sub type {
  "Plane";
}

=pod

=back

=head2 Class and object methods

See the L<SUGR::Element(3)|SUGR::Element> manual page for inherited
methods.

=cut


# distance to origin
sub distorigin {
  my ($self) = @_;
  my $pdlvec = $self->vec->slice("0:2,(0)");
  (sqrt($self->vec->at(3)/PDL::inner($pdlvec,$pdlvec)))->at(0);
}

sub euclid_scale {
  my ($self) = shift;
  return sqrt(PDL::inner($self->vec->slice("0:2,(0)"),$self->vec->slice("0:2,(0)")));
}

# use by euclidize
sub Jacobian_norm_euclid {
  my ($self) = @_;
  my @A = $self->vec->list;
  my $A1A2A3 = $A[0]**2+$A[1]**2+$A[2]**2;
  die $self." is not a euclidean element." if $A1A2A3==0;
  my $A1A2A3_32 = (($A1A2A3)**(3/2));

  mpdl([[-$A[0]**2/($A1A2A3_32)+1/(sqrt($A1A2A3)), 
	 -$A[0]*$A[1]/($A1A2A3_32),
	 -$A[0]*$A[2]/($A1A2A3_32), 
	 0], 
	[-$A[0]*$A[1]/($A1A2A3_32), 
	 -$A[1]**2/($A1A2A3_32)+1/(sqrt($A1A2A3)), 
	 -$A[1]*$A[2]/($A1A2A3_32), 
	 0], 
	[-$A[0]*$A[2]/($A1A2A3_32), 
	 -$A[1]*$A[2]/($A1A2A3_32), 
	 -$A[2]**2/($A1A2A3_32)+1/(sqrt($A1A2A3)), 
	 0], 
	[-$A[3]*$A[0]/($A1A2A3_32), 
	 -$A[3]*$A[1]/($A1A2A3_32), 
	 -$A[3]*$A[2]/($A1A2A3_32), 
	 1/(sqrt($A1A2A3))]]);
}


sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian for euclidean scaling
  my $M = SUGR::Element::Identity(4);
  ($_ = $M->diagonal(0,1)->PDL::slice("3")) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_homogeneous_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->slice("0:2")**2));
}

sub norm_of_euclidean_part {
  my $self = shift;
  PDL::sum(abs($self->vec->slice("3")));
}


sub Constraint_Euclidean {
  my ($A) = @_;
  $A = (ref($A) =~ /SUGR/) ? $A->vec : $A;
  my $ret = vpdl([ 0.5*PDL::sum($A->slice("0:2")**2)  - 0.5 ]);
  return $ret;
}
sub Jacobian_Constraint_Euclidean {
  my ($A) = @_;
  my $v = (ref($A) =~ /SUGR/) ? $A->vec : $A;
  my $J = vzeroes(4);
  ($_ = $J->slice("0:2")) .= $v->slice("0:2");
  return ~$J;
}


sub Constraint_Sphere {
  my ($A,$squared_radius) = @_;
  $A = (ref($A) =~ /SUGR/) ? $A->vec : $A;
  $squared_radius ||=1;
  return vpdl([ 0.5*PDL::sum($A**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($A) = @_;
  return (ref($A) =~ /SUGR/) ? ~($A->vec) : ~$A;
}

return 1 if caller;

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Estimation(3), SUGR::Element(3), PDL::Matrix(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

my $A = sugr_Plane([1,2,3,1],mones(4,4));

print $A;
print $A->matrix;
print $A->Jacobian_Constraint_Euclidean();
print $A->Jacobian_Constraint_Sphere();
$A->scale_euclidean_part(0.1);
print $A->Jacobian_Constraint_Euclidean();
