=head1 NAME

SUGR::Point -- class for uncertain points in 3D

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Point->VERSION

=head1 SYNOPSIS

  use SUGR;

  $X = new SUGR::Line([$xc,$yc,$zc,1],[[0.01,0,0,0]
				       [0,0.01,0,0]
				       [0,0,0.01,0],
				       [0,0,0   ,0]]);

=head1 DESCRIPTION

=head2 Parameter vector

The parameter vector is a homogeneous 4-vector. Setting the last
element to 1, the first three elements refer to the Euclidean
coordinates $xc, $yc and $zc.

=head2 Constructor new()

The C<new> constructor is quite flexible and allows several different
creation methods. Note that the arguments can be swapped.

=over 4

=item - 

Construction by specifying parameter vector and covariance matrix. 

The parameter vector can be specified either as a PDL::Matrix object
or as a reference to a Perl array (the covariance matrix then is a
reference to Perl array of arrays, see perldoc perldsc for more
details).

You may omit the covariance matrix, putting up with the lack of error
propagation for this element.

=item -

Construct from three planes $A1, $A2, $A3

  # ref($A1) eq "SUGR::Plane"
  # ref($A2) eq "SUGR::Plane"
  # ref($A3) eq "SUGR::Plane"
  $x = new SUGR::Point($A1,$A2,$A3);

=item -

Construct from a 3D plane $A and a 3D line $L

  # ref($A) eq "SUGR::Point"
  # ref($L) eq "SUGR::Line"
  $x = new SUGR::Point($A,$L);

=item - 

Estimation of a 2D point from sets of observations. The observations
may be incident 2D lines or equal 2D points. The observations will be
specified in a hash. See also <SUGR::Estimation(3)|SUGR::Estimation>
for more details on the input and the technique.

  $X = new SUGR::Point({
			"IncidentLine" => [ $L1, $L2, $L3 ],
			"IncidentPlane" => [ $A1, $A2, $A3 ],
			"EqualPoint" => [ $X1, $X2 ]
		       });

=cut

package SUGR::Point;


use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

no warnings 'redefine';

use Exporter;
use SUGR::Element;
use SUGR::Estimation ();
$VERSION = 0.01;
@ISA = qw(Exporter SUGR::Element);

@EXPORT = qw(sugr_Point); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL::Math;
use PDL::Matrix;
use SUGR::Basic qw/Pi/;

# this is a constructor that can be called like it is
sub sugr_Point { SUGR::Point->new(@_) }

sub ConstructFromHASH {
  my ($arghash) = @_;
  
  my $result = SUGR::Estimation::Estimate("Point",$arghash);

  return $result->{Point};
}

# returns sugr_Point-object from sugr_Plane $P and sugr_Line $L
sub ConstructFromPlaneLine {
  my ($A,$L) = @_;
  my $X   = ~$A->Pi x $L->vec;
  my $Cov = undef;
  $Cov = defined($A->cov) ?
    ($L->Gammadual x $A->cov x  ~$L->Gammadual) :  undef;
  $Cov += defined($L->cov) ?
    (~$A->Pi x $L->cov x $A->Pi ) :  undef;
  return sugr_Point($X,$Cov);
}
sub ConstructFromLinePlane { ConstructFromPlaneLine(@_[1,0]) }

sub ConstructFromPlanePlanePlane {
  my ($A,$B,$C) = @_;
  my $L = new SUGR::Line($A,$B);
  return ConstructFromPlaneLine($C,$L);
}

sub ConstructFromInverseProjectiveCameraPoint2d {
  my ($P_inv,$x) = @_;

  my $X   = $P_inv->matrix x $x->vec;
  my $Cov = undef;
#   $Cov += defined($P_inv->cov) ?
#     ($L->Gammadual x $A->cov x  ~$L->Gammadual) :  undef;
#  $Cov += defined($x->cov) ?
#     ($P_inv->matrix x $x->cov x ~$P_inv->matrix ) :  undef;
  return sugr_Point($X,$Cov);

}

sub ConstructFromPoint2dInverseProjectiveCamera {
  ConstructFromInverseProjectiveCameraPoint2d(@_[1,0]);
}

# methods
# this is actually an estimation !!!!
sub ConstructFromImagePoints {
  my ($x1,$P1s,$x2,$P2s) = @_;
  
  my @L;

  # maybe we have to compute the line projection matrix from the point matrix
  $P1s = LineProjectionFromPointProjection($P1s) if (join(",",$P1s->mdims) =~ /3,4/);
  $P2s = LineProjectionFromPointProjection($P2s) if (join(",",$P2s->mdims) =~ /3,4/);

  # compute the point rays
  push @L , sugr_Line($x1,$P1s);
  push @L , sugr_Line($x2,$P2s);
  
  # compute 3D point out of the two 3D lines
  # no covariance matrix at the moment
  my $M = mzeroes(4,4);
  for (my $i=0;$i<2;$i++) {
    $M += ~(SUGR::Element::Gamma($L[$i]->vec)) x SUGR::Element::Gamma($L[$i]->vec);
  }
  my ($evector,$evalue) = eigens($M);
  my $iev = qsorti $evalue;
  # eigenvector for smallest eigenvalue
  my $sev = $evector->slice(",(".$iev->at(0).")");

  # return computed point
  sugr_Point(vpdl($sev/$sev->at(3)));
}

# compute projection matrix for lines
sub LineProjectionFromPointProjection {
  my ($P) = @_;

  my $r1 = ~$P->slice("(0),");
  my $r2 = ~$P->slice("(1),");
  my $r3 = ~$P->slice("(2),");

  my $rs1 = SUGR::Element::Pidual($r2) x $r3;
  my $rs2 = SUGR::Element::Pidual($r3) x $r1;
  my $rs3 = SUGR::Element::Pidual($r1) x $r2;

  my $Ps = mzeroes(3,6);
  ($_ = $Ps->slice("0:0,0:5")) .= ~$rs1;
  ($_ = $Ps->slice("1:1,0:5")) .= ~$rs2;
  ($_ = $Ps->slice("2:2,0:5")) .= ~$rs3;

  return $Ps;
}

sub matrix {
  ~Pi($_[0]);
}

# a dual of a point is a plane
sub dual {
  my ($self) = @_;
  my $copy = $self->copy();
  bless $copy,"SUGR::Plane";
}

sub type {
  "Point";
}

=pod

=back

=head2 Class and object methods

See the L<SUGR::Element(3)|SUGR::Element> manual page for inherited
methods.

=cut

# distance to origin 
# TODO: can be put in base class using
# norm_of_euclidean_part and norm_of_homogeneous_part
sub distorigin {
  my ($self) = @_;
  my $pdlvec = $self->vec->slice("0:2,(0)");
  (sqrt(PDL::inner($pdlvec,$pdlvec))/$self->vec->at(3))->at(0);
}

# distance from one element to the other
# TODO: must optimized ( no euclidize call)
sub distance {
  my ($self,$other) = @_;
  my $pdlvec = $self->euclidize->vec->slice("0:2,(0)") 
    - $other->euclidize->vec->slice("0:2,(0)");
  (sqrt(PDL::inner($pdlvec,$pdlvec)->at(0)));
}

sub euclid_scale {
  my ($self) = shift;
  return $self->vec->at(3);
}

# needed for euclidize
sub Jacobian_norm_euclid {
  my ($self) = @_;
  my @x = $self->vec->list;
  die $self." is not a euclidean element." if $x[3]==0;
  mpdl([[1/$x[3], 0      , 0      , -$x[0]/($x[3]**2)], 
	[0,       1/$x[3], 0      , -$x[1]/($x[3]**2)], 
	[0      , 0      , 1/$x[3], -$x[2]/($x[3]**2)], 
	[0      , 0      , 0      ,                0 ]]);
}

sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian
  my $M = SUGR::Element::Identity(4);
  ($_ = $M->diagonal(0,1)->PDL::slice("0:2")) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_euclidean_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->slice("0:2")**2));
}

sub norm_of_homogeneous_part {
  my $self = shift;
  PDL::sum(abs($self->vec->slice("3")));
}

sub Constraint_Euclidean {
  my ($X) = @_;
  $X = (ref($X) =~ /SUGR/) ? $X->vec : $X;
  return vpdl([$X->at(3) - 1]);
}
sub Jacobian_Constraint_Euclidean {
  ~vpdl([0,0,0,1]);
}

sub Constraint_Sphere {
  my ($X,$squared_radius) = @_;
  $X = (ref($X) =~ /SUGR/) ? $X->vec : $X;
  $squared_radius ||=1;
  return vpdl([ 0.5*PDL::sum($X**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($X) = @_;
  $X = (ref($X) =~ /SUGR/) ? $X->vec : $X;
  ~$X;
}

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Estimation(3), SUGR::Element(3), PDL::Matrix(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

return 1 if caller;

################################################################################
# Test code
use SUGR::Line;

my $X = sugr_Point([1,2,3,1],mones(4,4));

print "Point : ",$X;
print "matrix representation: ",$X->matrix;
print "Jacobian of Euclidean Constraint ",$X->Jacobian_Constraint_Euclidean();
print "Jacobian of Homogeneous Constraint ",$X->Jacobian_Constraint_Sphere();
$X->scale_euclidean_part(0.1);
print $X->Jacobian_Constraint_Euclidean();

my $X1 = sugr_Point([1,2,3,1],mones(4,4));
my $X2 = sugr_Point([4,2,3,1],mones(4,4));
my $X3 = sugr_Point([1,2,5,1],mones(4,4));
my $X4 = sugr_Point([4,2,6,1],mones(4,4));

my $L1 = new SUGR::Line($X1,$X2);
my $L2 = new SUGR::Line($X3,$X4);

my $XL = new SUGR::Point({
			  "IncidentLine" => [$L1,$L2],
			 });
print $XL;

#the following code just tests whether a direct solution yields the
#same result as an estimation with minimal set of observations

#use SUGR::Plane;
#use SUGR::Line;
#use SUGR::Cuboid;

# my %Objects = generate_cuboid(no_points => 1,
# 			      no_planes => 1,
# 			      no_lines => 1,
# 			      random_seed => [4711,222712]);
# %Objects = transform_cuboid(translation => [1,2,2],
#  			    rotation    => [0.5,0.7,1.3],
#  			    %Objects);

# my $X_cube  = $Objects{Points};
# my $L_cube  = $Objects{Lines};
# my $A_cube  = $Objects{Planes};


# my $Xs = new SUGR::Point({
# 			  "IncidentPlane" => [$A_cube->[0]->[0]],
# 			  "IncidentLine" => [$L_cube->[4]->[0]],
# 			 });

# print "Estimated: $Xs\n-------------------------------------\n";

# print "Directly constructed: ",sugr_Point($A_cube->[0]->[0],$L_cube->[4]->[0]);

