=head1 NAME

SUGR::Point2d -- class for uncertain 2D points

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Point2d->VERSION

=head1 SYNOPSIS

  use SUGR;

  $x = new SUGR::Point2d([$xc,$yc,1],[[0.01,0,0]
				      [0,0.01,0]
				      [0,0,0   ]]);

=head1 DESCRIPTION

=head2 Parameter vector

The parameter vector is a homogeneous 2-vector. Setting the last
coordinate to 1, the first two coordinates are equal to the Euclidean
coordinates $xc, $yc.

=head2 Constructor new()

The C<new> constructor is quite flexible and allows several different
creation methods. Note that the arguments may be swapped.

=over 4

=item - 

Construction by specifying parameter vector and covariance matrix. 

The parameter vector can be specified either as a PDL::Matrix object
or as a reference to a Perl array (the covariance matrix then is a
reference to Perl array of arrays, see perldoc perldsc for more
details).

You may omit the covariance matrix, putting up with the lack of error
propagation for this element.

=item -

Construct from two 2D lines

  # ref($l1) eq "SUGR::Line2d"
  # ref($l2) eq "SUGR::Line2d"
  $x = new SUGR::Point2d($l1,$l2);

=item -

Construct from 3D point and pin-hole camera

  # ref($X) eq "SUGR::Point"
  # ref($P) eq "SUGR::ProjectiveCamera"
  $x = new SUGR::Point2d($X,$P);

=item - 

Estimation of a 2D point from sets of observations. The observations
may be incident 2D lines or equal 2D points. The observations will be
specified in a hash. See also <SUGR::Estimation(3)|SUGR::Estimation>
for more details on the input and the technique.

  $x = new SUGR::Point2d({
			  "IncidentLine2d" => [ $l1, $l2, $l3 ],
			  "EqualPoint2d" => [ $x1, $x2 ]
			 });

=cut

package SUGR::Point2d;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::Element;
use SUGR::Estimation ();
$VERSION = SUGR->VERSION;
@ISA = qw(Exporter SUGR::Element);

@EXPORT = qw(sugr_Point2d); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL::Matrix;
use SUGR::Basic qw/:ALL/;

######################################################################
#
# Constructors
#

# this is a constructor that can be called like it is
sub sugr_Point2d { SUGR::Point2d->new(@_); }

sub ConstructFromHASH {
  my ($arghash) = @_;
  
  my $result = SUGR::Estimation::Estimate("Point2d",$arghash);

  return $result->{Point2d};
}

# returns sugr_Point2d-object from two sugr_Line2d-objects
sub ConstructFromLine2dLine2d {
  my ($l,$m) = @_;
  
  my $x   = S($l) x $m->vec;
  my $cov = (defined($l->cov) && defined($m->cov)) ?
    (S($l) x $m->cov x ~S($l) +
     S($m) x $l->cov x ~S($m)  ) :
       undef;
  return sugr_Point2d($x,$cov);
}

# project 3D point into image
sub ConstructFromPointMatrix {
    my ($X,$P) = @_;

    my $x = $P x $X->vec;
    my $cov =  defined($X->cov) ? 
      ($P x $X->cov x ~$P) :
	undef;

    return sugr_Point2d($x,$cov);
}
sub ConstructFromMatrixPoint { ConstructFromPointMatrix(@_[1,0]) }

sub ConstructFromPointProjection {
  my ($X,$P) = @_;

  my $x = $P->matrix x $X->vec;

  my $Cov = defined($X->cov) ?
    ($P->matrix x $X->cov x ~$P->matrix) :
      undef;

  if (defined($P->cov)) {
    my $J = mzeroes(3,12);
    ($_ = $J->slice("0,0:3"))  .= ~$X->vec;
    ($_ = $J->slice("1,4:7"))  .= ~$X->vec;
    ($_ = $J->slice("2,8:11")) .= ~$X->vec;

#    print "test projection of point:",$x->clump(2)," - ",clump($J x $P->vec("point"),2),"\n";

    $Cov += ( $J x $P->cov x ~$J); # this is also works if
                                   # $Cov was not defined before!
  } 
  
  return sugr_Point2d($x,$Cov);
}
sub ConstructFromProjectionPoint { ConstructFromPointProjection(@_[1,0]) }

sub ConstructFromPointProjectiveCamera {
  ConstructFromPointProjection(@_)
}
sub ConstructFromProjectiveCameraPoint {
  ConstructFromPointProjection(@_[1,0])
}

=pod

=back

=head2 Class and object methods

See the L<SUGR::Element(3)|SUGR::Element> manual page for inherited
methods.

=cut

#
# methods
#

# make the following a package global so that one does not need to create a pdl
# vector each time this method is called;
my $euclidean_nullspace = vpdl([0,0,1]);
sub _euclidean_nullspace {
  return $euclidean_nullspace;
}


# matrix representation of the entity
sub matrix {
  S($_[0]);
}

sub dual {
  my ($self) = @_;
  my $p = $self->copy;
  bless $p, "SUGR::Line2d";
}

sub type {
  "Point2d";
}

sub euclid_scale {
  my ($self) = shift;
  return $self->vec->at(2);
}

# needed for euclidize
sub Jacobian_norm_euclid {
  my ($self) = @_;
  my @x = $self->vec->list;
  die $self." is not a euclidean element." if $x[2]==0;
  mpdl([[1/$x[2], 0      , -$x[0]/($x[2]**2)], 
	[0      , 1/$x[2], -$x[1]/($x[2]**2)], 
	[0      , 0      ,                0 ]]);
}


sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian
  my $M = SUGR::Element::Identity(3);
  ($_ = $M->diagonal(0,1)->PDL::slice("0:1")) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_euclidean_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->slice("0:1")**2));
}

sub norm_of_homogeneous_part {
  my $self = shift;
  PDL::sum(PDL::abs($self->vec->slice("2")));
}

sub Constraint_Euclidean {
  my ($x) = @_;
  $x = (ref($x) =~ /SUGR/) ? $x->vec : $x;
  return vpdl([$x->at(2) - 1]);
}
sub Jacobian_Constraint_Euclidean {
  ~vpdl([0,0,1]);
}

sub Constraint_Sphere {
  my ($x,$squared_radius) = @_;
  $x = (ref($x) =~ /SUGR/) ? $x->vec : $x;
  $squared_radius ||=1;
  return vpdl([ 0.5*sum($x**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($x) = @_;
  return  (ref($x) =~ /SUGR/) ? ~$x->vec : ~$x;
}

return 1 if caller;

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Estimation(3), SUGR::Element(3), PDL::Matrix(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

# test code

my $x = new SUGR::Point2d([1,2,1],mones(3,3));

print $x;
print $x->matrix;
print $x->Jacobian_Constraint_Euclidean();
print $x->Jacobian_Constraint_Sphere();
$x->scale_euclidean_part(0.1);
print $x->Jacobian_Constraint_Euclidean();
