=head1 NAME

SUGR::ProjectiveCamera -- class for the projective transformation from
3D space to a 2D plane

=head1 NOT DOCUMENTED YET

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Element(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

package SUGR::ProjectiveCamera;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
use SUGR::TransformationNew;

$VERSION = SUGR->VERSION;
@ISA = qw(Exporter SUGR::TransformationNew);

@EXPORT = qw(sugr_ProjectiveCamera); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use SUGR::Basic qw/:ALL/;

# this is a constructor that can be called like it is
sub sugr_ProjectiveCamera { 
  SUGR::ProjectiveCamera->new(@_);
}

sub ConstructFromHASH { 
  my @args = @_;

  # we may want to specify the camera parameters explicitly and build
  # the projection matrix from there. Then we can pass a hash to this
  # function and everything is cool.
  die "Error when calling constructor from HASH " unless (ref($args[0]) =~ /HASH/);

  my %camera = %{$args[0]};
  #    print "$_ -> $camera{$_}\n" foreach (keys %camera);

  ###################################################
  # process all arguments in the hash
  my $X0  = $camera{X0};
  my $R   = $camera{R};
  my $c   = defined($camera{c})  ? $camera{c}  : 1;
  my $pH  = defined($camera{pH}) ? $camera{pH} : vpdl([0,0]);
  my $s   = defined($camera{s})  ? $camera{s}  : 0;
  my $m   = defined($camera{m})  ? $camera{m}  : 0;

  # maybe we have information on the uncertainty of the elements,
  # we store them in the following variables
  my $X0_cov  = defined($camera{X0_cov})  ? $camera{X0_cov} : undef;
  my $R_cov   = defined($camera{R_cov})   ? $camera{R_cov}  : undef;;
  my $c_var   = defined($camera{c_var})   ? $camera{c_var}  : undef;
  my $pH_cov  = defined($camera{pH_cov})  ? $camera{pH_cov} : undef
    my $s_var   = defined($camera{s_var})   ? $camera{s_var}  : undef;
  my $m_var   = defined($camera{m_var})   ? $camera{m_var}  : undef;

  # check if $X0 or $pH are SUGR elements
  if (ref($X0) =~ /SUGR/) {
    my $X0_euclid = $X0->copy->euclidize;
    $X0 = $X0_euclid->vec->slice("0:2"); # maybe this is a memory leak?!
    if (defined($X0_euclid->cov)) {
      $X0_cov = $X0_euclid->cov->slice("0:2,0:2");
    }
  }
  if (ref($pH) =~ /SUGR/) {
    my $pH_euclid = $pH->copy->euclidize;
    $pH = $pH->vec->slice("0:1"); # maybe this is a memory leak?!
    if (defined($pH_euclid->cov)) {
      $pH_cov = $pH_euclid->cov->slice("0:1,0:1");
    }
  }


  ########################################################
  # now do actual computation of the projection matrix
  my $P = mzeroes(3,4);
  ($_ = $P->slice("0:2,0:2")) .= $R; # rotation 
  ($_ = $P->slice("0:2,3:3")) .= -$R x $X0; # translation
  # calibration matrix
  my $K = SUGR::Basic::Identity(3);
  $K->set(0,0,$c);
  $K->set(0,1,$c*$s);
  $K->set(1,1,$c*(1+$m));
  ($_ = $K->slice("0:1,2:2")) .= $pH;

  $P = $K x $P;
    
  ##################################################
  # now covariances, as the input is defined this way, we assume
  # some parameters to be uncorrelated: e.g. projection center $X0
  # and rotation matrix $R are uncorrelated
  my $Pideal_cov = mzeroes(12,12);
  # exterior parameters
  my @x0_id = (3,7,11);
  if (defined($X0_cov)) {
    # error propagation for rotated projection center -RX0
    my $RX0_cov = $R x $X0_cov x ~$R;
    foreach my $i (0..2) {
      foreach my $j (0..2) {
	$Pideal_cov->set($x0_id[$i],$x0_id[$j],$RX0_cov->at($i,$j)); 
      }
    } 
  }
  my @R_id = (0,1,2,4,5,6,8,9,10);
  if (defined($R_cov)) {
    foreach my $i (0..8) {
      foreach my $j (0..8) {
	$Pideal_cov->set($R_id[$i],$R_id[$j],$R_cov->at($i,$j)); 
      }
    }
  }   

  # internal parameters 
  my $K_cov = mzeroes(9,9);
  if (defined($c_var)) {
    $K_cov->set(0,0,$c_var);
    $K_cov->set(1,1,$c_var);
  }
  if (defined($s_var)) {
    $K_cov->set(0,1,($s**2*$c_var+ $c**2 * $s_var));
  }
  # replace element (1,1) if variance for m is defined
  if (defined($m_var)) {
    $K_cov->set(1,1,($m**2*$c_var+ $c**2 * $m_var));
  }
  my @pH_id = (2,5);
  if (defined($pH_cov)) {
    foreach my $i (0..1) {
      foreach my $j (0..1) {
	$K_cov->set($pH_id[$i],$pH_id[$j],$pH_cov->at($i,$j)); 
      }
    }
  }   

  # Now this becomes really ugly, but what the heck. We have to
  # define the calibration matrix $K for the _VECTOR_ of the
  # projection matrix,i.e. $P->vec. 
  my $Kmatrix = kroneckerproduct( $K,Identity(4));
  my $Transpose33 = permutation_vec_transpose(3,3);
  my $Transpose34 = permutation_vec_transpose(3,4);
  my $Pmatrix = $Transpose34 x kroneckerproduct(SUGR::Basic::Identity(3),~$P);

  #     my $P_cov = $Kmatrix x $Pideal_cov x ~$Kmatrix +
  #       $Pmatrix x ($Transpose33 x $K_cov) x ~$Pmatrix;
  my $P_cov = $Kmatrix x $Pideal_cov x ~$Kmatrix + 
    $Pmatrix x $K_cov x ~$Pmatrix;

  # as arguments for the actual constructor, we need the projection
  # matrix ...
  @args = ($P);
  # ... and the covariance matrix if it's not a null matrix
  push @args,$P_cov if $P_cov->abs->sum != 0;

  SUGR::ProjectiveCamera->new(@args);
}

# the following two constructor helper functions may be rewritten less
# redundantly, since they actually only check dimensions, but what the
# heck!

# constructor with no covariance matrix
sub ConstructFromMatrix { 
  ConstructFromMatrixMatrix(@_);
}

# this is the actual constructor !
sub ConstructFromMatrixMatrix {
  my ($matrix,$cov) = @_;
  croak("no transformation matrix given") if !defined($matrix);

  # check dimensions
  my ($r,$c) = $matrix->mdims;
  if ($r*$c==12) {
    $matrix = $matrix->reshape(3,4);
  }
  else {
    croak("first argument should be a (3,4) matrix, is $matrix");
  }

  my $self = defined($cov) ?
    SUGR::TransformationNew::ConstructFromMatrixMatrix($matrix,$cov) :
	SUGR::TransformationNew::ConstructFromMatrix($matrix);

  # this is so that I don't have to compute the projection matrix for
  # lines over and over again.
  ($self->{vec_line},$self->{cov_line}) = LineProjection($self->{vec},$self->{cov});

  return $self;
}

sub type {
  "ProjectiveCamera";
}


sub result_type {
  my ($self,$type) = @_;
  $type ||= "point"; # default value
  $type = lc($type);
  
  # TODO: no error checking yet!
  return ($type =~ /point/) ? "SUGR::Point2d" : "SUGR::Line2d";
}


# compute projection matrix for lines, usually we normalize the
# outcoming projection matrix. If this is not desired, specify 1 as
# the third argument
sub LineProjection {
  my ($Pvec,$Pcov,$nonormalize) = @_;

  my $r1 = $Pvec->slice("0:3");
  my $r2 = $Pvec->slice("4:7");
  my $r3 = $Pvec->slice("8:11");

  my $rs1 = SUGR::Element::Pidual($r2) x $r3;
  my $rs2 = SUGR::Element::Pidual($r3) x $r1;
  my $rs3 = SUGR::Element::Pidual($r1) x $r2;

  my $Ps = mzeroes(3,6);
  ($_ = $Ps->slice("0:0,0:5")) .= ~$rs1;
  ($_ = $Ps->slice("1:1,0:5")) .= ~$rs2;
  ($_ = $Ps->slice("2:2,0:5")) .= ~$rs3;

  my $J = mzeroes(18,12);
  ($_ = $J->slice("0:5,8:11"))  .= SUGR::Element::Pidual($r2);
  ($_ = $J->slice("6:11,0:3"))  .= SUGR::Element::Pidual($r3);
  ($_ = $J->slice("12:17,4:7")) .= SUGR::Element::Pidual($r1);

  my $Cov = defined($Pcov) ? $J x $Pcov x ~$J : undef;

  return ($Ps->reshape(18,1),$Cov) if $nonormalize;

  # normalize $Ps
  my $factor = 1/(PDL::max(PDL::abs($Ps)));
  $Ps  *= $factor;
  $Cov *= $factor**2 if defined($Cov);

  return ($Ps->reshape(18,1),$Cov);
}


my @euclid_indices = [0,5,10];
my @homo_indices = [1..4,6..9,11];

sub euclid_scale {
  my ($self) = shift;
  return sqrt(PDL::inner($self->vec->dice(@euclid_indices)->clump(-1),
			 $self->vec->dice(@euclid_indices)->clump(-1)));
}

# needed for euclidize
sub Jacobian_norm_euclid {
  my ($self) = @_;
  warn "Projection::Transformation::Jacobian_norm_euclid: don't use this, I don't know what happens here!\n";

  my @P = $self->vec->list;
  my $P0P5P10 = $P[0]**2+$P[5]**2+$P[10]**2;
  die $self." is not a euclidean element." if $P0P5P10==0;
  my $P0P5P10_32 = (($P0P5P10)**(3/2));

  my $J = mzeroes(12,12);

  $J->set(0,0, -1/$P0P5P10_32*$P[0]*$P[0]+1/(sqrt($P0P5P10)));
  $J->set(0,5, -1/$P0P5P10_32*$P[0]*$P[5]);
  $J->set(0,10, -1/$P0P5P10_32*$P[0]*$P[10]);

  $J->set(1,0, -1/$P0P5P10_32*$P[1]*$P[0]);
  $J->set(1,1, 1/(sqrt($P0P5P10)));
  $J->set(1,5, -1/$P0P5P10_32*$P[1]*$P[5]);
  $J->set(1,10, -1/$P0P5P10_32*$P[1]*$P[10]);

  $J->set(2,0, -1/$P0P5P10_32*$P[2]*$P[0]);
  $J->set(2,2, 1/(sqrt($P0P5P10)));
  $J->set(2,5, -1/$P0P5P10_32*$P[2]*$P[5]);
  $J->set(2,10, -1/$P0P5P10_32*$P[2]*$P[10]);

  $J->set(3,0, -1/$P0P5P10_32*$P[3]*$P[0]);
  $J->set(3,3, 1/(sqrt($P0P5P10)));
  $J->set(3,5, -1/$P0P5P10_32*$P[3]*$P[5]);
  $J->set(3,10, -1/$P0P5P10_32*$P[3]*$P[10]);
  $J->set(3,11, 0.0);

  $J->set(4,0, -1/$P0P5P10_32*$P[4]*$P[0]);
  $J->set(4,4, 1/(sqrt($P0P5P10)));
  $J->set(4,5, -1/$P0P5P10_32*$P[4]*$P[5]);
  $J->set(4,10, -1/$P0P5P10_32*$P[4]*$P[10]);

  $J->set(5,0, -1/$P0P5P10_32*$P[0]*$P[5]);
  $J->set(5,5, -1/$P0P5P10_32*$P[5]*$P[5]+1/(sqrt($P0P5P10)));
  $J->set(5,10, -1/$P0P5P10_32*$P[5]*$P[10]);

  $J->set(6,0, -1/$P0P5P10_32*$P[6]*$P[0]);
  $J->set(6,5, -1/$P0P5P10_32*$P[6]*$P[5]);
  $J->set(6,6, 1/(sqrt($P0P5P10)));
  $J->set(6,10, -1/$P0P5P10_32*$P[6]*$P[10]);

  $J->set(7,0, -1/$P0P5P10_32*$P[7]*$P[0]);
  $J->set(7,5, -1/$P0P5P10_32*$P[7]*$P[5]);
  $J->set(7,7, 1/(sqrt($P0P5P10)));
  $J->set(7,10, -1/$P0P5P10_32*$P[7]*$P[10]);

  $J->set(8,0, -1/$P0P5P10_32*$P[8]*$P[0]);
  $J->set(8,5, -1/$P0P5P10_32*$P[8]*$P[5]);
  $J->set(8,8, 1/(sqrt($P0P5P10)));
  $J->set(8,10, -1/$P0P5P10_32*$P[8]*$P[10]);

  $J->set(9,0, -1/$P0P5P10_32*$P[9]*$P[0]);
  $J->set(9,5, -1/$P0P5P10_32*$P[9]*$P[5]);
  $J->set(9,9, 1/(sqrt($P0P5P10)));
  $J->set(9,10, -1/$P0P5P10_32*$P[9]*$P[10]);

  $J->set(10,0, -1/$P0P5P10_32*$P[0]*$P[10]);
  $J->set(10,5, -1/$P0P5P10_32*$P[5]*$P[10]);
  $J->set(10,10, -1/$P0P5P10_32*$P[10]*$P[10]+1/(sqrt($P0P5P10)));

  $J->set(11,0, -1/$P0P5P10_32*$P[11]*$P[0]);
  $J->set(11,5, -1/$P0P5P10_32*$P[11]*$P[5]);
  $J->set(11,10, -1/$P0P5P10_32*$P[11]*$P[10]);
  $J->set(11,11, 1/(sqrt($P0P5P10)));

  return $J;
}

sub scale_euclidean_part {
  my ($self,$scalar) = @_;

  # generate Jacobian
  my $M = SUGR::Element::Identity(12);
  ($_ = $M->diagonal(0,1)->dice(@euclid_indices)) .= $scalar;

  $self->setvec($M x $self->vec);
  $self->setcov($M x $self->cov x ~$M) if defined($self->cov);
  return $self;
}

sub norm_of_euclidean_part {
  my $self = shift;
  sqrt(PDL::sum($self->vec->dice(@euclid_indices)**2));
}

sub norm_of_homogeneous_part {
  my $self = shift;
  PDL::sum(abs($self->vec->dice(@homo_indices)));
}

sub Constraint_Euclidean {
  my ($P) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  my $ret =  vpdl([ 0.5*PDL::sum($P->dice(@homo_indices)**2) - 0.5 ]);
  return $ret;
}
sub Jacobian_Constraint_Euclidean {
  my ($P) = @_;
  my $v = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  my $J = vzeroes(12);
  ($_ = $J->dice(@homo_indices)) .= $v->dice(@homo_indices);
  
  return ~$J;
}

sub Constraint_Sphere {
  my ($P,$squared_radius) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  $squared_radius ||=1;
  return vpdl([ 0.5*PDL::sum($P**2) - 0.5*$squared_radius]);
}
sub Jacobian_Constraint_Sphere {
  my ($P) = @_;
  $P = (ref($P) =~ /SUGR/) ? $P->vec : $P;
  ~$P;
}

return 1 if caller;

#################################################################
# Test code
use PDL::Matrix;
use SUGR::Point;
use SUGR::Cuboid;
use SUGR::Point2d;
use SUGR::Line2d;
#use PDL::Graphics::PGPLOT;
use SUGR::MyListUtil qw/max min/;

my %Objects = generate_cuboid(no_points   => 1,
			      no_planes   => 1,
			      no_lines    => 1,
			      standard_deviation => [0.01,0.01,0.01],
			      random_seed => [4711,222712]);

%Objects = transform_cuboid(translation => [1,2,2],
			    rotation    => [0.5,0.7,1.3],
			    %Objects);


my $X_cube  = $Objects{Points};
my $L_cube  = $Objects{Lines};
my $A_cube  = $Objects{Planes};
my $X_cube_ideal  = $Objects{Ideal_Points};
my $L_cube_ideal  = $Objects{Ideal_Lines};
my $A_cube_ideal  = $Objects{Ideal_Planes};


my $X0 = sugr_Point([1.5,1.5,10,1],
		    0.01**2*mpdl([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,0]]));

my $R  = mpdl([[1,0,0],
	       [0,1,0],
	       [0,0,1]]);
my $P = sugr_ProjectiveCamera({ 
				       "X0"    => $X0,
				       "R"     => $R,
				       "c"     => 1,
				       "c_var" => 0.001,
				      });

print $P;
#print $P->cov;
#print $P->euclidize;
my (@xs,@ls);
my @Cube = (
	    [0,1,2,3], # face 1
	    [0,4,5,1], # face 2
	    [1,5,6,2], # face 3
	    [2,6,7,3], # face 4
	    [3,7,4,0], # face 5
	    [4,5,6,7], # face 6
	   );

print "Projecting noisy cube points ...\n";
foreach my $X_cubepoint (map { $_->[0] } (@$X_cube)) {
#   push @xs, sugr_Point2d($X_cubepoint,$P)->euclidize; 
  push @xs, $P->transform($X_cubepoint)->euclidize; 

#   print "<<<<<<<<",
#     sugr_Point2d($X_cubepoint,$P)->euclidize,
#     $P->transform($X_cubepoint)->euclidize,"<<<<<<<<<<\n" ;
#   print $xs[-1]->vec->slice("0:1")->clump(-1),"\n";
}

foreach my $L_cubeline (map { $_->[0] } (@$L_cube)) {
#  push @ls, sugr_Line2d($L_cubeline,$P)->euclidize; 
  push @ls, $P->transform($L_cubeline)->euclidize; 
}

my @o_points = ($P, $xs[0], $X_cube->[0]->[0]);
my @o_lines  = ($P, $ls[0], $L_cube->[0]->[0]);

my ($factor_points,@o_points_normalized) = normalize_sugr_objects(@o_points);
my ($factor_lines, @o_lines_normalized)  = normalize_sugr_objects(@o_lines);


if ($factor_points !=1 || $factor_lines !=1) {
  print "normalizing factor: points = $factor_points, lines = $factor_lines\n";
#  use SUGR::Relation;
  foreach my $i (0..2) {
    my $pointdiff = 
      PDL::sum(SUGR::Relation::Jacobian_IncidentProjectionPoint2dPoint((map { $_->vec } @o_points),$i) x $o_points[$i]->vec -
	       SUGR::Relation::Jacobian_IncidentProjectionPoint2dPoint((map { $_->vec } @o_points_normalized),$i) x $o_points_normalized[$i]->vec);
    my $linediff = 
      PDL::sum(SUGR::Relation::Jacobian_IncidentProjectionLine2dLine((map { $_->vec } @o_lines),$i) x $o_lines[$i]->vec -
	       SUGR::Relation::Jacobian_IncidentProjectionLine2dLine((map { $_->vec } @o_lines_normalized),$i) x $o_lines_normalized[$i]->vec);
    print "difference original - normalized with $i th Jacobian of points: $pointdiff\n";
    print "difference original - normalized with $i th Jacobian of lines:  $linediff\n";
    die "SCRIPT STOPPED... CHECK normalize_sugr_objects for camera elemnts\n" if (abs($linediff) > 1e-10 ||
									   abs($pointdiff) > 1e-10);
  }
}

my $xmin = min map { $_->vec->at(0) } @xs;
my $ymin = min map { $_->vec->at(1) } @xs;
my $xmax = max map { $_->vec->at(0) } @xs;
my $ymax = max map { $_->vec->at(1) } @xs;

my $i=1;
my $win = PDL::Graphics::PGPLOT::Window->new({Device => '/xs', 
					      BORDER => { TYPE => 'rel', VALUE => 0.05 },
					      justify => 1});
$win->env( $xmin, $xmax, $ymin, $ymax);

foreach my $line (@Cube) {
  my $x = PDL->new(map { $xs[$_]->vec->at(0) } @$line);
  my $y = PDL->new(map { $xs[$_]->vec->at(1) } @$line);
  $win->poly($x,$y, {COLOR=> $i++, FILL=> 2});
  $win->hold;
}


__END__
