=head1 NAME

SUGR::Relation -- base class for geometrical relations 

=head1 NOT DOCUMENTED YET

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut


package SUGR::Relation;

use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;

@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request

use SUGR::Basic qw/:ALL/;
use PDL;
use PDL::Matrix;
use PDL::Math;
use PDL::Slatec;

$VERSION = SUGR->VERSION;

my $relation_cnt = 0;

# abstract method, do not call this
sub new {
  my ($protoverse, @objects) = @_;
  my $class = ref($protoverse) || $protoverse;

  die "do not call this from $class" unless $class->can("relation_type");

  my $relation_type = $class->relation_type;
  
  no strict "refs";
  my %test_result = &$relation_type(@objects);
  my $self = { %test_result };
  $self->{pair_ids} = [ map { $_->id } @objects];
  $self->{id} = $class."_".($relation_cnt++);
  bless $self, $class;
}

sub stringify {
  my $self = shift;
  join(",",$self->id,$self->bool,$self->testvalue);
}
use overload q{""} => \&stringify;

sub id :lvalue {
  $_[0]->{id}
}

sub bool :lvalue { 
  $_[0]->{bool}
}

sub testvalue :lvalue { 
  $_[0]->{testvalue}
}

sub testratio :lvalue { 
  $_[0]->{testratio}
}

############################################################
#
# Tests for Geometrical Relations

# general testing code for all 18 tests, this loop generates all 18
# testing methods for equality, incidence orthogonality and
# parallelity. 
foreach my $Test qw/Equal Incident Orthogonal Parallel/ {
  no strict "refs"; # so that we can safely define a closure

  foreach my $object_types (@{$possible_test_types{$Test}}) {
    # generating testnames like "IncidentPointLine"
    my $Testname = $Test.join("",@{$object_types});
    # now define the closure, 
    # see also Programming Perl, 3rd Edition, p.261 f.
    *$Testname = sub { 
      @_ = grep { defined } @_; # make sure that we only use
                                # defined arguments
      my ($alpha,@o_in) = @_[-1..$#_-1]; # for the argument list, take
                                         # the last argument as the
                                         # first 
      my ($factor,@o) = normalize_sugr_objects(@o_in);
      my ($o1,$o2) = @o;
      my @o_vec = map { $_->vec } @o;

	
      # determine (restricted) Jacobians
      my $Jacobian_func = "Jacobian_$Testname";
      my $degree_of_freedoms = SUGR::Basic::DegFreedom($Testname);
      my @linear_indep_rows = ChooseRowIndices(&$Jacobian_func(@o_vec,0),
					       $degree_of_freedoms);


      # get Jacobians and choose those rows which are largest
      my @J;
      foreach my $i (0..$#o) {
	$J[$i] = ChooseRows(&$Jacobian_func(@o_vec,$i),
			    @linear_indep_rows);
#	print $J[$i];
      }
      
      # compute distance $d with any of the given Jacobians, let's say 0
      my $d = $J[0] x $o[0]->vec;
      
#      print "D == $d\n";

      # compute covariance matrix if possible
      my $cov = undef;
      foreach my $i (0..$#o) {
	next if !defined($o[$i]->cov);
	defined($cov) ? 
	  ($cov = $cov + ($J[$i] x $o[$i]->cov x ~$J[$i])) :
	    ($cov = ($J[$i] x $o[$i]->cov x ~$J[$i]));
#	print "o($i)->cov ",$o[$i]->cov," J($i) $J[$i] cov($i) $cov";
      }

      # just compare the result to package variable $eps if there was
      # no covariance matrix given
      return( bool => all((~$d x $d) < eps*$factor)) if (!defined($cov) || (ref($cov) !~ PDL));

      # compute the testvalue from covariance matrix
      my $test = ~$d x $cov->inv x $d;

      # check with chi-square value and we are done
      my $chis=chisqr(DegFreedom($Testname),$alpha);
      return ( bool => ($test->at(0,0)< $chis),
	       testratio => ($test->at(0,0)/ $chis),
	       testvalue => sqrt($test->at(0,0))
	     );
    };
    # that's it, we are done with the sub definition
    
    # maybe we have a different order for the name of the function
    # call, like IncidentLinePoint additionally of
    # IncidentPointLine. Then we just swap the arguments
    for my $i (0..$#{@{$object_types}}) {
      for my $j (($i+1)..$#{@{$object_types}}) {
	next if ($object_types->[$i] eq $object_types->[$j]);
	# swap id's temporarily
	my @o_types = @$object_types;
	@o_types[$j,$i] = @o_types[$i,$j];
	my $Testname_rev = $Test.join("",@o_types);
	*$Testname_rev = sub { @_[$j,$i] = @_[$i,$j]; &$Testname(@_) };
      }
    }
  }
}

# the overloading of the polymorphic operators works as the
# overloading trick for the constructor. Now I only do this four times
# for the four test types.
foreach my $func qw(Incident Equal Orthogonal Parallel) {
  no strict "refs";
  # define closure, i.e. a function with name $func
  *$func = sub {
    my ($op1,$op2,$op3,$op4) = @_;
    my ($s1,$s2,$s3) = (undef,undef,"");
    ($s1 = ref($op1)) =~ s/SUGR::(.+)/$1/;
    ($s2 = ref($op2)) =~ s/SUGR::(.+)/$1/;
    $op3 = 0.95 if !defined($op3);
    if (ref($op3) =~ /SUGR/) {
      $op4 = 0.95 if !defined($op4);
      ($s3 = ref($op3)) =~ s/SUGR::(.+)/$1/;
    } 
    die "no known SUGR objects given to $func" if ($s1.$s2.$s3 eq '');
    my $code = "$func$s1$s2$s3";
    no strict "refs";
    return () if !exists(&$code);
    return &$code($op1,$op2,$op3,$op4);
  };
  my $op_func = "op_$func";
  *$op_func = sub {
    my ($op1,$op2) = @_;
    my %test = &$func($op1,$op2);
    return $test{bool};
  };
  push @EXPORT,$func;
}

################################################################################
#
# Jacobians for all relations
#

# Equality
foreach my $type qw(Point2dPoint2d Line2dLine2d PointPoint LineLine PlanePlane) {
  my $Jacobian_Equaltype = "Jacobian_Equal$type";
  no strict "refs";
  # closure
  *$Jacobian_Equaltype  = sub { 
    my ($E1,$E2,$diff) = @_;
    my $neg = ($diff==0) ? 1 : -1;
    ($E1,$E2) = ($E2,$E1) if ($diff==0);
    return $neg * SUGR::Basic::C($E1,$E2);
  }
}

# make shortcuts for the above functions, so that
# Jacobian_Equal($x1,$x2) works
foreach my $func qw(Incident Equal Orthogonal Parallel) {
  no strict "refs";
  my $Jacobian_func = "Jacobian_$func";
  # create closure
  *$Jacobian_func = sub {
    my ($op1,$op2,$op3,$op4) = @_;
    my ($s1,$s2,$s3) = (undef,undef,"");
    my @args = ($op1->vec,$op2->vec);
    ($s1 = ref($op1)) =~ s/SUGR::(.+)/$1/;
    ($s2 = ref($op2)) =~ s/SUGR::(.+)/$1/;
    if (defined($op3) && ref($op3) =~ /SUGR/) {
      ($s3 = ref($op3)) =~ s/SUGR::(.+)/$1/;
      push @args, $op3->vec;
    }
    push @args, ($s3 eq "") ? $op3 : $op4; 

    my $code = "Jacobian_$func$s1$s2$s3";
    no strict "refs";
    return &$code(@args);
  };
  push @EXPORT_OK,"Jacobian_$func";
}

my @not = (1,0); # helps for calling Jacobian_xxx vice versa

# the same as @not, but for trilinear relations, a hack
my @permute210 = (2,1,0);
my @permute102 = (1,0,2);
my @permute120 = (1,2,0);
my @permute021 = (0,2,1);
my @permute201 = (2,0,1);

################################################################################
#
# Incidence 
#

# -> first the easy cases of incidence (with one-dimensional dot
# product)
sub Jacobian_IncidentDotProduct {
  my ($E1,$E2,$diff) = @_;
  my $Ediff = ($diff==0)  ? $E2 : $E1;
  (ref($Ediff) =~ /SUGR/) ? ~$Ediff->vec : ~$Ediff;
}
foreach my $type qw(PointPlane PlanePoint  
		    Line2dPoint2d Point2dLine2d) {
  my $Jacobian_Incidenttype = "Jacobian_Incident$type";
  no strict "refs";
  *$Jacobian_Incidenttype = sub { Jacobian_IncidentDotProduct(@_) };
}
sub Jacobian_IncidentLineLine {
  my ($E1,$E2,$diff) = @_;
  my $Ediff = ($diff==0) ? $E2 : $E1;
  (ref($Ediff) =~ /SUGR/) ? ~$Ediff->vec->rotate(3) : ~$Ediff->rotate(3);
}

# now the ones with Gamma and Pi matrices
sub Jacobian_IncidentPointLine {
  my ($X,$L,$diff) = @_;
  return ($diff==0) ? -Gamma($L) : transpose(Pidual($X));
}

sub Jacobian_IncidentLinePoint {
  Jacobian_IncidentPointLine(@_[1,0],$not[$_[2]]);
}

sub Jacobian_IncidentPlaneLine {
  my ($A,$L,$diff) = @_;
  return ($diff==0) ? -Gammadual($L) : transpose(Pi($A));
}

sub Jacobian_IncidentLinePlane {
  Jacobian_IncidentPlaneLine(@_[1,0],$not[$_[2]]);
}


# EXPERIMENTAL: 
# - incidence for 2D point, 3D point and projection matrix
sub Jacobian_IncidentProjectiveCameraPoint2dPoint {
  my ($P,$x,$X,$diff) = @_;

  # maybe the vectors $x and $X will be regarded as one unique
  # vector. Then we return a doubled Jacobian
  my $doublepack = 0;
  if ((!defined($x) || $x->dims == (0)) && defined($X) && join(",",$X->mdims) =~ /7,1/) {
    # correct indexing ?!?!
    $x = $X->slice("4:6")->copy;    $X = $X->slice("0:3");
    $doublepack = 1;
  }
  if ((!defined($X) || $X->dims == (0)) && defined($x) && join(",",$x->mdims) =~ /7,1/) {
    $X = $x->slice("3:6")->copy;    $x = $x->slice("0:2");    
    $doublepack = 1;
  }
#  print "($P,$x,$X,$diff)\n";

  # ...
  # need to do this for all combinations.
  if ($diff==0) {
    return S($x) x kroneckerproduct(Identity(3),~$X)
  }
  elsif ($diff==1) {
    my $ret = $doublepack ? 
      0.5*append(- S($P->reshape(3,4) x $X),
		 S($x) x $P->reshape(3,4))
      : - S($P->reshape(3,4) x $X);
    return $ret;
  }
  elsif ($diff==2) {
    my $ret = $doublepack ? 
      0.5*append(- S($P->reshape(3,4) x $X),
		 S($x) x $P->reshape(3,4))
	:  S($x) x $P->reshape(3,4);
    return $ret;
  }
  warn "something went wrong calculating Jacobian_IncidentProjectiveCameraPoint2dPoint";
  return undef;
}
push @EXPORT_OK,"Jacobian_IncidentProjectiveCameraPoint2dPoint";
# yuck, but I want to ensure that all combinations are valid
# functions! The yucky thing ist that I have to take care of the
# fourth argument which determines the variable for the Jacobian
sub Jacobian_IncidentPoint2dProjectiveCameraPoint {
  Jacobian_IncidentProjectiveCameraPoint2dPoint(@_[1,0,2],$permute102[$_[3]]);
}
sub Jacobian_IncidentPoint2dPointProjectiveCamera {
#  print "---->",join(",",@_),"\n","new index: ",$permute120[$_[3]],"<------\n";
  Jacobian_IncidentProjectiveCameraPoint2dPoint(@_[2,0,1],$permute120[$_[3]]);
}
sub Jacobian_IncidentPointPoint2dProjectiveCamera {
  Jacobian_IncidentProjectiveCameraPoint2dPoint(@_[2,1,0],$permute210[$_[3]]);
}
sub Jacobian_IncidentPointProjectiveCameraPoint2d {
  Jacobian_IncidentProjectiveCameraPoint2dPoint(@_[1,2,0],$permute201[$_[3]]);
}
sub Jacobian_IncidentProjectiveCameraPointPoint2d {
  Jacobian_IncidentProjectiveCameraPoint2dPoint(@_[0,2,1],$permute021[$_[3]]);
}

# EXPERIMENTAL: 
# - incidence for 2D line, 3D line and projection matrix
sub Jacobian_IncidentProjectiveCameraLine2dLine {
  my ($P,$l,$L,$diff) = @_;

  # maybe the vectors $l and $L will be regarded as one unique
  # vector. Then we return a doubled Jacobian
  my $doublepack = 0;
  if ((!defined($l) || $l->dims == (0)) && defined($L) && join(",",$L->mdims) =~ /9,1/) {
    # correct indexing ?!?!
    $l = $L->slice("6:8")->copy;    $L = $L->slice("0:5");
    $doublepack = 1;
  }
  if ((!defined($L) || $L->dims == (0)) && defined($l) && join(",",$l->mdims) =~ /9,1/) {
    $L = $l->slice("3:8")->copy;    $l = $l->slice("0:2");    
    $doublepack = 1;
  }

  if ($diff==0) {
    return -Gammadual($L) x kroneckerproduct(~$l,SUGR::Basic::Identity(4));
  }
  elsif ($diff==1) {
    my $ret = $doublepack ? 
      0.5*append( -Gammadual($L) x ~$P->reshape(3,4),
		  transpose(Pi($P->reshape(3,4)->transpose x $l))) 
	: -Gammadual($L) x ~$P->reshape(3,4);
    return $ret;
  }
  elsif ($diff==2) {
    my $ret = $doublepack ? 
      0.5*append( -Gammadual($L) x ~$P->reshape(3,4),
		  transpose(Pi($P->reshape(3,4)->transpose x $l))) 
	: transpose(Pi($P->reshape(3,4)->transpose x $l));
    return $ret;
  }
  warn "something went wrong calculating Jacobian_IncidentProjectiveCameraLine2dLine";
  return undef;
}
push @EXPORT_OK,"Jacobian_IncidentProjectiveCameraLine2dLine";
sub Jacobian_IncidentLine2dProjectiveCameraLine {
  Jacobian_IncidentProjectiveCameraLine2dLine(@_[1,0,2],$permute102[$_[3]]);
}
sub Jacobian_IncidentLine2dLineProjectiveCamera {
#  print "---->",join(",",@_),"\n","new index: ",$permute120[$_[3]],"<------\n";
  Jacobian_IncidentProjectiveCameraLine2dLine(@_[2,0,1],$permute120[$_[3]]);
}
sub Jacobian_IncidentLineLine2dProjectiveCamera {
  Jacobian_IncidentProjectiveCameraLine2dLine(@_[2,1,0],$permute210[$_[3]]);
}
sub Jacobian_IncidentLineProjectiveCameraLine2d {
  Jacobian_IncidentProjectiveCameraLine2dLine(@_[1,2,0],$permute201[$_[3]]);
}
sub Jacobian_IncidentProjectiveCameraLineLine2d {
  Jacobian_IncidentProjectiveCameraPoint2dPoint(@_[0,2,1],$permute021[$_[3]]);
}

# EXPERIMENTAL: 
# - Jacobian for EqualTransformationNewPointPoint
sub Jacobian_EqualHomographyPointPoint {
  my ($H,$X,$Y,$diff) = @_;

  # maybe the vectors $X and $Y will be regarded as one unique
  # vector. Then we'll return a doubled Jacobian
  my $doublepack = 0;
  if ((!defined($X) || $X->dims == (0)) && defined($Y) && join(",",$Y->mdims) =~ /8,1/) {
    # correct indexing ?!?!
    $X = $Y->slice("0:3")->copy;    $Y = $Y->slice("4:7");
    $doublepack = 1;
  }
  if ((!defined($Y) || $Y->dims == (0)) && defined($X) && join(",",$X->mdims) =~ /8,1/) {
    $Y = $X->slice("4:7")->copy;    $X = $X->slice("0:3");    
    $doublepack = 1;
  }
#  print "($H,$X,$Y,$diff)\n";

  # ...
  # need to do this for all combinations.
  if ($diff==0) {
    return -Pi($Y) x kroneckerproduct(Identity(4),~$X)
  }
  elsif ($diff==1) {
    my $ret = $doublepack ? 
      0.5*append( -Pi($Y) x $H->reshape(4,4),
		   Pi($H->reshape(4,4) x $X)
		)
	: -Pi($Y) x $H->reshape(4,4) ;
    return $ret;
  }
  elsif ($diff==2) {
    my $ret = $doublepack ? 
      0.5*append( -Pi($Y) x $H->reshape(4,4),
		   Pi($H->reshape(4,4) x $X)
		)
	:  Pi($H->reshape(4,4) x $X);
    return $ret;
  }
  warn "something went wrong calculating Jacobian_EqualHomographyPointPoint";
  return undef;
}
push @EXPORT_OK,"Jacobian_EqualHomographyPointPoint";
# yuck, but I want to ensure that all combinations are valid
# functions! The yucky thing ist that I have to take care of the
# fourth argument which determines the variable for the Jacobian
sub Jacobian_EqualPointHomographyPoint {
  Jacobian_EqualHomographyPointPoint(@_[1,0,2],$permute102[$_[3]]);
}
sub Jacobian_EqualPointPointHomography {
#  print "---->",join(",",@_),"\n","new index: ",$permute120[$_[3]],"<------\n";
  Jacobian_EqualHomographyPointPoint(@_[2,0,1],$permute120[$_[3]]);
}
# sub Jacobian_EqualPointPointHomography {
#   Jacobian_EqualHomographyPointPoint(@_[2,1,0],$permute210[$_[3]]);
# }
# sub Jacobian_EqualPointHomographyPoint {
#   Jacobian_EqualHomographyPointPoint(@_[1,2,0],$permute201[$_[3]]);
# }
# sub Jacobian_EqualHomographyPointPoint {
#   Jacobian_EqualTransformationNewPointPoint(@_[0,2,1],$permute021[$_[3]]);
# }


################################################################################
#
# Orthogonality
#
sub Jacobian_OrthogonalLine2dLine2d {
  my ($l,$m,$diff) = @_;
  my $linediff = ($diff==0) ? $m : $l;
  $linediff = (ref($linediff) =~ /SUGR/) ? $linediff->vec->copy : $linediff->copy;
  $linediff->set(2,0);
  return ~$linediff;
}

sub Jacobian_OrthogonalLineLine {
  my ($L,$M,$diff) = @_;
  my $Linediff = ($diff==0) ? $M : $L;
  $Linediff = (ref($Linediff) =~ /SUGR/) ? $Linediff->vec->copy : $Linediff->copy;
  ($_ = $Linediff->slice("3:5")) .= 0;
  return ~$Linediff;
}

sub Jacobian_OrthogonalPlanePlane {
  my ($A,$B,$diff) = @_;
  my $Planediff = ($diff==0) ? $B : $A;
  $Planediff = (ref($Planediff) =~ /SUGR/) ? $Planediff->vec->copy : $Planediff->copy;
  ($_ = $Planediff->slice("3:3")) .= 0;
  return ~$Planediff;
}

sub Jacobian_OrthogonalPlaneLine {
  Jacobian_OrthogonalLinePlane(@_[1,0],$not[$_[2]])
}

sub Jacobian_OrthogonalLinePlane {
  my ($L,$A,$diff) = @_;
  my $Diff = ($diff==0) ? $A : $L;
  $Diff = (ref($Diff) =~ /SUGR/) ? $Diff->vec : $Diff;
  my $neg = ($diff==1) ? 1 : -1;
  my $S = $neg*S($Diff->slice("0:2"));
  my $J = ($diff==0) ? mzeroes(6,6) : mzeroes(4,4);
  ($_ = $J->slice("0:2,0:2")) .= $S;
  return $J;
}

################################################################################
#
# Parallelity
#
sub Jacobian_ParallelLine2dLine2d {
  my ($l,$m,$diff) = @_;
  my $neg = ($diff==1) ? 1 : -1;
  my $Diff = ($diff==0) ? $m : $l;
  $Diff = (ref($Diff) =~ /SUGR/) ? $Diff->vec : $Diff;
  my $J = vzeroes(3);
  $J->set(0, $Diff->at(1));
  $J->set(1,-$Diff->at(0)); 
  return ~($neg * $J);
}

sub Jacobian_ParallelLineLine {
  my ($L,$M,$diff) = @_;
  my $Diff = ($diff==0) ? $M : $L;
  $Diff = (ref($Diff) =~ /SUGR/) ? $Diff->vec : $Diff;
  my $neg = ($diff==1) ? 1 : -1;
  my $J = mzeroes(6,6);
  ($_ = $J->slice("0:2,0:2")) .= SUGR::Basic::S( $Diff->slice("0:2") );
  return $neg*$J;
}

sub Jacobian_ParallelPlanePlane {
  my ($A,$B,$diff) = @_;
  my $Diff = ($diff==0) ? $B : $A;
  $Diff = (ref($Diff) =~ /SUGR/) ? $Diff->vec : $Diff;
  my $neg = ($diff==1) ? 1 : -1;
  my $J = mzeroes(4,4);
  ($_ = $J->slice("0:2,0:2")) .= $neg*S($Diff->slice("0:2"));
  return $J;
}

sub Jacobian_ParallelPlaneLine {
  my ($A,$L,$diff) = @_;
  my $Diff = ($diff==0) ? $L : $A;
  $Diff = (ref($Diff) =~ /SUGR/) ? $Diff->vec : $Diff;
  $Diff = $Diff->slice("0:2");
  my $J = ($diff==0) ? vzeroes(4) : vzeroes(6);
  ($_ = $J->slice("0:2")) .= $Diff;
  return ~$J;
}

sub Jacobian_ParallelLinePlane {
  Jacobian_ParallelPlaneLine(@_[1,0],$not[$_[2]])
}



##################################################
#
# function for relations build by their Jacobians
#
# now we can e.g. call func_Incident($X,$L) to obtain $D
foreach my $func qw(Incident Equal Orthogonal Parallel) {
  my $func_func ="func_$func";
  no strict "refs";
  *$func_func =  sub  {
    my ($E1,$E2) = @_;
    my ($s1,$s2) = ( ref($E1), ref($E2));
    $s1 =~ s/^.*:://;
    $s2 =~ s/^.*:://;
    my $code = "func_$func$s1$s2";
    no strict "refs";
    return &$code($E1,$E2,0);
  };
  push @EXPORT_OK,"func_$func";
}

foreach my $RelationType (keys %possible_test_types) {
  foreach my $objects (@{$possible_test_types{$RelationType}}) {
    my @type_object_combination;
    push @type_object_combination, $RelationType.join("",@$objects);
    push @type_object_combination, $RelationType.join("",@$objects[1,0]) if $objects->[0] ne $objects->[1];

    foreach my $func (@type_object_combination) {
      my $func_func ="func_$func";
      no strict "refs";
      *$func_func =  sub  {
	my ($E1,$E2) = @_;
	my $v = (ref($E2) =~ /SUGR/) ? $E2->vec : $E2;
	my $code = "Jacobian_$func";
	no strict "refs";
	return &$code($E1,$E2,0) x $v; # here we actually compute the
	# function value
      };
      #  push @EXPORT_OK,"func_$func";
    }
  }
}

%EXPORT_TAGS = (ALL => [@EXPORT,@EXPORT_OK]);



return 1 if caller;

################################################################################
#
# Test code
#

{

require SUGR::Point2d;
require SUGR::Point;
require SUGR::Line2d;
require SUGR::Line;
require PDL::Matrix;

require SUGR::ProjectiveCamera;


############################################################
#
# Test Jacobians for Collinearity Equations
#

my $P = new SUGR::ProjectiveCamera(2+mpdl(random(4,3)), #
				      mones(12,12));

my $X = new SUGR::Point(vpdl(random(4)),mones(4,4));
my $x = new SUGR::Point2d(vpdl(random(3)),mones(3,3));
#my $x = new SUGR::Point2d($P,$X);

my $L = new SUGR::Line(vsequence(6),mones(6,6));
my $l = new SUGR::Line2d([1,2,1],mones(3,3));

print "Jacobian for collinearity equations:";
my @obj_points  = ( $P->vec,$x->vec,$X->vec );
my @obj_points_tmp  = ( $P->vec,~append(~$x->vec,~$X->vec),undef);
foreach my $i (0..2) {
  print Jacobian_IncidentProjectiveCameraPoint2dPoint(@obj_points,$i) x $obj_points[$i];
  print Jacobian_IncidentProjectiveCameraPoint2dPoint(@obj_points_tmp,$i) x $obj_points_tmp[$i] if $i<2;
}

print "Jacobian for coplanarity equations:";
my @obj_lines  = ( $P->vec,$l->vec,$L->vec );
my @obj_lines_tmp  = ( $P->vec,~append(~$l->vec,~$L->vec),undef);
foreach my $i (0..2) {
  print Jacobian_IncidentProjectiveCameraLine2dLine(@obj_lines,$i)     x $obj_lines[$i];
  print Jacobian_IncidentProjectiveCameraLine2dLine(@obj_lines_tmp,$i) x $obj_lines_tmp[$i] if $i<2;
}

############################################################
#
# Test Jacobians for Homography Equations
#

require SUGR::Homography;

my $H = new SUGR::Homography(msequence(4,4)+1,0.01*mones(16,16));
my $Y = $H->transform($X); #new SUGR::Point(vpdl(random(4)),mones(4,4));
@obj_points  = ( $H->vec,$X->vec,$Y->vec );
@obj_points_tmp  = ( $H->vec,~append(~$X->vec,~$Y->vec),undef);
print "Jacobian for homography equations (must be 5 identical vectors):\n";
foreach my $i (0..2) {
  print "$i\n";
  print Jacobian_EqualHomographyPointPoint(@obj_points,$i) x $obj_points[$i];
  print Jacobian_EqualHomographyPointPoint(@obj_points_tmp,$i) x $obj_points_tmp[$i] if $i<2;
}

# apply_constraints(0);
# my $x_empty = new SUGR::Point2d(vsequence(3),Identity(3));
# my $l_empty = new SUGR::Line2d (vsequence(3),Identity(3));

# my $x_euclid = new SUGR::Point2d(vsequence(3),Identity(3))->euclidize;
# my $l_euclid = new SUGR::Line2d (vsequence(3),Identity(3))->euclidize;

# apply_constraints(1);
# my $x_sphere = new SUGR::Point2d(vsequence(3),Identity(3));
# my $l_sphere = new SUGR::Line2d (vsequence(3),Identity(3));

# print "nullspace = euclidean : ",Incident($x_euclid,$l_euclid),"\n";
# print "nullspace = sphere    : ",Incident($x_sphere,$l_sphere),"\n";
# print "nullspace = empty     : ",Incident($x_empty,$l_empty),"\n";

############################################################
#
# test all known 3D relations using random cuboid
#
use SUGR::Cuboid;

print "\nTest all known 3D relations\n";

my %Objects = generate_cuboid(no_points => 10,
			      no_planes => 10,
			      no_lines  => 10,
			      random_seed => [4711,222712]);
%Objects = transform_cuboid(translation => [1,2,2],
 			    rotation    => [0.5,0.7,1.3],
 			    %Objects);

my $X_cube  = $Objects{Points};
my $L_cube  = $Objects{Lines};
my $A_cube  = $Objects{Planes};
my $X_cube_ideal  = $Objects{Ideal_Points};
my $L_cube_ideal  = $Objects{Ideal_Lines};
my $A_cube_ideal  = $Objects{Ideal_Planes};

#######




print "Projecting noisy cube points ...\n";
use SUGR::ProjectiveCamera;
my $X0 = new SUGR::Point([1.5,1.5,10,1],0.001*mones(4,4));
my $R  = mpdl([[1,0,0],
	       [0,1,0],
	       [0,0,1]]);
$P = sugr_ProjectiveCamera({ 
#				    "X0"    => vpdl([1.5,1.5,10]), 
				    "X0"    =>$X0,
				    "R"     => $R,
				    "c"     => 1.0,
				    "c_var" => 0.001,
				   });
my (@xs_cube,@ls_cube);
foreach my $corner (0..@$X_cube) {
  foreach my $X_cubepoint (@{$X_cube->[$corner]})  {
    push @{$xs_cube[$corner]}, new SUGR::Point2d($X_cubepoint,$P);
  }
  foreach my $L_cubeline (@{$L_cube->[$corner]})  {
    push @{$ls_cube[$corner]}, new SUGR::Line2d($L_cubeline,$P);
  }
}


my %Testobjects = ( 
     		   "EqualPointPoint"    => [$X_cube_ideal->[0],$X_cube->[0]],
    		   "EqualLineLine"      => [$L_cube_ideal->[0],$L_cube->[0]],
    		   "EqualPlanePlane"    => [$A_cube_ideal->[0],$A_cube->[0]],
    		   "IncidentPointLine"  => [$X_cube_ideal->[0],$L_cube->[0]],
    		   "IncidentPointPlane" => [$X_cube_ideal->[0],$A_cube->[0]],
    		   "IncidentLineLine"   => [$L_cube_ideal->[0],$L_cube->[1]],
    		   "IncidentLinePlane"  => [$L_cube_ideal->[0],$A_cube->[0]],
    		   "OrthogonalLineLine"   => [$L_cube_ideal->[0],$L_cube->[1]],
    		   "OrthogonalLinePlane"  => [$L_cube_ideal->[0],$A_cube->[2]],
    		   "OrthogonalPlanePlane" => [$A_cube_ideal->[0],$A_cube->[1]],
    		   "ParallelLineLine"     => [$L_cube_ideal->[0],$L_cube->[2]],
    		   "ParallelLinePlane"    => [$L_cube_ideal->[0],$A_cube->[3]],
    		   "ParallelPlanePlane"   => [$A_cube_ideal->[0],$A_cube->[5]],

 		   "IncidentProjectiveCameraPoint2dPoint"  => [[$P],$xs_cube[0],$X_cube_ideal->[0]],
 		   "IncidentProjectiveCameraLine2dLine"  => [[$P],$ls_cube[0],$L_cube_ideal->[0]],
    	   
  		   "NotIncidentPointLine"  => [$X_cube_ideal->[0],$L_cube->[6]],
  		   "NotIncidentLineLine"   => [$L_cube_ideal->[0],$L_cube->[6]],
 		   "NotEqualPointPoint"    => [$X_cube_ideal->[0],$X_cube->[2]],
 		   "NotEqualLineLine"      => [$L_cube_ideal->[0],$L_cube->[2]],
 		   "NotEqualPlanePlane"    => [$A_cube_ideal->[0],$A_cube->[3]],
 		   "NotIncidentProjectiveCameraPoint2dPoint"  => [[$P],$xs_cube[2],$X_cube_ideal->[0]],
 		   "NotIncidentProjectiveCameraLine2dLine"  => [[$P],$ls_cube[3],$L_cube_ideal->[0]],

		  );


foreach my $Test (keys %Testobjects) {
  print "$Test\n",(map { "-" } (1..length($Test))),"\n";

  my @O1 = @{$Testobjects{$Test}->[0]};
  my @O2 = @{$Testobjects{$Test}->[1]};
  my @O3 = defined($Testobjects{$Test}->[2]) ? @{$Testobjects{$Test}->[2]} : (undef) ;

  # do we expect the test to be accepted?
  my $wrong = ($Test =~ s/^Not//) ? 1 : 0;

  foreach my $O1 (@O1) {
    foreach my $O2 (@O2) {
      foreach my $O3 (@O3) {
	no strict "refs";
      
	my %Result = &$Test($O1,$O2,$O3,0.99);
	die "Test was rejected @{[ %Result ]}" if ((!$Result{bool} && !$wrong) ||
						   ($Result{bool} && $wrong));
	print "Test successful with testvalue $Result{testvalue}\n";
      }
    }
  }
  print "\n";
}
}
