=head1 NAME

SUGR::Scene -- A class for a set of SUGR objects

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Scene->VERSION

=head1 SYNOPSIS

  use SUGR::Scene;

  # create a new and empty scene, i.e. an 
  # emtpy bag of objects
  $Scene = SUGR::Scene->new;

  # create some SUGR objects;
  # [... create $L1 and $L2 ...]
  $X = new SUGR::Point($L1,$L2);

  # add these objects to the scene
  $Scene->add($X,$L1,$L2);

  # return the points in the scene
  @X = $Scene->Points;

=head1 DESCRIPTION

This module provides a class for a scene that is described by SUGR
elements. A possible scene could be a set of cameras with 2D points
and lines observed in the camera image and 3D lines, points and planes
referring to a polyhedral object.

An instance of this class is nothing but a bag of SUGR objects, that
can be accessed. You can either initialize this bag by your program or
you might want to use L<write_gxf> for storing the scene in a file,
see also <SUGR::Scene::GXF(3)|SUGR::Scene::GXF> for more details.

=head1 METHODS

=head2 Constructors and Initialization

=over 4

=cut

package SUGR::Scene;


our $VERSION = SUGR->VERSION;

# this will be used for an empty scene
my %empty_scene = (
		   "_Points2d" => [],
		   "_Lines2d" => [],
		   "_Points" => [],
		   "_Lines" => [],
		   "_Planes" => [],
		   "_Cameras" => [],
		   "_Homography" => [],
		   "_Homography2d" => [],
		   "_EqualRelations" => [],
		   "_IncidentRelations" => [],
		   "_ParallelRelations" => [],
		   "_OrthogonalRelations" => []
		  );

my @object_types = map { s/^_//; $_ } (keys %empty_scene);

# this is to translate a class name to the internal slot of the object
my %internal_slot = (
	    "SUGR::Point2d" => "_Points2d",
	    "SUGR::Line2d" =>"_Lines2d" ,
	    "SUGR::Point" => "_Points" ,
	    "SUGR::Line" => "_Lines" ,
	    "SUGR::Plane" => "_Planes" ,
	    "SUGR::ProjectiveCamera" => "_Cameras" ,
	    "SUGR::Homography" => "_Homography" ,
	    "SUGR::Homography2d" => "_Homography2d",
	    "SUGR::Relation::Equal" => "_EqualRelations",
	    "SUGR::Relation::Incident" => "_IncidentRelations",
	    "SUGR::Relation::Parallel" => "_ParallelRelations", 
	    "SUGR::Relation::Orthogonal" => "_OrthogonalRelations",
	   );

# counter for object ids
our $scene_cnt=1;

=item new(@objects)

The default constructor takes a list @objects of SUGR(3) objects as
input, see L<SUGR(3)|SUGR> for details. The input list may be empty. If it
is not empty, the above call is equivalent to;
  
  $Scene = SUGR::Scene->new;
  $Scene->add(@objects);

=cut


sub new {
  my ($protoverse, @objects) = @_;
  my $class = ref($protoverse) || $protoverse;

  my $self = { };
  $self->{$_} = [] foreach (keys %empty_hash);

  bless $self, $class;

  $self->{id} = $class."_".($scene_cnt++);
  $self->{type} = "Scene";
  $self->{_hash_of_ids} = {};
  
  $self->add(@objects);
  
  return $self;
}

=item add(@objects)

Adds a list of objects to the scene.

Example:

  push @objects, new SUGR::Point([1,2,3,1]);
  $Scene->add(@objects);

=cut

# distribute given list of objects to the internal slots
sub add {
  my ($self,@objects) = @_;

  foreach my $object (@objects) {
    push @{$self->{$internal_slot{ref($object)}}}, $object;
    $self->{_hash_of_ids}->{$object->id} = $object;
  }
}

=pod

=back

=head2 Class and object methods

=over 4

=item id()

Returns or sets the unique identifier of the whole scene. The L<new>
constructor will automatically assign a default id. This method takes
no arguments.

Example:

  $Scene = SUGR::Scene->new;
  $Scene->id = "My_Scene";
  print $Scene->id;

=cut 

sub id :lvalue {
  $_[0]->{id}
}

sub comment  {
  $_[0]->{comment} = $_[1] if $_[1];
  $_[0]->{comment};
}

sub type :lvalue {
  $_[0]->{type}
}

sub filename  {
  defined($self->{_filename}) ? $self->{_filename} : "";
}

sub output_filename  {
  my ($self,$name) = @_;
  $self->{_outputfilename} = $name if $name;
  defined($self->{_outputfilename}) ? $self->{_outputfilename} : "";
}

=item Points2d()
  Lines2d()
  Points()
  Lines()
  Planes()
  Cameras()
  Homography()
  Homography2d()
  EqualRelations()
  IncidentRelations()
  ParallelRelations()
  OrthogonalRelations()

These methods give access to the list of objects and entities present
in the actual scene. 

=cut

# define access methods for points, lines, etc.
foreach my $slot (@object_types) {
  my $lcslot = lc($slot);
  *$lcslot = *$slot = sub { 
    my ($self,$index) = @_; 
   
    return $self->{"_$slot"}->[$index] if defined($index);
    return @{$self->{"_$slot"}};
  }
}

=item object_id()

returns a list of all object_ids

=cut

sub object_id {
  my ($self,@ids) = @_;
  return keys %{$self->{_hash_of_ids}} unless @ids;
  my @existing_ids = grep { defined($self->{_hash_of_ids}->{$_}) } @ids;
  return @existing_ids ? @existing_ids : undef if wantarray;
  return @existing_ids ? $existing_ids[0] : undef;
}

=item object(@ids) , obj(@ids)

returns a list of objects from a list of object ids @ids. If @ids is
empty, all objects are returned.

obj(@ids) is an alias for object(@ids)

=cut

sub object {
  my ($self,@ids) = @_;
  my @existing_ids =  $self->object_id(@ids);
  @objects = map { $self->{_hash_of_ids}->{$_} } @existing_ids;
  return wantarray ? @objects : $objects[0];
}

# aliases
sub obj { object(@_); }
sub Object { Object(@_); }

=item object_types()

return the object types which have at least one element in the
scene. These object types can be used for accessing the list of
objects.

Example:

   foreach my $existing_object_type ($Scene->object_types) {
      # print out list of objects that exists in the scene
      print $Scene->$existing_object_type;
   }

=cut

sub object_types {
  my ($self) = @_;

  grep { scalar($self->$_) > 0 } (@object_types);
}

=item write_gxf($filename)

Writes the scene in a specialized format called GXF, see
<SUGR::Scene::GXF(3)|SUGR::Scene::GXF> for details. To load such a
scene, try

  $Scene = new SUGR::Scene::GXF($filename);

=cut

sub write_gxf {
  my ($self,$filename) = @_; 

  # take the given filename. If this is not present, then take the one
  # defined in the object itself.
  unless ($filename) {
    if ($self->output_filename) {
      $filename = $self->output_filename;
    }
    elsif ($self->filename) {
      ($filename = $self->filename) =~ s/\.gxf$/-sweeted.gxf/;
    }
    else {
      die "*** Error: filename not specified.\n" unless defined($filename);
    }
  } 
  

  print "Writing $filename ... ";
  open OUT,">$filename" or die "*** Error: cannot open $filename: $!\n";
  print OUT ";; generated by SUGR::Scene on ",scalar(localtime()),"\n;;\n";

  if ($self->filename) {
    print OUT "BLOCKSTART SCENEINFO\n";
    print OUT "\tTYPE SUGR\n";
    print OUT "\tCOMMENT \"".$self->comment."\"\n" if $self->comment;
    print OUT "\tGENERATOR \"SUGR::Scene::write_gxf\"\n";
    print OUT "\tORIGINAL_FILENAME \"",$self->filename,"\"\n" if $self->filename;
    print OUT "BLOCKEND SCENEINFO\n\n";
  }
  print OUT "BLOCKSTART OBJECTS\n";
  my @object_types = $self->object_types;
  foreach my $object_type (@object_types) {
    next if $object_type !~ /(POINTS|LINES|PLANES|POINTS_2D|LINES_2D)/i;
    my $elementlist_id = uc($object_type);
    print OUT "\tELEMENTLIST $elementlist_id ",scalar($self->$object_type),"\n";
    foreach my $object ($self->$object_type) {
      print OUT "\tELEMENT ",$object->id,"\n";
      print OUT "\t\tTYPE HOMOGENEOUS\n";
      print OUT "\t\tLIST_OF COORDS ",$dim{uc($object_type)},"\n";
      print OUT "\t\tCOORDS $_\n" foreach ($object->vec->list);
      if (defined($object->cov)) {
	print OUT "\t\tLIST_OF COVARIANCES ",$dim{uc($object_type)},"\n";
	foreach ($object->cov->transpose->dog) {
	  print OUT "\t\tCOVARIANCES ",join(" ",($_->list)),"\n";
	}
      }
      print OUT "\tEND_ELEMENT\n";
    }
    print OUT "\n";
  }
  print OUT "BLOCKEND OBJECTS\n\n\n";

  print OUT "BLOCKSTART TRANSFORMATIONS\n";
  foreach my $object_type (@object_types) {
    next unless $object_type =~ /(CAMERA|HOMOGRAPHIES|HOMOGRAPHIES_2D)/i;
    my $elementlist_id = uc($object_type);
    print OUT "\tELEMENTLIST $elementlist_id ",scalar($self->$object_type),"\n";
    foreach my $object ($self->$object_type) {
      print OUT "\tELEMENT ",$object->id,"\n";
      print OUT "\t\tTYPE HOMOGENEOUS\n";
      my $matrix = $object->matrix("points");
      my ($rowdim) = $matrix->mdims;
      print OUT "\t\tLIST_OF POINTTRANSFORMATION $rowdim\n";
      foreach ($matrix->dog) {
	print OUT "\t\tPOINTTRANSFORMATION ",join(" ",($_->list)),"\n";
      }
      if (defined($object->cov)) {
	print OUT "\t\tLIST_OF COVARIANCES ",$dim{uc($object_type)},"\n";
	foreach ($object->cov->transpose->dog) {
	  print OUT "\t\tCOVARIANCES ",join(" ",($_->list)),"\n";
	}
      }
      print OUT "\tEND_ELEMENT\n";
    }
    print OUT "\n";
  }
  print OUT "BLOCKEND TRANSFORMATIONS\n\n\n";

  print OUT "BLOCKSTART RELATIONS\n";
  foreach my $object_type (@object_types) {
    next if $object_type !~ /(EQUAL|INCIDENT|PARALLEL|ORTHOGONAL)/i;
    (my $elementlist_id = uc($object_type)) =~ s/RELATIONS//i;
    print OUT "\tELEMENTLIST $elementlist_id ",scalar($self->$object_type),"\n";
    foreach my $object ($self->$object_type) {
      print OUT "\tELEMENT $id\n";
      print OUT "\t\tPAIR ".join(" ",@{$object->{pair_ids}})."\n";
      print OUT "\t\tTEST ".($object->bool ? "TRUE" : "FALSE")."\n";
      print OUT "\t\tTESTVALUE ".$object->testvalue."\n";
      print OUT "\t\tTESTRATIO ".$object->testratio."\n";
      print OUT "\tEND_ELEMENT\n";
    }
    print OUT "\n";
  }
  print OUT "BLOCKEND RELATIONS\n";

  print OUT "END_OF_FILE\n";
  close OUT;


  print "done.\n";

  return 1;
}

use overload q{""} => \&stringify;

sub stringify { 
  my ($self) = shift;
  print "Scene: ",$self->id,"\n";

  foreach my $object_type ($self->object_types) {
    my $no_elements = scalar($self->$object_type);
    print " - $object_type: $no_elements element(s)\n";
  }
  print " - List of Object-IDs:\n";
  print "    $_\n" foreach (sort $self->object_id);
}

return 1 if caller;

=pod

=back

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

L<SUGR::Scene::GXF>

=head1 AUTHOR

Stephan Heuel (stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

# test code

print "Version: ",SUGR::Scene->VERSION,"\n";

$X = new SUGR::Point([1,2,3,4]);
$S = new SUGR::Scene($X);

print "class: ",ref($S),"\n";


$n = $S->Points;

print $S;

print $S->Points($n-1);

foreach (qw/SUGR::Point_1 Point_1/) {
  print "$_ ";
  print $S->object_id($_) ? "exists\n" : "does not exist\n";
}


print "\nList of objects\n";
print join(",",$S->object);

print "\nList of objects\n";
print $S->object_types;


