=head1 NAME

SUGR::Scene::GXF -- class for creating scenes with GXF(sugr), the
generic exchange format for SUGR elements

=head1 VERSION

This document refers to version 0.1.0, the actual version can be
checked with SUGR::Scene::GXF->VERSION

=head1 SYNOPSIS

  use SUGR::Scene::GXF;

  # parse a GXF(sugr) file and perform possible constructions
  $Objects = new SUGR::Scene::GXF("/path/to/my-scene.gxf");

  # get the list of points
  @X = $Objects->points;

  # compute a new 3D line from the first two points
  $L = new SUGR::Line($X[0],$X[1]);

  # add the previously computed line to the scene
  $Objects->add($L);

  # write the new scene to disk
  $Objects->write("/path/to/my-new-scene-new.gxf");

=head1 OVERVIEW

This module provides read and write methods for accessing the
L<SUGR|SUGR> library without actually using Perl. See also the program
L<sugr(1)|sugr> which reads a GXF(sugr) file and computes new elements
resp. tests relations between them and produce another GXF(sugr) file
as output.

=head1 GXF - Generic Exchange Format

=head2 History

The GXF was primarily designed and used at the Institute of
Photogrammetry, Bonn for specifying the output file of the feature
extraction system FEX, see
http://www.ipb.uni-bonn.de/ipb/projects/fex/fex.html for details. It
uses a syntax similar to mark-up languages such as XML. 

We denote the specialization of GXF to SUGR with GXF(sugr) as opposed
to GXF(fex), which refers to the specialization of GXF to the FEX
program.

=head2 What kind of data can GXF(sugr) describe?

The entities that can be described are 

=over 4

=item -

points and (infinite) lines in 2D

=item -

points, lines and planes in 3D

=item -

projective transformations in 2D and 3D and 3D->2D

=back

These entities can be specified by a parameter vector and possibly
(and advisably) a covariance matrix for this parameter vector. The
covariance matrix refers to the parameter vector of the element.

=head2 What operations can I perform using GXF(sugr)?

Once a SUGR dataset is specified you can construct new entities from
the given data, either directly by intersection or join:

=over 4

=item - 

2D point 

intersection of two 2D lines

=item - 

2D point 

projection of 3D point using a pin-hole camera

=item -

2D line 

join of two 2D points

=item - 

2D line 

projection of 3D line using a pin-hole camera

=item -

3D point 

intersection of three 3D planes

=item -

3D point

intersection of a 3D plane and a 3D line

=item - 

3D line

join of two 3D points

=item -

3D line  

intersection of two 3D planes

=item -

3D plane 

join of three 3D points 

=item -

3D plane 

join of a 3D point and a 3D line

=back

Additionally one may estimate an unknown entity based on observed
entities. An example is the intersection of many lines in one point,
as it happens when reconstructing an unknown 3D point, which is
observed by two or more cameras.

Furthermore you can test relations between given entities. The
currently supported relations are 

=over 4

=item -

identity

=item -

incidence

=item -

parallelity

=item -

orthogonality

=back

=head1 METHODS

Please check the inherited methods from L<SUGR::Scene|SUGR::Scene>.

=over 4

=cut

package SUGR::Scene::GXF;

use base qw/SUGR::Scene/;

use SUGR;
use SUGR::Relation::Equal;
use SUGR::Relation::Incident;
use SUGR::Relation::Parallel;
use SUGR::Relation::Orthogonal;
use SUGR::Scene;
use PDL;

our $VERSION = SUGR->VERSION;

my %class = ( POINTS_2D => "SUGR::Point2d",
	      LINES_2D => "SUGR::Line2d",
	      POINTS => "SUGR::Point",
	      LINES => "SUGR::Line",
	      PLANES => "SUGR::Plane",
	      CAMERAS => "SUGR::ProjectiveCamera",
	      HOMOGRAPHIES => "SUGR::Homography",
	      HOMOGRAPHIES_2D => "SUGR::Homography2D",
	    );

my %dim = ( POINTS_2D => 3,
	    LINES_2D => 3,
	    POINTS => 4,
	    LINES => 6,
	    PLANES => 4,
	    CAMERAS => 12
	  );

=pod

=item 

new

You may construct a scene by reading the objects specified in a
GXF(sugr) file.

  $Scene = new SUGR::Scene::GXF("/path/to/my.gxf")

=cut

sub new {
  my ($protoverse, $filename_or_handle) = @_;
  my $class = ref($protoverse) || $protoverse;

  $self = SUGR::Scene->SUPER::new;
  bless $self, $class;
  
  my ($fh,$filename);
  if (fileno($filename_or_handle)) {
    print "Reading from filehandle  ... ";
    $fh = $filename_or_handle;
    $self->{_filename} = "FILEHANDLE";
    $self->comment("read by Perl filehandle");
  }
  else {
    $filename = $filename_or_handle;
    print "Reading $filename ... ";
    $self->{_filename} = $filename;
    
    open $fh,"$filename" or die "*** Error: cannot open $filename: $!\n";
    $self->comment("read from $filename");
  }
  

  my $current_block = "";
  # this is the main reading block
  while (<$fh>) {
    next if /^\s*;/; # skip comments
    
    $current_block = $1    if (/BLOCKSTART\s+(\w+)/);
    $current_block = "" if (/BLOCKEND\s+(\w+)/);
    
    # right now we only care about elementlists and treat all the
    # same, so we actually ignore the BLOCKs right now
    $self->elementlist($fh,$current_block,$1,$2) if (/ELEMENTLIST (\w+) (\d+)/);
   
    if ($current_block =~ /CONTROLPARAMETERS/) {
      $self->block_control($fh,$current_block);
      $current_block = "";
      next;
    }
    
  }
  close $fh;
  print "done.\n";

  return $self;
}

sub write {
  my $self = shift;
  $self->write_gxf(@_);
}


################################################################################
#
# funcs for parsing a GXF file
#

sub elementlist {
  my ($self,@arg) = @_; # save arguments

  my $current_block = $arg[1];
  return $self->elementlist_objects(@arg) if $current_block =~ /(OBJECTS|TRANSFORMATIONS)/;
  return $self->elementlist_constructions(@arg) if $current_block =~ /CONSTRUCTIONS/;
  return $self->elementlist_relations(@arg) if $current_block =~ /RELATIONS/;
}

sub block_control {
  my ($self,$fh,$current_block) = @_;
  
  while (<$fh>) {
    next if /^\s*;/; # skip comments

    $self->comment($1) if /COMMENT \"([^\"]+)\"/;

    if (/TYPE\s+(\w+)/) {
      die "*** Error: this GXF file is not a GXF(sugr) file\n" if $1 ne "SUGR";
    }
    $self->output_filename($1) if /FILE\s+\"([^\"]+)\"/;
    last if /BLOCKEND/;
  }
}

sub elementlist_constructions {
  my ($self,$fh,$current_block,$element_type,$no_of_elements) = @_;

  my $cnt = 0;
  my $id = "";
  my %construction = ();
  while (<$fh>) {
    next if /^\s*;/; # skip comments

    if (/ELEMENT\s+(\w+)/) {
      $id = $1;
      
      die "*** Parse error: Multiple object id: $id\n" if $self->object_id($id);
      next;
    }

    if (/LIST_OF (\w+)\s+(\d+)/) {
      die "*** Parse Error no id is known: $_\n" if $id eq "";
      my $construction_type = $1;
      my $objects_no = $2;
      foreach my $i (0..$objects_no-1) {
	my $line = <$fh>;
	if ( $line =~ /^\s+$construction_type\s+(\w+)/ ) {
	  push @{$construction{$construction_type}}, $1;
	}
	else { die "*** Parse error at line $line, expected LIST_OF element\n"; }
      }
      next;
    }
    
    if (/END_ELEMENT/) {
      die "*** Error while parsing $_\n" if scalar(keys %construction) == 0;
      
      if (defined($construction{DIRECT})) {
	my @objects = $self->object(@{$construction{DIRECT}});
	my $O = eval "new $class{$element_type}(\@objects)";
	die $@ if $@;
	$O->id = $id;
	$self->add($O);
      }
      else {
#	print "Constructing $id with ";
	my %estimation_hash =  ();
	foreach my $key (keys %construction) {
	  my $relation = ucfirst(lc($key));
	  foreach my $object_id (@{$construction{$key}}) {
	    die "*** Error: $object_id is not known: $_\n" unless defined($self->object_id($object_id));
	    my $object = $self->object($object_id);
	    my $estimation_key = "$relation".$object->type;
#	    print "$estimation_key => ",$object,"\n";
	    push @{$estimation_hash{$estimation_key}}, $object;
	  }
	}
	die "*** Parse Error: no observations given for estimation: $_\n" 
	  if scalar(keys %estimation_hash) == 0;
	my $O =  eval "new $class{$element_type}({\%estimation_hash})";
	die $@ if $@;
	$O->id = $id;
	$self->add($O);
	die "*** Error: problems when estimating \"$id\": $@" if $@;
#	print "Estimated : $self->{$id}\n";
      }

      %construction = ();
      $id = "";
      $cnt++;
    }
    last if $cnt == $no_of_elements;
  }
}

sub elementlist_objects {
  my ($self,$fh,$current_block,$element_type,$no_of_elements) = @_;

  if (!defined($class{$element_type})) {
    warn "*** Warning: elementtype $element_type is not known\n";
    return 0;
  }
  my $cnt = 0;
  my ($id,$vec,$cov) = (undef,undef,undef);
  while (<$fh>) {
    next if /^\s*;/; # skip comments

    if (/ELEMENT\s+(\w+)/) {
      $id = $1;
      die "*** Parse error: Multiple object id: $id\n" if defined($self->object_id($id));

      $vec = vones($dim{$element_type}); 
      $cov = mzeroes($dim{$element_type},$dim{$element_type});
      next;
    }

    if (/LIST_OF COORDS\s+(\d+)/) {
      my $coords_no = $1;
      foreach my $i (0..$coords_no-1) {
	my $line = <$fh>;
	if ( $line =~ /^\s+COORDS\s+([+\-\d.eE]+)/ ) { $vec->set($i,$1) }
	else { die "*** Parse error at line $line, expected LIST_OF element\n"; }
      }
      next;
    }

    if (/LIST_OF POINTTRANSFORMATION\s+(\d+)/) {
      my $matrix_no = $1;
      my @numbers = ();
      foreach my $i (0..$matrix_no-1) {
	my $line = <$fh>;
	if ($line =~ /^\s+POINTTRANSFORMATION\s+([+\-\d.eE\s]+)$/) {
	  push @numbers , split(" ",$1);
	}
	else { die "*** Parse error at line $line, expected LIST_OF element\n"; }
      }
      ($_ = $vec->slice("0:$#numbers")) .= vpdl [ @numbers ];
      next;
    }

    if (/LIST_OF COVARIANCES\s+(\d+)/) {
      my $cov_no = $1;
      foreach my $i (0..$cov_no-1) {
	my $line = <$fh>;
	if ($line =~ /^\s+COVARIANCES\s+([+\-\d.eE\s]+)$/) {
	  my @numbers = split(" ",$1);
	  ($_ = $cov->slice("($i),0:$#numbers")) .= pdl [ @numbers ];
	}
	else { die "*** Parse error at line $line, expected LIST_OF element\n"; }
      }
      next;
    }

    if (/END_ELEMENT/) {
      die "*** Error while parsing $_\n" if grep { !defined } ($id,$vec,$cov);
      # make the covariance matrix undef if it's equal to zero.
      $cov = undef if $cov->abs->sum == 0;
      # redundant representation
      my $O =  eval "new $class{$element_type}(\$vec,\$cov)";
      die $@ if $@;
      $O->id = $id;
      $self->add($O);
      $cnt++;
      ($id,$vec,$cov) = (undef,undef,undef)
    }
    last if $cnt == $no_of_elements;
  }
  return 1;
}


sub elementlist_relations {
  my ($self,$fh,$current_block,$element_type,$no_of_elements) = @_;

  my $cnt = 0;
  my $id = undef;
  my @test_objects = ();
  my $test_objects = "";
  while (<$fh>) {
    next if /^\s*;/; # skip comments

    if (/ELEMENT\s+(\w+)/) {
      $id = $1;
      
      die "*** Parse error: Multiple object id: $id\n" if defined($self->object_id($id));
      next;
    }

    if (/PAIR\s+(\w+)\s+(\w+)/) {
      die "*** Parse Error id's are not known: $1,$2\n" unless (defined($self->object_id($1)) &&
								defined($self->object_id($2)));
      @test_objects = $self->object($1,$2);
      $test_objects = "$1 $2";
      next;
    }
    
    if (/END_ELEMENT/) {
      die "*** Error while parsing $_\n" if @test_objects==0;
      
      my $relation_type = "SUGR::Relation::".ucfirst(lc($element_type));

      my $relation = $relation_type->new(@test_objects);
      $relation->id = $id;

      $self->add($relation);
      
      @test_objects = ();
      $test_objects = "";
      $id = undef;
      $cnt++;
    }
    last if $cnt == $no_of_elements;
  }
}

1;


=head1 MINI TUTORIAL

=head2 A minimal example

We first start with a minimal, but still useful GXF(sugr) file,
defining two points in 3D.

  BLOCKSTART OBJECTS

    ELEMENTLIST POINTS 2
    ELEMENT Point_1
       TYPE EUCLIDEAN
       LIST_OF COORDS 3
       COORDS  1.0
       COORDS -2.0
       COORDS  3.0
    END_ELEMENT    
    ELEMENT Point_2
       TYPE HOMOGENEOUS
       LIST_OF COORDS 4
       COORDS  4.0
       COORDS -3.0
       COORDS  1.0
       COORDS  2.0
    END_ELEMENT

  BLOCKEND OBJECTS

  END_OF_FILE

This defines two 3D points which are called C<Point_1> and
C<Point_2>. They are referred to as identifiers and can be used later
on when constructing new elements.

The file consists of one block called C<OBJECTS>; in general, a block
starts with C<< BLOCKSTART <blockname> >> and ends with C<< BLOCKEND
<blockname> >>. Within the block, there can be elements (comparable to
a record), lists (comparable to an array) and tags (comparable to a 
fixed variable or fixed set of variables).

In this example, we have one list of elements called C<POINTS>. A list
of elements usually starts with the header C<< ELEMENTLIST <listname>
<number_of_listelements> >>, followed by C<number_of_listelements>
element entries. Each entry are delimited with C<< ELEMENT <id> >> and
C<< END_ELEMENT >>, where the start delimiter contains the unique
identifier for an element. This identifier has to be truly unique
across the whole GXF(sugr) file.

Between the element delimiters, the values for the elements are
listed. For the two points we may specify the C<TYPE> of coordinates,
which could either be C<EUCLIDEAN> or C<HOMOGENEOUS>. Then the actual
coordinates are specified using a list. 

A list contains of a header C<< LIST_OF <listname>
<number_of_listelements> >> and C<< <number_of_listelements> >> lines;
each line starts with the C<< <listname> >> and the list values, in
our case the coordinates of the object.

=head2 Introducing Uncertainty

Assume that we know the standard deviations of our parameter vector
For simplicity of this example, we assume the parameters to be not
correlated. We can specify a covariance matrix in an elementlist,
again using the C<< LIST_OF >> syntax:

    ; snippet from the last example 
    ELEMENT Point_1
       TYPE EUCLIDEAN
       LIST_OF COORDS 3
       COORDS  1.0
       COORDS -2.0
       COORDS  3.0
       LIST_OF COVARIANCES 3 ; 3 x 3 matrix
       COVARIANCES 0.01 0 0 
       COVARIANCES 0 0.01 0 
       COVARIANCES 0 0 0.01 
    END_ELEMENT    

Nothing fancy here. Note that one can include comments into a GXF
file, the comment character is C<;> as in Lisp.

=head2 Constructing new elements

(not written yet, meanwhile check example files)

=head2 Testing relations between elements

(not written yet, meanwhile check example files)

=head2 Complex Example file

More complex example GXF(sugr) files can be found the SUGR
distribution, see C<example/scene.gxf> or C<example/cube.gxf>.

=head1 BUGS

Probably many, but not documented yet.

=head1 TODO

=over 4

=item -

documentation

=back

=head1 SEE ALSO

SUGR::Scene(3), SUGR(3)

=head1 AUTHOR

Stephan Heuel (stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut
