=head1 NAME

SUGR::Transformation -- class for projective transformations of points, lines and planes.

(this class is deprecated! Do not use this class in future implementations)

=head1 VERSION

  This document refers to version 0.01

=head1 SYNOPSIS
 
 $TX = ... # a 4x4 matrix containing projective point transformation 
 $T = new SUGR::Transformation($TX);
 $LT = $T->transform($L); # transforms the line according 
                          # to point transformation
 $AT = $A->transform($T); # transforms the plane according
                          # to point transformation

=head1 DESCRIPTION


=head2 Overview

=head2 Constructurs and initializaton

=head2 Class and object methods

=head2 Something else?!

=head2 Literature

  see Note 'Homogeneous Transformation of Points, Planes and Lines in
  3D' by Wolfgang Frstner, 2000

=head1 DIAGNOSTICS

=over 4

=item "error messages?!"

=back

=head1 BUGS

=head1 TODO

=head1 FILES

=head1 SEE ALSO

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000 Stephan Heuel All rights reserved. There is no
warranty. You are allowed to redistribute this software /
documentation under certain conditions. For details, see the file
COPYING in the PDL distribution. If this file is separated from the
PDL distribution, the copyright notice should be included in the file.

=cut

package SUGR::Transformation;


use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 0.01;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL;
use PDL::Matrix;
use PDL::Math;
use PDL::Slatec;
use SUGR::Basic qw/Pidual/;

my $Q = mpdl [ [0,0,0,1,0,0],
	       [0,0,0,0,1,0],
	       [0,0,0,0,0,1],
	       [1,0,0,0,0,0],
	       [0,1,0,0,0,0],
	       [0,0,1,0,0,0] ];

# Constructors
#

# this is the constructor taking a 4x4 matrix (point/plane
# transformations) or a 6x6 matrix (line transformation) as input
sub new {
  my ($protoverse, @arg) = @_;
  my $class = ref($protoverse) || $protoverse;
  my $type = $arg[1] || "point";

  croak("transformation matrix is no PDL::Matrix") if ref($arg[0]) ne "PDL::Matrix";
  my ($r,$c) = $arg[0]->mdims;
  my $self;

  if ($type =~ m/point2d/i) {
    croak("transformation matrix is not 3x3 (assumably a 2D point transformation) ") if ($r != 3 || $c !=3);
    $self->{"SUGR::Point2d"} = $arg[0] ; 
    $self->{"SUGR::Line2d"}  = Point2dToLine2dTransform($self->{"SUGR::Point2d"}) ; 
    $self->{dims} = 2;
  }
  elsif ($type =~ m/line2d/i) {
    croak("transformation matrix is not 3x3 (assumably a 2d line transformation)") if ($r != 3 || $c !=3);
    $self->{"SUGR::Line2d"}  = $arg[0] ; 
    $self->{"SUGR::Point2d"} = Line2dToPoint2dTransform($self->{"SUGR::Line2d"});
    $self->{dims} = 2;
  }
  elsif ($type =~ m/point/i) {
    croak("transformation matrix is not 4x4 (assumably a point transformation) ") if ($r != 4 || $c !=4);
    $self->{"SUGR::Point"} = $arg[0] ; 
    $self->{"SUGR::Line"}  = PointToLineTransform($self->{"SUGR::Point"}) ; 
    $self->{"SUGR::Plane"} = PointToPlaneTransform($self->{"SUGR::Point"}) ; 
    $self->{dims} = 3;
  }
  elsif ($type =~ m/line/i) {
    croak("transformation matrix is not 6x6 (assumably a point transformation)") if ($r != 6 || $c !=6);
    $self->{"SUGR::Line"}  = $arg[0] ; 
    $self->{"SUGR::Point"} = LineToPointTransform($self->{"SUGR::Line"});
    $self->{"SUGR::Plane"} = PointToPlaneTransform($self->{"SUGR::Point"}) ; 
    $self->{dims} = 3;
  }
  elsif ($type =~ m/point/i) {
    croak("transformation matrix is not 4x4 (assumably a plane transformation)") if ($r != 4 || $c !=4);
    $self->{"SUGR::Plane"} = $arg[0] ; 
    $self->{"SUGR::Point"} = PlaneToPointTransform($self->{"SUGR::Plane"}) ; 
    $self->{"SUGR::Line"}  = PointToLineTransform($self->{"SUGR::Point"}) ; 
    $self->{dims} = 3;
  }
  elsif ($type =~ m/pointprojection/i) {
    croak("transformation matrix is not 3x4 (i.e. no point projection)") if ($r != 3 || $c !=4);
    $self->{"SUGR::Plane"} = undef;
    $self->{"SUGR::Point"} = PlaneToPointTransform($self->{"SUGR::Plane"}) ; 
    $self->{"SUGR::Line"}  = PointToLineTransform($self->{"SUGR::Point"}) ; 
    $self->{dims} = undef;
  }
  elsif ($type =~ m/lineprojection/i) {
    croak("transformation matrix is not 3x6 (i.e. no line projection)") if ($r != 3 || $c !=6);
    $self->{"SUGR::Plane"} = undef;
    $self->{"SUGR::Point"} = PlaneToPointTransform($self->{"SUGR::Plane"}) ; 
    $self->{"SUGR::Line"}  = PointToLineTransform($self->{"SUGR::Point"}) ; 
    $self->{dims} = undef;
  }
  else {
    croak("unknown transformation for ".$type);
  }

  bless $self,$class;
}

sub Translation {
  my ($tvec,$type) = @_;
  $type ||= "point";
  
  croak("translation vector is no PDL::Matrix") if ref($tvec) ne "PDL::Matrix";
  croak("BUG: sorry, translation vector must be only for points") if $type ne "point";

  my $T = SUGR::Element::Identity(4);
  ($_ = $T->slice("0:2,3")) .= $tvec;
  
  return new SUGR::Transformation($T,$type);
}

sub create_TranslationPointLine {
   my ($X,$L) = @_;

   $X = $X->copy->euclidize();

   # translate so that $X is origin 
   my $T1 = Translation(-$X->vec->slice("0:2"));
   my $LT1 = $L->transform($T1);
   # compute point closest to origin
   my $D = $LT1->point_closest_to_origin->euclidize;
   my $T2 = Translation($D->vec->slice("0:2"));

   return $T2;
}
push @EXPORT,"create_TranslationPointLine";

# creates a new SUGR::Transformation object, which is the inverse of
# the given one
sub inverse {
  my ($self) = @_;

  return undef if !defined($self->{dims});

  my $type = $self->{dims} == 3 ? "Point" : "Point2d";

  return new SUGR::Transformation( $self->{"SUGR::$type"}->matinv,$type);

}
sub PointToLineTransform {
  my ($M) = @_ ; # M is the point transform 
  my $N = mzeroes(6,6);

  my @ind1 = (1,2,0,3,3,3);
  my @ind2 = (2,0,1,0,1,2);
  for (my $i=0;$i<6;$i++) {
    my $L = Pidual(~$M->slice($ind1[$i].",")) x ~$M->slice($ind2[$i].",");
    ($_ = $N->slice("$i,")) .= ~$L->rotate(3); # store the dual
  }
  return $Q x $N;
#  return ~SL x  kroneckerproduct($M,$M) x $SL;
}

sub LineToPointTransform {
  my ($M) = @_ ; # M is the point transform 
  croak("LineToPointTransform not implemented yet");
}
sub PointToPlaneTransform {
  my ($M) = @_ ; # M is the point transform 
  return ~$M->inv();
}
sub PlaneToPointTransform {
  my ($M) = @_ ; # M is the plane transform 
  return ~$M->inv();
}

sub Point2dToLine2dTransform {
  my ($M) = @_ ; # M is the point transform 
  return ~$M->inv();
}
sub Line2dToPoint2dTransform {
  my ($M) = @_ ; # M is the line transform 
  return ~$M->inv();
}


# here we do the transformation depending on the type of SUGR::Element
sub transform {
  my ($Transform,$Element) = @_;
  
  my $type = ref($Element);
  # some testing
  croak("first arg is not a transform") if ref($Transform) !~ /SUGR::Transformation/;
  croak("second arg is neither point, line nor a plane") 
      if ($type !~ /SUGR::Point|SUGR::Plane|SUGR::Line|SUGR::Line2d|SUGR::Point2d/);

  my $vec = $Transform->{$type} x $Element->vec;
  return eval "new $type(\$vec)" if (!defined($Element->cov));
  my $cov = $Transform->{$type} x $Element->cov x ~$Transform->{$type};
  return eval "new $type(\$vec,\$cov)";
}

# define accessing methods
foreach my $func qw(Point2d Line2d Point Line Plane) {
  my $code = << "EOE";
sub $func {
  my \$self = shift;
  return \$self->{\"SUGR::$func\"};
}
EOE
  eval $code;
}

use overload q{""} => \&stringify;

sub stringify {
  my ($self) = shift;
  my $type = ($self->{dims}==3) ? "SUGR::Point" : "SUGR::Point2d";
  return "$type-Transformation: ".$self->{$type};
}

return 1 if caller;

####### Test #########
use Math::Random qw(random_uniform);
use SUGR::Point;
use SUGR::Line;
use SUGR::Basic qw/Gamma/;

for (0..20) {
  my $omega = random_uniform(1,0,90)*(3.14159/180);
  my $phi   = random_uniform(1,0,90)*(3.14159/180);
  my $kappa = random_uniform(1,0,90)*(3.14159/180);
  my $tx = random_uniform(1,-10,10);
  my $ty = random_uniform(1,-10,10);
  my $tz = random_uniform(1,-10,10);
  
  my $T = mzeroes(4,4);
  my $R = mpdl([
		[cos($omega),-sin($omega), 0 ],
		[sin($omega), cos($omega), 0 ],
		[           0,          0, 1 ]
	       ]) x
		 mpdl([
		       [ cos($phi) ,0, sin($phi)  ],
		       [ 0         ,1,         0  ],
		       [-sin($phi), 0, cos($phi) ]
		      ]) x
			mpdl([
			      [1,          0,          0 ],
			      [0,cos($omega),-sin($omega)],
			      [0,sin($omega), cos($omega)]
			     ]);
    
  ($_ = $T->slice("0:3,3")) .= vpdl([$tx,$ty,$tz,1]);
  ($_ = $T->slice("0:2,0:2")) .= $R;


  my $TR = new SUGR::Transformation($T);
    
  my $X1 = $TR->transform(sugr_Point([1 , 0, 2,1]));
  my $X2 = $TR->transform(sugr_Point([2 , 1, 1,1]));
  my $X3 = $TR->transform(sugr_Point([1 , 1, 0,1]));
  my $L = sugr_Line($X1,$X2);
    
  my $TLP = create_TranslationPointLine($X3,$L);
  my $X3T = $TLP->transform($X3);
  my $X3C = $L->point_closest_to_point($X3);

  print "point on line 1 ? ",(sum(-Gamma($L) x $X3T->vec) < 1e-12),"\n";
  print "point on line 2 ? ",(sum(-Gamma($L) x $X3C->vec) < 1e-12),"\n";
}

1;
