=head1 NAME

SUGR::TransformationNew -- abstract class for linear transformation of entities

(!!! at some point will be renamed to SUGR::Transformation !!)

=head1 NOT DOCUMENTED YET

=head1 BUGS

Probably many, but not documented yet.

=head1 SEE ALSO

SUGR(3), SUGR::Element(3)

=head1 AUTHOR

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000/2001 Stephan Heuel and Institute for Photogrammetry,
University of Bonn. All rights reserved. There is no warranty. You are
allowed to redistribute this software / documentation under certain
conditions, see the GNU Public License, GPL for details.

=cut

package SUGR::TransformationNew;

use strict;
use Carp;
use vars qw(@EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

our @ISA = qw(SUGR::Element Exporter);
use Exporter;
use SUGR::Element;
use SUGR::Basic qw/Identity/;
use PDL::Matrix qw/kroneckerproduct mpdl/;

$VERSION = SUGR->VERSION;

@EXPORT = qw(sugr_Transformation); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

my %valid_dimensions = ( "3 3"   => "2D->2D",
			 "4 4"   => "3D->3D",
			 "3 4"   => "3D->2D(projective)",
			 "2 4"   => "3D->2D(affine)",
			 "4 3"   => "2D->(3D-Plane)"
		       );


#
# this is a constructor that can be called like it is
sub sugr_Transformation { SUGR::Transformation->new(@_) }

sub ConstructFromMatrix { 
  ConstructFromMatrixMatrix(@_);

  my ($matrix) = @_;
  croak("no transformation matrix given") if !defined($matrix);
  my @dims = $matrix->mdims;
  croak("unknown dimension for transformation matrix") 
    if !defined($valid_dimensions{"@dims"});

  my $self = SUGR::Element::ConstructFromMatrix($matrix->reshape($dims[0]*$dims[1],1));

  $self->{dims} = [@dims];

  # TODO: automatically calculate line transformation from given point
  # transformation
#  ($self->{vec_line}) = LineTransformation($self->{vec});
  # for the matrix dimension of lines we need to replace 4 by 6
  $self->{dims_line} = [map { $_==4 ? 6 : $_; } @dims];

  return $self;
}

# this is the actual constructor !
sub ConstructFromMatrixMatrix {
  my ($matrix,$cov) = @_;

  croak("no matrix given") if !defined($matrix);
  my @dims = $matrix->mdims;
  croak("unknown dimension for transformation matrix") 
    if !defined($valid_dimensions{"@dims"});
  croak("covariance matrix is no PDL::Matrix, is ".ref($cov)) 
    if (defined($cov) && ref($cov) !~ /PDL::Matrix/);
  
  my $self = defined($cov) ? 
    SUGR::Element::ConstructFromMatrixMatrix($matrix->reshape($dims[0]*$dims[1],1),$cov) :
	SUGR::Element::ConstructFromMatrix($matrix->reshape(12,1));

  $self->{dims} = [@dims];

  # TODO: automatically calculate line transformation from given point
  # transformation
#  ($self->{vec_line},$self->{cov_line}) = LineTransformation($self->{vec},$self->{cov});
  # for the matrix dimension of lines we need to replace 4 by 6
  $self->{dims_line} = [map { $_==4 ? 6 : $_; } @dims];

  return $self;
}

sub ConstructFromARRAY {
  my (@arg) = @_;
  ConstructFromMatrix(mpdl($arg[0]));
}
sub ConstructFromARRAYARRAY {
  my (@arg) = @_;
  ConstructFromMatrixMatrix(mpdl($arg[0]),mpdl($arg[1]));
}
sub ConstructFromARRAYMatrix {
  my (@arg) = @_;
  ConstructFromMatrixMatrix(mpdl($arg[0]),$arg[1]);
}
sub ConstructFromMatrixArray {
  my (@arg) = @_;
  ConstructFromMatrixMatrix($arg[0],mpdl($arg[1]));
}

# how to copy objects
sub copy {
  my ($self) = @_;
  
  my $copy = $self->SUGR::Element::copy;
  foreach my $additional_key qw(dims dims_line vec_line cov_line) {
    $copy->{$additional_key} = $self->{$additional_key}
  }
  return $copy;
}

sub type {
  "Transformation";
}

# method return matrix representation of element
sub matrix {
  my ($self,$type) = @_;
  $type ||= "point"; # default value
  $type = lc($type);

  my $key = ($type =~ /point/) ? "" : "_$type";
  $key =~ s/sugr:://;

  $self->{"vec$key"}->reshape(@{$self->{"dims$key"}});
}


sub vec {
  my ($self,$type) = @_;
  $type ||= "point"; # default value
  $type = lc($type);

  my $key = ($type =~ /point/) ? "" : "_$type";

  $self->{"vec$key"};
}

sub cov {
  my ($self,$type) = @_;
  $type ||= "Point"; # default value
  # $type ignored at the moment
  $type ||= "point"; # default value
  $type = lc($type);

  my $key = ($type =~ /point/) ? "" : "_$type";

  $self->{"cov$key"};
}

use overload q{""} => \&stringify;

# override method from SUGR::Element
sub is_transform {
  return 1;
}

sub result_type {
  warn "call to abstract method result_type\n";
  undef;
}

# here we do the transformation depending on the type of SUGR::Element
sub transform {
  my ($Transform,$Element) = @_;
  
  my $type = ref($Element);
  # some testing
  croak("first arg is not a SUGR transform") if !$Transform->is_transform;
  croak("second arg is neither point, line nor a plane") 
      if ($type !~ /SUGR::Point|SUGR::Plane|SUGR::Line|SUGR::Line2d|SUGR::Point2d/);

  my $result_type = $Transform->result_type($type);
  my $vec = $Transform->matrix($type) x $Element->vec;
  return eval "new $result_type(\$vec)" if (!defined($Element->cov));
  my $cov = $Transform->matrix($type) x $Element->cov x ~$Transform->matrix($type);
  # here we ignore that the transformation has a covariance, too.
  if (defined($Transform->cov($type))) {
    my ($rows,undef) = $Transform->matrix($type)->mdims;
    my $J = kroneckerproduct(Identity($rows),~$Element->vec($type));
    $cov += $J x $Transform->cov($type) x ~$J;
  }

  return eval "new $result_type(\$vec,\$cov)";
}

sub stringify { 
  my ($self) = shift;

  (my $s = ref($self)." Matrix::\n") =~ s/SUGR:://;

  foreach my $type (qw/point/) { # qw/point line/
    $s .= "- $type-transformation ";
    $s .= $valid_dimensions{"@{$self->{dims}}"}." : ";
    $s .= $self->matrix($type);
  }

  my %reshape = ( point => [3,4], line => [3,6] );
  foreach my $type (qw/point/) { # qw/point line/
    if (defined($self->cov($type))) {
      $s .= "Variances of $type-transformation matrix: ";
      $s .=  $self->cov($type)->diagonal(0,1)->reshape(@{$self->{dims}})."\n";
    } 
  }
  return $s;
}

return 1 if caller;

####### Test #########

