;;; sugrshell.el --- specialized comint.el for running the sugrshell.

;;; This is based on shell.el and only slightly changed so that the
;;; minimal features of the sugrshell run. 

;; Author: Olin Shivers <shivers@cs.cmu.edu> then
;;	Simon Marshall <simon@gnu.ai.mit.edu>
;; Maintainer: Pete Ware <ware@cis.ohio-state.edu>


;; $Docid: Oct. 9th 2001  17:45  Stephan Heuel $

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This file defines a a shell-in-a-buffer package (shell mode) built
;; on top of comint mode.  This is a derived version from shell.el and
;; provides a comint mode for the sugrshell, a shell for statistically
;; uncertaing geometric reasoning.

;; Note that not there are a lot of old functions and comments from
;; shell.el which make no sense in the sugrshell mode. Please
;; disregard them.

;; Since this mode is built on top of the general command-interpreter-in-
;; a-buffer mode (comint mode), it shares a common base functionality, 
;; and a common set of bindings, with all modes derived from comint mode.
;; This makes these modes easier to use.

;; For documentation on the functionality provided by comint mode, and
;; the hooks available for customising it, see the file comint.el.
;; For further information on shell mode, see the comments below.


;; Read the rest of this file for more information.

;; Customization and Buffer Variables
;; ===========================================================================
;; 

;;; Code:

(require 'comint)

(defgroup sugrshell nil
  "Running sugrshell from within Emacs buffers"
  :group 'processes
  :group 'unix)

(defgroup sugrshell-directories nil
  "Directory support in sugrshell mode"
  :group 'sugrshell)

(defgroup sugrshell-faces nil
  "Faces in sugrshell buffers"
  :group 'sugrshell)

;;;###autoload
(defvar sugrshell-prompt-pattern "^sugr> *"
  "Regexp to match prompts in the inferior sugrshell.
Defaults to \"^[^#$%>\\n]*[#$%>] *\", which works pretty well.
This variable is used to initialise `comint-prompt-regexp' in the 
sugrshell buffer.

The pattern should probably not match more than one line.  If it does,
Sugrshell mode may become confused trying to distinguish prompt from input
on lines which don't start with a prompt.

This is a fine thing to set in your `.emacs' file.")

(defcustom sugrshell-completion-fignore nil
  "*List of suffixes to be disregarded during file/command completion.
This variable is used to initialize `comint-completion-fignore' in the sugrshell
buffer.  The default is nil, for compatibility with most sugrshells.
Some people like (\"~\" \"#\" \"%\").

This is a fine thing to set in your `.emacs' file."
  :type '(repeat (string :tag "Suffix"))
  :group 'sugrshell)

;jwz: turned this off; it's way too broken.
(defvar sugrshell-delimiter-argument-list nil ;'(?\| ?& ?< ?> ?\( ?\) ?\;
  "List of characters to recognise as separate arguments.
This variable is used to initialize `comint-delimiter-argument-list' in the
sugrshell buffer.  The value may depend on the operating system or sugrshell.

This is a fine thing to set in your `.emacs' file.")

(defvar sugrshell-file-name-chars
  (if (memq system-type '(ms-dos windows-nt))
      "~/A-Za-z0-9_^$!#%&{}@`'.()-"
    "~/A-Za-z0-9+@:_.$#%,={}-")
  "String of characters valid in a file name.
This variable is used to initialize `comint-file-name-chars' in the
sugrshell buffer.  The value may depend on the operating system or sugrshell.

This is a fine thing to set in your `.emacs' file.")

(defvar sugrshell-file-name-quote-list
  (if (memq system-type '(ms-dos windows-nt))
      nil
    (append sugrshell-delimiter-argument-list '(?\  ?\* ?\! ?\" ?\' ?\`)))
  "List of characters to quote when in a file name.
This variable is used to initialize `comint-file-name-quote-list' in the
sugrshell buffer.  The value may depend on the operating system or sugrshell.

This is a fine thing to set in your `.emacs' file.")

(defvar sugrshell-dynamic-complete-functions
  '(comint-replace-by-expanded-history
    sugrshell-dynamic-complete-environment-variable
    sugrshell-dynamic-complete-command
    sugrshell-replace-by-expanded-directory
    comint-dynamic-complete-filename)
  "List of functions called to perform completion.
This variable is used to initialise `comint-dynamic-complete-functions' in the
sugrshell buffer.

This is a fine thing to set in your `.emacs' file.")

(defcustom sugrshell-command-regexp "[^;]+"
  "*Regexp to match a single command."
  :type 'regexp
  :group 'sugrshell)

(defcustom sugrshell-completion-execonly t
  "*If non-nil, use executable files only for completion candidates.
This mirrors the optional behavior of tcsh.

Detecting executability of files may slow command completion considerably."
  :type 'boolean
  :group 'sugrshell)

(defcustom sugrshell-multiple-sugrshells nil
  "*If non-nil, each time sugrshell mode is invoked, a new sugrshell is made"
  :type 'boolean
  :group 'sugrshell)

;(defcustom sugrshell-popd-regexp "popd"
;  "*Regexp to match subsugrshell commands equivalent to popd."
;  :type 'regexp
;  :group 'sugrshell-directories)

;(defcustom sugrshell-pushd-regexp "pushd"
;  "*Regexp to match subsugrshell commands equivalent to pushd."
;  :type 'regexp
;  :group 'sugrshell-directories)

;(defcustom sugrshell-pushd-tohome nil
;  "*If non-nil, make pushd with no arg behave as \"pushd ~\" (like cd).
;This mirrors the optional behavior of tcsh."
;  :type 'boolean
;  :group 'sugrshell-directories)

;(defcustom sugrshell-pushd-dextract nil
;  "*If non-nil, make \"pushd +n\" pop the nth dir to the stack top.
;This mirrors the optional behavior of tcsh."
;  :type 'boolean
;  :group 'sugrshell-directories)

;(defcustom sugrshell-pushd-dunique nil
;  "*If non-nil, make pushd only add unique directories to the stack.
;This mirrors the optional behavior of tcsh."
;  :type 'boolean
;  :group 'sugrshell-directories)

;(defcustom sugrshell-cd-regexp "# cd"
;  "*Regexp to match subsugrshell commands equivalent to cd."
;  :type 'regexp
;  :group 'sugrshell-directories)

(defcustom sugrshell-chdrive-regexp
  (if (memq system-type '(ms-dos windows-nt)) 
      ; NetWare allows the five chars between upper and lower alphabetics.
      "[]a-zA-Z^_`\\[\\\\]:"
    nil)
  "*If non-nil, is regexp used to track drive changes."
  :type '(choice regexp
		 (const nil))
  :group 'sugrshell-directories)

(defcustom explicit-sugrshell-file-name "sugrshell"
  "*If non-nil, is file name to use for explicitly requested inferior sugrshell."
  :type '(choice (const :tag "None" nil) file)
  :group 'sugrshell)

;(defcustom explicit-sugrshell-args '("-tk")
(defcustom explicit-sugrshell-args '(" ")
  "*List of arguments to pass to \"sugrshell\" on startup in `sugrshell-mode'."
  :type '(repeat (string :tag "Argument"))
  :group 'sugrshell)

;(defcustom explicit-bash-args '("-i")
;  "*List of arguments to pass to \"bash\" on startup in `sugrshell-mode'."
;  :type '(repeat (string :tag "Argument"))
;  :group 'sugrshell)

;(defcustom explicit-csh-args
;  (if (eq system-type 'hpux)
;      ;; -T persuades HP's csh not to think it is smarter
;      ;; than us about what terminal modes to use.
;      '("-i" "-T")
;    '("-i"))
;  "*Args passed to inferior sugrshell by M-x sugrshell, if the sugrshell is csh.
;Value is a list of strings, which may be nil."
;  :type '(repeat (string :tag "Argument"))
;  :group 'sugrshell)

;(defcustom explicit-ksh-args '("-i")
;  "*List of arguments to pass to \"ksh\" on startup in `sugrshell-mode'."
;  :type '(repeat (string :tag "Argument"))
;  :group 'sugrshell)

;(defcustom explicit-pdsh-args '("-i")
;  "*List of arguments to pass to \"pdksh\" on startup in `sugrshell-mode'."
;  :type '(repeat (string :tag "Argument"))
;  :group 'sugrshell)

;(defcustom explicit-tcsh-args '("-i")
;  "*List of arguments to pass to \"tcsh\" on startup in `sugrshell-mode'."
;  :type '(repeat (string :tag "Argument"))
;  :group 'sugrshell)

;(defcustom explicit-zsh-args '("-i")
;  "*List of arguments to pass to \"zsh\" on startup in `sugrshell-mode'."
;  :type '(repeat (string :tag "Argument"))
;  :group 'sugrshell)

(defcustom sugrshell-input-autoexpand 'history
  "*If non-nil, expand input command history references on completion.
This mirrors the optional behavior of tcsh (its autoexpand and histlit).

If the value is `input', then the expansion is seen on input.
If the value is `history', then the expansion is only when inserting
into the buffer's input ring.  See also `comint-magic-space' and
`comint-dynamic-complete'.

This variable supplies a default for `comint-input-autoexpand',
for Sugrshell mode only."
  :type '(choice (const nil) (const input) (const history))
  :type 'sugrshell)

;(defvar sugrshell-dirstack nil
;  "List of directories saved by pushd in this buffer's sugrshell.
;Thus, this does not include the sugrshell's current directory.")

;(defvar sugrshell-dirtrackp t
;  "Non-nil in a sugrshell buffer means directory tracking is enabled.")

;(defvar sugrshell-last-dir nil
;  "Keep track of last directory for ksh `cd -' command.")

;(defvar sugrshell-dirstack-query nil
;  "Command used by `sugrshell-resync-dir' to query the sugrshell.")

(defvar sugrshell-mode-map nil)
(if (not sugrshell-mode-map)
    (let ((map (make-keymap)))
      (set-keymap-parents map (list comint-mode-map))
      (set-keymap-name map 'sugrshell-mode-map)
      (define-key map "\C-c\C-f" 'sugrshell-forward-command)
      (define-key map "\C-c\C-b" 'sugrshell-backward-command)
      (define-key map "\t" 'comint-dynamic-complete)
      (define-key map "\M-?"  'comint-dynamic-list-filename-completions)
      ;; XEmacs: this is a pretty common operation for those of us
      ;; who use directory aliases ...  someone shoot me if they
      ;; don't like this binding.  Another possibility is C-c C-s
      ;; but that's way awkward.
      ;; July-5-1997, Bang! -slb
      (unless (featurep 'infodock)
	(define-key map "\M-\C-m" 'sugrshell-resync-dirs))
      (setq sugrshell-mode-map map)))

(defcustom sugrshell-mode-hook nil
  "*Hook for customising Sugrshell mode."
  :type 'hook
  :group 'sugrshell)


;; font-locking
(defcustom sugrshell-prompt-face 'sugrshell-prompt-face
  "Face for sugrshell prompts."
  :type 'face
  :group 'sugrshell-faces)
(defcustom sugrshell-option-face 'sugrshell-option-face
  "Face for command line options."
  :type 'face
  :group 'sugrshell-faces)
(defcustom sugrshell-output-face 'sugrshell-output-face
  "Face for generic sugrshell output."
  :type 'face
  :group 'sugrshell-faces)
(defcustom sugrshell-output-2-face 'sugrshell-output-2-face
  "Face for grep-like output."
  :type 'face
  :group 'sugrshell-faces)
(defcustom sugrshell-output-3-face 'sugrshell-output-3-face
  "Face for [N] output where N is a number."
  :type 'face
  :group 'sugrshell-faces)

(make-face sugrshell-prompt-face)
(make-face sugrshell-option-face)
(make-face sugrshell-output-face)
(make-face sugrshell-output-2-face)
(make-face sugrshell-output-3-face)

(defun sugrshell-font-lock-mode-hook ()
  (or (face-differs-from-default-p sugrshell-prompt-face)
      (copy-face 'font-lock-keyword-face sugrshell-prompt-face))
  (or (face-differs-from-default-p sugrshell-option-face)
      (copy-face 'font-lock-comment-face sugrshell-option-face))
  (or (face-differs-from-default-p sugrshell-output-face)
      (copy-face 'italic sugrshell-output-face))
  (or (face-differs-from-default-p sugrshell-output-2-face)
      (copy-face 'font-lock-string-face sugrshell-output-2-face))
  (or (face-differs-from-default-p sugrshell-output-3-face)
      (copy-face 'font-lock-string-face sugrshell-output-3-face))
  ;; we only need to do this once
  (remove-hook 'font-lock-mode-hook 'sugrshell-font-lock-mode-hook))
(add-hook 'font-lock-mode-hook 'sugrshell-font-lock-mode-hook)

(defvar sugrshell-prompt-pattern-for-font-lock nil
  "If non-nil, pattern to use to font-lock the prompt.
When nil, sugrshell-prompt-pattern will be used.  Set this to a regular
expression if you want the font-locked pattern to be different then
the sugrshell's prompt pattern.")

(defvar sugrshell-font-lock-keywords
  (list '(eval . (cons (if sugrshell-prompt-pattern-for-font-lock
			   sugrshell-prompt-pattern-for-font-lock
			 sugrshell-prompt-pattern)
		       sugrshell-prompt-face))
;	'("[ \t]\\([+-][^ \t\n>]+\\)" 1 sugrshell-option-face)
	'("\[[^\n]*\]" . sugrshell-output-3-face)
	'("^[ \t]*[][][ \t]*$" . sugrshell-output-3-face)
	'("^[^ \t\n]+:.*" . sugrshell-output-2-face)
	'("^[^\n]+.*$" . sugrshell-output-face))
  "Additional expressions to highlight in Sugrshell mode.")
(put 'sugrshell-mode 'font-lock-defaults '(sugrshell-font-lock-keywords t))


;;; Basic Procedures

(put 'sugrshell-mode 'mode-class 'special)

(defun sugrshell-mode ()
  "Major mode for interacting with an inferior sugrshell.
\\<sugrshell-mode-map>\\[comint-send-input] after the end of the process' output sends the text from
    the end of process to the end of the current line.
\\[comint-send-input] before end of process output copies the current line minus the
    prompt to the end of the buffer and sends it (\\[comint-copy-old-input] just copies
    the current line).
\\[send-invisible] reads a line of text without echoing it, and sends it to
    the sugrshell.  This is useful for entering passwords.  Or, add the function
    `comint-watch-for-password-prompt' to `comint-output-filter-functions'.

If you want to make multiple sugrshell buffers, rename the `*sugrshell*' buffer
using \\[rename-buffer] or \\[rename-uniquely] and start a new sugrshell.

If you want to make sugrshell buffers limited in length, add the function
`comint-truncate-buffer' to `comint-output-filter-functions'.

If you accidentally suspend your process, use \\[comint-continue-subjob]
to continue it.

`cd', `pushd' and `popd' commands given to the sugrshell are watched by Emacs to
keep this buffer's default directory the same as the sugrshell's working directory.
While directory tracking is enabled, the sugrshell's working directory is displayed
by \\[list-buffers] or \\[mouse-buffer-menu] in the `File' field.
\\[dirs] queries the sugrshell and resyncs Emacs' idea of what the current 
    directory stack is.
\\[dirtrack-toggle] turns directory tracking on and off.

\\{sugrshell-mode-map}
Customization: Entry to this mode runs the hooks on `comint-mode-hook' and
`sugrshell-mode-hook' (in that order).  Before each input, the hooks on
`comint-input-filter-functions' are run.  After each sugrshell output, the hooks
on `comint-output-filter-functions' are run.

Variable `sugrshell-multiple-sugrshells' will automatically generate a new sugrshell each
time it is invoked.

Variables `sugrshell-cd-regexp', `sugrshell-chdrive-regexp', `sugrshell-pushd-regexp' 
and `sugrshell-popd-regexp' are used to match their respective commands, 
while `sugrshell-pushd-tohome', `sugrshell-pushd-dextract' and `sugrshell-pushd-dunique' 
control the behavior of the relevant command.

Variables `comint-completion-autolist', `comint-completion-addsuffix',
`comint-completion-recexact' and `comint-completion-fignore' control the
behavior of file name, command name and variable name completion.  Variable
`sugrshell-completion-execonly' controls the behavior of command name completion.
Variable `sugrshell-completion-fignore' is used to initialise the value of
`comint-completion-fignore'.

Variables `comint-input-ring-file-name' and `comint-input-autoexpand' control
the initialisation of the input ring history, and history expansion.

Variables `comint-output-filter-functions', a hook, and
`comint-scroll-to-bottom-on-input' and `comint-scroll-to-bottom-on-output'
control whether input and output cause the window to scroll to the end of the
buffer."
  (interactive)
  (comint-mode)
  (setq major-mode 'sugrshell-mode)
  (setq mode-name "Sugrshell")
  (use-local-map sugrshell-mode-map)
  (make-local-variable 'comint-prompt-regexp)
  (setq comint-prompt-regexp sugrshell-prompt-pattern)
  (setq comint-completion-fignore sugrshell-completion-fignore)
  (make-local-variable 'comint-delimiter-argument-list)
  (setq comint-delimiter-argument-list sugrshell-delimiter-argument-list)
  (setq comint-file-name-chars sugrshell-file-name-chars)
  (make-local-variable 'comint-after-partial-filename-command)
  (setq comint-after-partial-filename-command 'sugrshell-after-partial-filename)
  (make-local-variable 'comint-get-current-command)
  (setq comint-get-current-command 'sugrshell-get-current-command)
  (make-local-variable 'comint-dynamic-complete-command-command)
  (setq comint-dynamic-complete-command-command 'sugrshell-dynamic-complete-command)
  (setq comint-file-name-quote-list sugrshell-file-name-quote-list)
  (setq comint-dynamic-complete-functions sugrshell-dynamic-complete-functions)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start comint-prompt-regexp)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(sugrshell-font-lock-keywords t))
  (make-local-variable 'sugrshell-dirstack)
  (setq sugrshell-dirstack nil)
  (make-local-variable 'sugrshell-last-dir)
  (setq sugrshell-last-dir nil)
  (make-local-variable 'sugrshell-dirtrackp)
  (setq sugrshell-dirtrackp t)
  (add-hook 'comint-input-filter-functions 'sugrshell-directory-tracker nil t)
  (setq comint-input-autoexpand sugrshell-input-autoexpand)
  ;; This is not really correct, since the sugrshell buffer does not really
  ;; edit this directory.  But it is useful in the buffer list and menus.
  (make-local-variable 'list-buffers-directory)
  (setq list-buffers-directory (expand-file-name default-directory))
  ;; sugrshell-dependent assignments.
  (let ((sugrshell (file-name-nondirectory (car
		 (process-command (get-buffer-process (current-buffer)))))))
    (setq comint-input-ring-file-name
	  (or (getenv "SUGRHISTFILE")
	      "~/.sugr_hist"))
    (if (or (equal comint-input-ring-file-name "")
	    (equal (file-truename comint-input-ring-file-name)
		   (file-truename "/dev/null")))
	(setq comint-input-ring-file-name nil))
    (setq sugrshell-dirstack-query
	  (cond ((string-equal sugrshell "sugrshell") "# pwd")
		(t "# dirs"))))
  (run-hooks 'sugrshell-mode-hook)
  (comint-read-input-ring t)
  (sugrshell-dirstack-message))


;;;###autoload
(defun sugrshell ()
  "Run an inferior sugrshell, with I/O through buffer *sugrshell*.
If buffer exists but sugrshell process is not running, make new sugrshell.
If buffer exists and sugrshell process is running, 
 just switch to buffer `*sugrshell*'.
Program used comes from variable `explicit-sugrshell-file-name',
 or (if that is nil) from the ESUGRSHELL environment variable,
 or else from SUGRSHELL if there is no ESUGRSHELL,
 or if there is no SUGRSHELL then from variable `sugrshell-file-name',
 or, if all of the above are nil, \"/bin/sh\" is used.
If a file `~/.emacs_SUGRSHELLNAME' exists, it is given as initial input
 (Note that this may lose due to a timing error if the sugrshell
  discards input when it starts up.)
The buffer is put in Sugrshell mode, giving commands for sending input
and controlling the subjobs of the sugrshell.  See `sugrshell-mode'.
See also the variable `sugrshell-prompt-pattern'.

To specify a coding system for converting non-ASCII characters
in the input and output to the sugrshell, use \\[universal-coding-system-argument]
before \\[sugrshell].  You can also specify this with \\[set-buffer-process-coding-system]
in the sugrshell buffer, after you start the sugrshell.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

The sugrshell file name (sans directories) is used to make a symbol name
such as `explicit-csh-args'.  If that symbol is a variable,
its value is used as a list of arguments when invoking the sugrshell.
Otherwise, one argument `-i' is passed to the sugrshell.

\(Type \\[describe-mode] in the sugrshell buffer for a list of commands.)"
  (interactive)
  (let ((buffer "*sugrshell*")
	(buffer-name (if sugrshell-multiple-sugrshells
			 "*sugrshell*"
		       "sugrshell")))
  (cond ((or sugrshell-multiple-sugrshells
	     (not (comint-check-proc buffer)))
	 (let* ((prog (or (getenv "SUGRSHELL")
			  explicit-sugrshell-file-name
			  "/usr/local/bin/sugrshell"))
		(name (file-name-nondirectory prog))
		(startfile (concat "~/.emacs_" name))
		(xargs-name (intern-soft (concat "explicit-" name "-args"))))
	   (setq buffer (set-buffer (apply 'make-comint buffer-name prog
					   (if (file-exists-p startfile)
					       startfile)
					   (if (and xargs-name
						    (boundp xargs-name))
					       (symbol-value xargs-name)
;					     '("-i")
					     ))))
	   (sugrshell-mode))))
  (prog1
      buffer
    (pop-to-buffer buffer)
    (if sugrshell-multiple-sugrshells
	(rename-buffer (generate-new-buffer-name "*sugrshell*"))))))

;;; Don't do this when sugrshell.el is loaded, only while dumping.
;;;###autoload (add-hook 'same-window-buffer-names "*sugrshell*")

;;; Directory tracking
;;;
;;; This code provides the sugrshell mode input sentinel
;;;     SUGRSHELL-DIRECTORY-TRACKER
;;; that tracks cd, pushd, and popd commands issued to the sugrshell, and
;;; changes the current directory of the sugrshell buffer accordingly.
;;;
;;; This is basically a fragile hack, although it's more accurate than
;;; the version in Emacs 18's sugrshell.el. It has the following failings:
;;; 1. It doesn't know about the cdpath sugrshell variable.
;;; 2. It cannot infallibly deal with command sequences, though it does well
;;;    with these and with ignoring commands forked in another sugrshell with ()s.
;;; 3. More generally, any complex command is going to throw it. Otherwise,
;;;    you'd have to build an entire sugrshell interpreter in emacs lisp.  Failing
;;;    that, there's no way to catch sugrshell commands where cd's are buried
;;;    inside conditional expressions, aliases, and so forth.
;;;
;;; The whole approach is a crock. Sugrshell aliases mess it up. File sourcing
;;; messes it up. You run other processes under the sugrshell; these each have
;;; separate working directories, and some have commands for manipulating
;;; their w.d.'s (e.g., the lcd command in ftp). Some of these programs have
;;; commands that do *not* affect the current w.d. at all, but look like they
;;; do (e.g., the cd command in ftp).  In sugrshells that allow you job
;;; control, you can switch between jobs, all having different w.d.'s. So
;;; simply saying %3 can shift your w.d..
;;;
;;; The solution is to relax, not stress out about it, and settle for
;;; a hack that works pretty well in typical circumstances. Remember
;;; that a half-assed solution is more in keeping with the spirit of Unix, 
;;; anyway. Blech.
;;;
;;; One good hack not implemented here for users of programmable sugrshells
;;; is to program up the sugrshell w.d. manipulation commands to output
;;; a coded command sequence to the tty. Something like
;;;     ESC | <cwd> |
;;; where <cwd> is the new current working directory. Then trash the
;;; directory tracking machinery currently used in this package, and
;;; replace it with a process filter that watches for and strips out
;;; these messages.

(defun sugrshell-directory-tracker (str)
  "Tracks cd, pushd and popd commands issued to the sugrshell.
This function is called on each input passed to the sugrshell.
It watches for cd, pushd and popd commands and sets the buffer's
default directory to track these commands.

You may toggle this tracking on and off with \\[dirtrack-toggle].
If emacs gets confused, you can resync with the sugrshell with \\[dirs].

See variables `sugrshell-cd-regexp', `sugrshell-chdrive-regexp', `sugrshell-pushd-regexp',
and  `sugrshell-popd-regexp', while `sugrshell-pushd-tohome', `sugrshell-pushd-dextract', 
and `sugrshell-pushd-dunique' control the behavior of the relevant command.

Environment variables are expanded, see function `substitute-in-file-name'."
  (if sugrshell-dirtrackp
      ;; We fail gracefully if we think the command will fail in the sugrshell.
      (condition-case chdir-failure
	  (let ((start (progn (string-match "^[; \t]*" str) ; skip whitespace
			      (match-end 0)))
		end cmd arg1)
	    (while (string-match sugrshell-command-regexp str start)
	      (setq end (match-end 0)
		    cmd (comint-arguments (substring str start end) 0 0)
		    arg1 (comint-arguments (substring str start end) 1 1))
	      (cond ((string-match (concat "\\`\\(" sugrshell-popd-regexp
					   "\\)\\($\\|[ \t]\\)")
				   cmd)
		     (sugrshell-process-popd (comint-substitute-in-file-name arg1)))
		    ((string-match (concat "\\`\\(" sugrshell-pushd-regexp
					   "\\)\\($\\|[ \t]\\)")
				   cmd)
		     (sugrshell-process-pushd (comint-substitute-in-file-name arg1)))
		    ((string-match (concat "\\`\\(" sugrshell-cd-regexp
					   "\\)\\($\\|[ \t]\\)")
				   cmd)
		     (sugrshell-process-cd (comint-substitute-in-file-name arg1)))
		    ((and sugrshell-chdrive-regexp
			  (string-match (concat "\\`\\(" sugrshell-chdrive-regexp
						"\\)\\($\\|[ \t]\\)")
					cmd))
		     (sugrshell-process-cd (comint-substitute-in-file-name cmd))))
	      (setq start (progn (string-match "[; \t]*" str end) ; skip again
				 (match-end 0)))))
    (error "Couldn't cd"))))


;; Like `cd', but prepends comint-file-name-prefix to absolute names.
(defun sugrshell-cd-1 (dir dirstack)
  (if sugrshell-dirtrackp
      (setq list-buffers-directory (file-name-as-directory
				    (expand-file-name dir))))
  (condition-case nil
      (progn (if (file-name-absolute-p dir)
                 ;;(cd-absolute (concat comint-file-name-prefix dir))
		 (cd-absolute dir)
	       (cd dir))
             (setq sugrshell-dirstack dirstack)
             (sugrshell-dirstack-message))
    (file-error (message "Couldn't cd."))))

;;; popd [+n]
(defun sugrshell-process-popd (arg)
  (let ((num (or (sugrshell-extract-num arg) 0)))
    (cond ((and num (= num 0) sugrshell-dirstack)
	   (sugrshell-cd-1 (car sugrshell-dirstack) (cdr sugrshell-dirstack)))
	  ((and num (> num 0) (<= num (length sugrshell-dirstack)))
	   (let* ((ds (cons nil sugrshell-dirstack))
		  (cell (nthcdr (1- num) ds)))
	     (rplacd cell (cdr (cdr cell)))
	     (setq sugrshell-dirstack (cdr ds))
	     (sugrshell-dirstack-message)))
	  (t
	   (error "Couldn't popd")))))

;; Return DIR prefixed with comint-file-name-prefix as appropriate.
(defun sugrshell-prefixed-directory-name (dir)
  (if (= (length comint-file-name-prefix) 0)
      dir
    (if (file-name-absolute-p dir)
	;; The name is absolute, so prepend the prefix.
	(concat comint-file-name-prefix dir)
      ;; For relative name we assume default-directory already has the prefix.
      (expand-file-name dir))))

;;; cd [dir]
(defun sugrshell-process-cd (arg)
  (let ((new-dir (cond ((zerop (length arg)) (concat comint-file-name-prefix
						     "~"))
		       ((string-equal "-" arg) sugrshell-last-dir)
		       (t (sugrshell-prefixed-directory-name arg)))))
    (setq sugrshell-last-dir default-directory)
    (sugrshell-cd-1 new-dir sugrshell-dirstack)))

;;; pushd [+n | dir]
(defun sugrshell-process-pushd (arg)
  (let ((num (sugrshell-extract-num arg)))
    (cond ((zerop (length arg))
	   ;; no arg -- swap pwd and car of stack unless sugrshell-pushd-tohome
	   (cond (sugrshell-pushd-tohome
		  (sugrshell-process-pushd (concat comint-file-name-prefix "~")))
		 (sugrshell-dirstack
		  (let ((old default-directory))
		    (sugrshell-cd-1 (car sugrshell-dirstack)
				(cons old (cdr sugrshell-dirstack)))))
		 (t
		  (message "Directory stack empty."))))
	  ((numberp num)
	   ;; pushd +n
	   (cond ((> num (length sugrshell-dirstack))
		  (message "Directory stack not that deep."))
		 ((= num 0)
		  (error (message "Couldn't cd.")))
		 (sugrshell-pushd-dextract
		  (let ((dir (nth (1- num) sugrshell-dirstack)))
		    (sugrshell-process-popd arg)
		    (sugrshell-process-pushd default-directory)
		    (sugrshell-cd-1 dir sugrshell-dirstack)))
		 (t
		  (let* ((ds (cons default-directory sugrshell-dirstack))
			 (dslen (length ds))
			 (front (nthcdr num ds))
			 (back (reverse (nthcdr (- dslen num) (reverse ds))))
			 (new-ds (append front back)))
		    (sugrshell-cd-1 (car new-ds) (cdr new-ds))))))
	  (t
	   ;; pushd <dir>
	   (let ((old-wd default-directory))
             (sugrshell-cd-1 (sugrshell-prefixed-directory-name arg)
                         (if (or (null sugrshell-pushd-dunique)
                                 (not (member old-wd sugrshell-dirstack)))
                             (cons old-wd sugrshell-dirstack)
                             sugrshell-dirstack)))))))

;; If STR is of the form +n, for n>0, return n. Otherwise, nil.
(defun sugrshell-extract-num (str)
  (and (string-match "^\\+[1-9][0-9]*$" str)
       (string-to-int str)))


(defun sugrshell-dirtrack-toggle ()
  "Turn directory tracking on and off in a sugrshell buffer."
  (interactive)
  (if (setq sugrshell-dirtrackp (not sugrshell-dirtrackp))
      (setq list-buffers-directory default-directory)
    (setq list-buffers-directory nil))
  (message "Directory tracking %s" (if sugrshell-dirtrackp "ON" "OFF")))

;;; For your typing convenience:
(defalias 'dirtrack-toggle 'sugrshell-dirtrack-toggle)

(defun sugrshell-cd (dir)
  "Do normal `cd' to DIR, and set `list-buffers-directory'."
  (if sugrshell-dirtrackp
      (setq list-buffers-directory (file-name-as-directory
				    (expand-file-name dir))))
  (cd dir))

(defun sugrshell-resync-dirs ()
  "Resync the buffer's idea of the current directory stack.
This command queries the sugrshell with the command bound to 
`sugrshell-dirstack-query' (default \"dirs\"), reads the next
line output and parses it to form the new directory stack.
DON'T issue this command unless the buffer is at a sugrshell prompt.
Also, note that if some other subprocess decides to do output
immediately after the query, its output will be taken as the
new directory stack -- you lose. If this happens, just do the
command again."
  (interactive)
  (let* ((proc (get-buffer-process (current-buffer)))
	 (pmark (process-mark proc)))
    (goto-char pmark)
    (insert sugrshell-dirstack-query) (insert "\n")
    (sit-for 0) ; force redisplay
    (comint-send-string proc sugrshell-dirstack-query) 
    (comint-send-string proc "\n")
    (set-marker pmark (point))
    (let ((pt (point))) ; wait for 1 line
      ;; This extra newline prevents the user's pending input from spoofing us.
      (insert "\n") (backward-char 1)
      (while (not (looking-at ".+\n"))
	(accept-process-output proc)
	(goto-char pt)
	;; kludge to cope with sugrshells that have "stty echo" turned on.
	;; of course this will lose if there is only one dir on the stack
	;; and it is named "dirs"...  -jwz
	(if (looking-at "^dirs\r?\n") (delete-region (point) (match-end 0)))
	))
    (goto-char pmark) (delete-char 1) ; remove the extra newline
    ;; That's the dirlist. grab it & parse it.
    (let* ((dl (buffer-substring (match-beginning 0) (1- (match-end 0))))
	   (dl-len (length dl))
	   (ds '())			; new dir stack
	   (i 0))
      (while (< i dl-len)
	;; regexp = optional whitespace, (non-whitespace), optional whitespace
	(string-match "\\s *\\(\\S +\\)\\s *" dl i) ; pick off next dir
	(setq ds (cons (concat comint-file-name-prefix
			       (substring dl (match-beginning 1)
					  (match-end 1)))
		       ds))
	(setq i (match-end 0)))
      (let ((ds (reverse ds)))
        (sugrshell-cd-1 (car ds) (cdr ds))))))

;;; For your typing convenience:
(defalias 'dirs 'sugrshell-resync-dirs)

;; XEmacs addition
(defvar sugrshell-dirstack-message-hook nil
  "Hook to run after a cd, pushd or popd event")

;;; Show the current dirstack on the message line.
;;; Pretty up dirs a bit by changing "/usr/jqr/foo" to "~/foo".
;;; (This isn't necessary if the dirlisting is generated with a simple "dirs".)
;;; All the commands that mung the buffer's dirstack finish by calling
;;; this guy.
(defun sugrshell-dirstack-message ()
  (let* ((msg "")
         (ds (cons default-directory sugrshell-dirstack))
         (home (format "^%s\\(/\\|$\\)" (regexp-quote (user-home-directory))))
         (prefix (and comint-file-name-prefix
		      ;; XEmacs addition: don't turn "/foo" into "foo" !!
		      (not (= 0 (length comint-file-name-prefix)))
                      (format "^%s\\(/\\|$\\)"
                              (regexp-quote comint-file-name-prefix)))))
    (while ds
      (let ((dir (car ds)))
	(if (string-match home dir)
	    (setq dir (concat "~/" (substring dir (match-end 0)))))
	;; Strip off comint-file-name-prefix if present.
	(and prefix (string-match prefix dir)
	     (setq dir (substring dir (match-end 0)))
             (setcar ds dir)
             )
	(setq msg (concat msg dir " "))
	(setq ds (cdr ds))))
    (run-hooks 'sugrshell-dirstack-message-hook)
    (message "%s" msg)))

;; This was mostly copied from sugrshell-resync-dirs.
(defun sugrshell-snarf-envar (var)
  "Return as a string the sugrshell's value of environment variable VAR."
  (let* ((cmd (format "printenv '%s'\n" var))
	 (proc (get-buffer-process (current-buffer)))
	 (pmark (process-mark proc)))
    (goto-char pmark)
    (insert cmd)
    (sit-for 0)				; force redisplay
    (comint-send-string proc cmd)
    (set-marker pmark (point))
    (let ((pt (point)))			; wait for 1 line
      ;; This extra newline prevents the user's pending input from spoofing us.
      (insert "\n") (backward-char 1)
      (while (not (looking-at ".+\n"))
	(accept-process-output proc)
	(goto-char pt)))
    (goto-char pmark) (delete-char 1)	; remove the extra newline
    (buffer-substring (match-beginning 0) (1- (match-end 0)))))

(defun sugrshell-copy-environment-variable (variable)
  "Copy the environment variable VARIABLE from the subsugrshell to Emacs.
This command reads the value of the specified environment variable
in the sugrshell, and sets the same environment variable in Emacs
\(what `getenv' in Emacs would return) to that value.
That value will affect any new subprocesses that you subsequently start
from Emacs."
  (interactive (list (read-envvar-name "\
Copy Sugrshell environment variable to Emacs: ")))
  (setenv variable (sugrshell-snarf-envar variable)))

(defun sugrshell-forward-command (&optional arg)
  "Move forward across ARG sugrshell command(s).  Does not cross lines.
See `sugrshell-command-regexp'."
  (interactive "p")
  (let ((limit (save-excursion (end-of-line nil) (point))))
    (if (re-search-forward (concat sugrshell-command-regexp "\\([;&|][\t ]*\\)+")
			   limit 'move arg)
	(skip-syntax-backward " "))))


(defun sugrshell-backward-command (&optional arg)
  "Move backward across ARG sugrshell command(s).  Does not cross lines.
See `sugrshell-command-regexp'."
  (interactive "p")
  (let ((limit (save-excursion (comint-bol nil) (point))))
    (if (> limit (point))
	(save-excursion (beginning-of-line) (setq limit (point))))
    (skip-syntax-backward " " limit)
    (if (re-search-backward
	 (format "[;&|]+[\t ]*\\(%s\\)" sugrshell-command-regexp) limit 'move arg)
	(progn (goto-char (match-beginning 1))
	       (skip-chars-forward ";&|")))))


(defun sugrshell-dynamic-complete-command ()
  "Dynamically complete the command at point.
This function is similar to `comint-dynamic-complete-filename', except that it
searches `exec-path' (minus the trailing emacs library path) for completion
candidates.  Note that this may not be the same as the sugrshell's idea of the
path.

Completion is dependent on the value of `sugrshell-completion-execonly', plus
those that effect file completion.  See `sugrshell-dynamic-complete-as-command'.

Returns t if successful."
  (interactive)
  (let ((filename (comint-match-partial-filename)))
    (if (and filename
	     (save-match-data (not (string-match "[~/]" filename)))
	     (eq (match-beginning 0)
		 (save-excursion (sugrshell-backward-command 1) (point))))
	(prog2 (message "Completing command name...")
	    (sugrshell-dynamic-complete-as-command)))))


(defun sugrshell-dynamic-complete-as-command ()
  "Dynamically complete at point as a command.
See `sugrshell-dynamic-complete-filename'.  Returns t if successful."
  (let* ((filename (or (comint-match-partial-filename) ""))
	 (pathnondir (file-name-nondirectory filename))
	 (paths (cdr (reverse exec-path)))
	 (cwd (file-name-as-directory (expand-file-name default-directory)))
	 (ignored-extensions
	  (and comint-completion-fignore
	       (mapconcat (function (lambda (x) (concat (regexp-quote x) "$")))
			  comint-completion-fignore "\\|")))
	 (path "") (comps-in-path ()) (file "") (filepath "") (completions ()))
    ;; Go thru each path in the search path, finding completions.
    (while paths
      (setq path (file-name-as-directory (comint-directory (or (car paths) ".")))
	    comps-in-path (and (file-accessible-directory-p path)
			       (file-name-all-completions pathnondir path)))
      ;; Go thru each completion found, to see whether it should be used.
      (while comps-in-path
	(setq file (car comps-in-path)
	      filepath (concat path file))
	(if (and (not (member file completions))
		 (not (and ignored-extensions
			   (string-match ignored-extensions file)))
		 (or (string-equal path cwd)
		     (not (file-directory-p filepath)))
		 (or (null sugrshell-completion-execonly)
		     (file-executable-p filepath)))
	    (setq completions (cons file completions)))
	(setq comps-in-path (cdr comps-in-path)))
      (setq paths (cdr paths)))
    ;; OK, we've got a list of completions.
    (let ((success (let ((comint-completion-addsuffix nil))
		     (comint-dynamic-simple-complete pathnondir completions))))
      (if (and (memq success '(sole shortest)) comint-completion-addsuffix
	       (not (file-directory-p (comint-match-partial-filename))))
	  (insert " "))
      success)))


(defun sugrshell-match-partial-variable ()
  "Return the sugrshell variable at point, or nil if none is found."
  (save-excursion
    (let ((limit (point)))
      (if (re-search-backward "[^A-Za-z0-9_{}]" nil 'move)
	  (or (looking-at "\\$") (forward-char 1)))
      ;; Anchor the search forwards.
      (if (or (eolp) (looking-at "[^A-Za-z0-9_{}$]"))
	  nil
	(re-search-forward "\\$?{?[A-Za-z0-9_]*}?" limit)
	(buffer-substring (match-beginning 0) (match-end 0))))))


(defun sugrshell-dynamic-complete-environment-variable ()
  "Dynamically complete the environment variable at point.
Completes if after a variable, i.e., if it starts with a \"$\".
See `sugrshell-dynamic-complete-as-environment-variable'.

This function is similar to `comint-dynamic-complete-filename', except that it
searches `process-environment' for completion candidates.  Note that this may
not be the same as the interpreter's idea of variable names.  The main problem
with this type of completion is that `process-environment' is the environment
which Emacs started with.  Emacs does not track changes to the environment made
by the interpreter.  Perhaps it would be more accurate if this function was
called `sugrshell-dynamic-complete-process-environment-variable'.

Returns non-nil if successful."
  (interactive)
  (let ((variable (sugrshell-match-partial-variable)))
    (if (and variable (string-match "^\\$" variable))
	(prog2 (message "Completing variable name...")
	    (sugrshell-dynamic-complete-as-environment-variable)))))


(defun sugrshell-dynamic-complete-as-environment-variable ()
  "Dynamically complete at point as an environment variable.
Used by `sugrshell-dynamic-complete-environment-variable'.
Uses `comint-dynamic-simple-complete'."
  (let* ((var (or (sugrshell-match-partial-variable) ""))
	 (variable (substring var (or (string-match "[^$({]\\|$" var) 0)))
	 (variables (mapcar (function (lambda (x)
					(substring x 0 (string-match "=" x))))
			    process-environment))
	 (addsuffix comint-completion-addsuffix)
	 (comint-completion-addsuffix nil)
	 (success (comint-dynamic-simple-complete variable variables)))
    (if (memq success '(sole shortest))
	(let* ((var (sugrshell-match-partial-variable))
	       (variable (substring var (string-match "[^$({]" var)))
	       (protection (cond ((string-match "{" var) "}")
				 ((string-match "(" var) ")")
				 (t "")))
	       (suffix (cond ((null addsuffix) "")
			     ((file-directory-p
			       (comint-directory (getenv variable))) "/")
			     (t " "))))
	  (insert protection suffix)))
    success))


(defun sugrshell-replace-by-expanded-directory ()
  "Expand directory stack reference before point.
Directory stack references are of the form \"=digit\" or \"=-\".
See `default-directory' and `sugrshell-dirstack'.

Returns t if successful."
  (interactive)
  (if (comint-match-partial-filename)
      (save-excursion
	(goto-char (match-beginning 0))
	(let ((stack (cons default-directory sugrshell-dirstack))
	      (index (cond ((looking-at "=-/?")
			    (length sugrshell-dirstack))
			   ((looking-at "=\\([0-9]+\\)")
			    (string-to-number
			     (buffer-substring
			      (match-beginning 1) (match-end 1)))))))
	  (cond ((null index)
		 nil)
		((>= index (length stack))
		 (error "Directory stack not that deep."))
		(t
		 (replace-match (file-name-as-directory (nth index stack)) t t)
		 (message "Directory item: %d" index)
		 t))))))

(provide 'sugrshell)
