=head1 NAME

  Statistics::AlgebraicMinimization -- algorithms for estimation using algebraic minimization. 

=head1 VERSION

  This document refers to version 0.01

=head1 SYNPOSIS
 
=head1 DESCRIPTION

=head2 Overview

=head2 Constructurs and initializaton

=head2 Class and object methods

=head2 Something else?!

=head2 Literature

=head1 DIAGNOSTICS

=over 4

=item "error messages?!"

=back

=head1 BUGS

=head1 TODO

=head1 FILES

=head1 SEE ALSO

=head1 AUTHOR(S)

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000 Stephan Heuel All rights reserved. There is no
warranty. You are allowed to redistribute this software /
documentation under certain conditions. For details, see the file
COPYING in the PDL distribution. If this file is separated from the
PDL distribution, the copyright notice should be included in the file.

=cut

package Statistics::AlgebraicMinimization;


use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 0.01;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL;
use PDL::Matrix;
use PDL::Math;
use PDL::Slatec;
use Statistics::GaussHelmert;

our $OUTPUT = 1;

# globals
my $number_of_iteration=10;
my $eps = 1e-2;


# this is like AlgebraicMinimizationBlocks_not_that_old, but the observations can be
# constrained in a second stage by the constraint g($y$) and its
# Jacobian $G$
sub AlgebraicMinimizationBlocks {
  my ($args) = @_;

  my ($y,$b0,$Sigma_yy,
      $func_w,$func_A,$func_B,
      $func_h,$func_H,
      $max_iterations) = (delete $args->{observations},
			  delete $args->{initial_guess},
			  delete $args->{covariance_observations},
			  delete $args->{observation_equations},
			  delete $args->{Jacobian_unknowns},
			  delete $args->{Jacobian_observations},
			  delete $args->{constraints},
			  delete $args->{Jacobian_constraints},
			  delete $args->{maximal_no_iterations},
			 );

  # deliberately ignore the following two options
  delete $args->{constraints_observations};
  delete $args->{Jacobian_constraints_observations};

  # warn for unknown options in argument hash
  foreach (keys %$args) {
    warn "*** Warning: Unknown option to AlgebraicMinimization: $_\n";
  }

  ################################################################################
  #
  # some preperations needed

  # set the maximum number of iterations
  $max_iterations ||= $number_of_iteration;

  # initialize $y^\hat$ and $b^\hat$.
  my @y_observed  = @$y; # these are my observations!

  my ($b,@b_iterations,$delta_b,$e,$Sigma_bb);

  my ($nU,$nW,$nH);
  $nW = &$func_w();		

  my $i=0;

  $b  = $b0->copy;
  @b_iterations = ($b); # here I store all iterations of the
  # unknown vector b
  $nU = ($b0->mdims())[0];	
  $nH = &$func_h();

  # assume Sigma_yy to be identity if it is not defined
  if (!defined($Sigma_yy)) {
    $Sigma_yy = [ map { my ($dim,undef,$no) = $y_observed[$_]->mdims();
			my $Id = mzeroes($dim,$dim,$no);
			(my $tmp = $Id->diagonal(0,1)) .= 1;
			$Id;
		       } (0..$#y_observed) ];
  }

  if ($nW + $nH < $nU) {
    warn "*** Warning: ",
      __FILE__," line ",__LINE__,"\n",
      "    # of unknowns ($nU) is larger than # of obs. ($nW) + # of constraints ($nH)\n";
    return undef;
  }
  my $sigma_hat;  

  if ($OUTPUT) {
    open LOG,">AlgebraicMinimization.log";
    print LOG "generated on ",scalar(localtime),"\n";
  }
  
  ################################################################################
  #
  # now start iterations!
  while ($i++ < $max_iterations) {
    if ($OUTPUT) {
      print LOG "Iteration $i\n";
      print "Iteration $i\n";
    } 
    
    
    my @A     = &$func_A([@y_observed],$b);
    my @B  = &$func_B([@y_observed],$b);
    my $H  = &$func_H($b);

    my @yhat = map { my ($dim,undef,$no) = $y_observed[$_]->mdims();
		     my $Id = mzeroes($dim,$dim,$no);
		     (my $tmp = $Id->diagonal(0,1)) .= 1;
		     ($Id - $Sigma_yy->[$_] x ~$B[$_] x
		      ($B[$_] x $Sigma_yy->[$_] x ~$B[$_])->matinv x
		      $B[$_]) x $y_observed[$_]
		    } (0..$#y_observed);
    my @Ahat = &$func_A([@yhat],$b);
    
    # $Sigma_ww^{-1}$.
    my @Sigma_ww_inv = map { ($B[$_] x $Sigma_yy->[$_] x ~$B[$_])->matinv } (0 .. $#B);
    
    
    my $M = mzeroes($nU,$nU);
    for (0..$#A) { $M += ~$Ahat[$_] x $Sigma_ww_inv[$_] x $A[$_] };
    my $MM = mzeroes($nU,$nU);
    for (0..$#A) { $MM += ~$A[$_] x $Sigma_ww_inv[$_] x $A[$_] };

    print LOG "A",@A,"B",@B,"Sigma_yy",@$Sigma_yy,"Sigma_ww_inv",@Sigma_ww_inv,"M",$M;
    
    # the following computation of $Sigma_bb, $Omega and $sigma_hat
    # refers to the $b of the last iteration!
    $Sigma_bb = SUGR::Element::pseudoinverse($M,~$H);

#    my $Omega = sum(~$b x $M x $b);;  
    my ($evector_omega,$evalue_omega) = eigens(~$MM x $MM);
    print "eigenvalues of MM: ",$evalue_omega,"\n";
    my $smallest_index_omega = (qsorti $evalue_omega)->at(0); # ... of eigenvalue vector
    my $Omega = sqrt($evalue_omega->slice("($smallest_index_omega)"));
    
    $sigma_hat = (($nW + $nH - $nU) > 0) ? ($Omega / ($nW + $nH  - $nU)) : 1;
    
    # compute smallest eigenvector 
    my ($evector,$evalue) = eigens(~$M x $M);
    my $smallest_index = (qsorti $evalue)->at(0); # ... of eigenvalue vector

    # eigenvector for smallest eigenvalue is the approximate value
    my $b_new = vpdl( $evector->slice(",($smallest_index)" ));
    
    my $delta_b = $b_new - $b;

    # compute criteria for a premature abortion of the iteration.
    # This is \delta(\beta)/sqrt(\sigma_\beta\beta) < 1e-2 means that the
    # corrections to the estimation is less than 1% w.r.t. to its
    # standard deviation.
    my ($abort_criterion,$abort_p);
    # only look at those entries where \sigma_\beta\beta is
    # significantly larger than 0
    my ($db_s,$sigma_bb_s) = where(abs($delta_b->clump(2)),
				   sqrt($Sigma_bb->diagonal(0,1)),
				   $Sigma_bb->diagonal(0,1) > 1e-16);
    $abort_criterion = max($db_s/$sigma_bb_s);
    $abort_p = ($abort_criterion < $eps);
    $b = $b_new;
    push @b_iterations,$b;

    # that's all computation for this iteration, but maybe we want to
    # output some of the results....
    if ( $OUTPUT) {
      if ($sigma_hat<0) {
	print "*** Warning: sigma_hat = $sigma_hat < 0, taking absolute value\n";
	$sigma_hat = abs($sigma_hat);
      }
      print LOG "sigma0 : ",sqrt($sigma_hat),"\n";
      print "sigma0 : ",sqrt($sigma_hat),"\n";

      # proof if Omega is equal to ...
      print LOG "Omega : $Omega\n";
      print "Omega = $Omega \n";

      my $h = &$func_h($b);	# new calculation of h
      print LOG "h(b$i) ".$h->slice(",(0)")."\n";
      print LOG "max | h(b$i) | ".PDL::max(abs($h->clump(2)))."\n";
      print "max | h(b$i) |  ".PDL::max(abs($h->clump(2)))."\n";

      print LOG "db($i) ".$delta_b->slice(",(0)")."\n";
      print LOG "b($i) ".$b->slice(",(0)")."\n";
      print LOG "b($i) (spherical)",($b->slice(",(0)")/sqrt(sum($b**2))),"\n";
      print "b($i) ".$b->slice(",(0)")."\n";
      print "b($i) (spherical)",($b->slice(",(0)")/sqrt(sum($b**2))),"\n";

      
      print LOG "max(db(i)/sigma_b(i)) $abort_criterion\n";
      print "max(db(i)/sigma_b(i)) $abort_criterion\n";

      print LOG "--------------------------------------------------\n";
      print "--------------------------------------------------\n";

    }
    
    # check if we can leave this estimation earlier
    last if ($abort_p); 
  }
  
  if ($OUTPUT) {
    close LOG;
  }
  # return the unknown and its covariance matrix, the estimated
  # observations.  Furthermore return an array with the unknown vector
  # for each iteration step and the number of iterations
  return { "estimated_unknown" => $b,
	   "sigma0_squared" => $sigma_hat,
	   "covariance_unknown" => $Sigma_bb, # without sigma_0 !
	   "estimated_unknowns_iterations" => [@b_iterations],
	   "number_of_iterations" => $i,
	 };
}
push @EXPORT,"AlgebraicMinimizationBlocks";

