=head1 NAME

  Statistics::GaussHelmert -- algorithms for estimation with Gauss-Helmert model

=head1 VERSION

  This document refers to version 0.02

=head1 SYNPOSIS
 
=head1 DESCRIPTION

=head2 Overview

=head2 Constructurs and initializaton

=head2 Class and object methods

=head2 Something else?!

=head Literature

=head1 DIAGNOSTICS

=over 4

=item "error messages?!"

=back

=head1 BUGS

=head1 TODO

=head1 FILES

=head1 SEE ALSO

=head1 AUTHOR(S)

Stephan Heuel
(stephan@heuel.org)

=head1 COPYRIGHT

Copyright (C) 2000 Stephan Heuel All rights reserved. There is no
warranty. You are allowed to redistribute this software /
documentation under certain conditions. For details, see the file
COPYING in the PDL distribution. If this file is separated from the
PDL distribution, the copyright notice should be included in the file.

=cut

package Statistics::GaussHelmert;


use strict;
use Carp;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 0.02;
@ISA = qw(Exporter);

@EXPORT = qw(); # symbols to autoexport, will be filled later on
@EXPORT_OK = qw(); #symbols to export on request
%EXPORT_TAGS = ( );

use PDL;
use PDL::Matrix;
use PDL::Math;
use PDL::Slatec;

our $OUTPUT = 0;

# default number of iterations
my $number_of_iteration=15;
# default abort criteria. This refers to the largest difference of
# change in the estimated vector divided by its standard deviation
my $eps = 1e-2;


# this is like GaussHelmertBlocks_not_that_old, but the observations can be
# constrained in a second stage by the constraint g($y$) and its
# Jacobian $G$.
sub GaussHelmertBlocks {
  my ($args) = @_;
  
  warn "I expect only one hash argument to GaussHelmertBlocks" if @_>1;

  # decode has arguments;
  my $y  = delete $args->{observations};
  my $b0 = delete $args->{initial_guess};
  my $Sigma_yy = delete $args->{covariance_observations};
  my $func_w = delete $args->{observation_equations};
  my $func_A = delete $args->{Jacobian_unknowns};
  my $func_B = delete $args->{Jacobian_observations};
  my $func_h = delete $args->{constraints};
  my $func_H = delete $args->{Jacobian_constraints};
  my $func_g = delete $args->{constraints_observations};
  my $func_G = delete $args->{Jacobian_constraints_observations};
  my $max_iterations = delete $args->{maximal_no_iterations};

  # warn for unknown options in argument hash
  foreach (keys %$args) {
    warn "*** Warning: Unknown option to GaussHelmertBlocks: $_\n";
  }

  ################################################################################
  #
  # some preperations needed

  # set the maximum number of iterations
  $max_iterations ||= $number_of_iteration;

  # initialize $y^\hat$ and $b^\hat$.
  my @y_observed  = @$y; # these are my observations!
  # in @y will be the estimated observations
  my @y = map { defined($_) ? $_->copy : undef } @y_observed;

  # now we set up variables such as unknowns $b, we keep track of the
  # unknowns per iteration (@b_iterations), the offerenc between two
  # consecutive unknowns $delta_b, the error $e and the coariance
  # matrix $Sigma_bb for the unknowns
  my ($b,@b_iterations,$delta_b,$e,$Sigma_bb);

  # get the number of unknowns, observations and constraints
  my ($nU,$nW,$nH);
  $nW = &$func_w();		

  # iteration 0
  my $i=0;

  # if the initial guess $b0 for the unknown is not defined and
  # therefore the Jacobian $A$ not defined either, we assume that only
  # the observations are supposed to be estimated.
  my ($only_observations,@A_init)=(0,undef);
  if (!defined($b0) && !defined($func_A)) {
    $only_observations=1;
    # build dummy observation and dummy Jacobian so that we do not
    # have to change too much code (for the sake of less
    # efficiency).
    $b0 = vpdl([0]); # the shortest unknown possible
    @A_init = map { (my $x = ~$y[$_]->copy) .= 0; $x; } (0..$#y);
    $nU=0;
    $nH=0;
  }
  else {
    $b  = $b0->copy;
    @b_iterations = ($b); # here I store all iterations of the
                          # unknown vector b
    $nU = ($b0->mdims())[0];	
    $nH = &$func_h();
  }

  # assume Sigma_yy to be identity if it is not defined
  if (!defined($Sigma_yy)) {
    $Sigma_yy = [ map { my ($dim,undef,$no) = $y[$_]->mdims();
			my $Id = mzeroes($dim,$dim,$no);
			(my $tmp = $Id->diagonal(0,1)) .= 1;
			$Id;
		      } (0..$#y) ];
  }
  
  if ($nW + $nH < $nU) {
    warn "*** Warning: ",
      __FILE__," line ",__LINE__,"\n",
      "    # of unknowns ($nU) is larger than # of obs. ($nW) + # of constraints ($nH)\n";
    return undef;
  }
  my $Omega_last_iteration = 0;
  my (@w,$sigma_hat);  

  if ($OUTPUT) {
    open LOG,">GaussHelmert.log";
    print LOG "generated on ",scalar(localtime),"\n";
    
    print LOG "Sigma_yy ",join(",",@$Sigma_yy);
  }
  
  ################################################################################
  #
  # now start iterations!
  while ($i++ < $max_iterations) {
    if ($OUTPUT) {
      print LOG "Iteration $i\n";
      print "Iteration $i\n";
    } 
    my @A  = ($only_observations) ? @A_init : &$func_A([@y],$b);
    my @B  = &$func_B([@y],$b);
    my $H  = ($only_observations) ? undef : &$func_H($b);

   if ($OUTPUT) { 
      foreach (0..$#y) { 
	print LOG "y$i part $_ ".$y[$_]->clump(3)."\n";
      } 
    } 

    # compute $c_w$.
    my @dy = map { $y_observed[$_] - $y[$_] } (0 .. $#y);
    my @cw = &$func_w([@y],$b);

    if ($OUTPUT) {
      my $ii = $i;
      print LOG "A($ii)",join(",",@A);
      print LOG "B($ii)",join(",",@B);
      print LOG "H($ii)",$H;
    }
    
    for (0..$#y) { $cw[$_] += $B[$_] x $dy[$_] }

    # $Sigma_ww^{-1}$.
#    print @B; print @$Sigma_yy;
    print LOG "\@B x ~\@B", join(",",map { ($B[$_] x $Sigma_yy->[$_] x ~$B[$_]) } (0 .. $#y)) if $OUTPUT;
    my @Sigma_ww_inv = map { ($B[$_] x $Sigma_yy->[$_] x ~$B[$_])->matinv } (0 .. $#y);

    if ($OUTPUT) { 
      @w = &$func_w([@y_observed],$b);	# new calculation of w for output purposes
      foreach (0..$#w) { 
	print LOG "cw(b$i,y$i) part $_ ".$cw[$_]->clump(3)."\n";
	print LOG "w(b$i,y) part $_ ".$w[$_]->clump(3)."\n";
	print LOG "Sigma_ww_inv(b$i,y) part $_ ".$Sigma_ww_inv[$_]."\n";
      } 
    } 

    if (!$only_observations) {
      my $ch = &$func_h($b);
      
      # building and solving normal equation system
      my $AtSwwA = mzeroes($nU,$nU);
      for (0..$#y) { $AtSwwA += msumover(~$A[$_] x $Sigma_ww_inv[$_] x $A[$_]) }
      my $M = ~append(~append($AtSwwA,~$H              ),
		      ~append($H    ,mzeroes($nH,$nH)));
      
      my $AtSwwcw  = mzeroes($nU);
      for (0..$#y) { $AtSwwcw += msumover(~$A[$_] x $Sigma_ww_inv[$_] x $cw[$_]) }
      
      my $s     = ~append( transpose(-$AtSwwcw),
			   transpose(-$ch)      );
      my $Minv  = $M->matinv;
      my $x     = $Minv x $s;
      
      # now compute the corrections to $\beta$.
      $delta_b = $x->slice("0:".($nU-1));
      $b	     = $b + $delta_b;
      push @b_iterations,$b;

      $Sigma_bb = $Minv->slice("0:".($nU-1).",0:".($nU-1));
    }
    else {
      $delta_b = vpdl([0]);
    }
    
    my @lambda = map { $Sigma_ww_inv[$_] x ($cw[$_] + $A[$_] x $delta_b) } (0..$#y);
    my @e      = map { $Sigma_yy->[$_] x ~$B[$_] x $lambda[$_] } (0..$#y);

    @y = map { $y_observed[$_] - $e[$_] } (0..$#y);

    # do we have to correct the observations y ?
    if (defined($func_g) && defined($func_G)) {

      my @ytmp = map { $_->copy } @y;

      if ( $OUTPUT ) {
	print LOG "-------> before correction of observations:\n";
	foreach (0..$#y) { 
	  print LOG "y($i) part $_ ".$y[$_]->clump(3)."\n";
	}
      }
      
      my $tmp = $OUTPUT;
      $OUTPUT=0;
      my $tmpresult= 
	GaussHelmertBlocks({
			    observations => [@ytmp],
			    observation_equations => $func_g,
			    Jacobian_observations => $func_G,
			    maximal_no_iterations => 5
			   });


      $OUTPUT=$tmp;
      my $y_est = $tmpresult->{estimated_observations};
      @y = @$y_est;

      if ( $OUTPUT ) {
	print LOG "-------->after correction of observations:\n";
	foreach (0..$#y) { 
	  print LOG "y($i) part $_ ".$y[$_]->clump(3)."\n";
	}
      }

      if ( $OUTPUT ) {
	print LOG "-------> division of before/after::\n";
	foreach (0..$#y) { 
	  print LOG "y($i) part $_ ".($y[$_]->clump(3)/$ytmp[$_]->clump(3))."\n";
	}
      }
    }

    # almost done ...
    my $Omega = 0;  
    foreach (0..$#y) { 
      $Omega += sum(~$e[$_] x ~$B[$_] x $lambda[$_]);
    };

    $sigma_hat = (($nW + $nH - $nU) > 0) ? ($Omega / ($nW + $nH - $nU)) : 1;

    # compute criteria for a premature abortion of the iteration.
    # This is \delta(\beta)/sqrt(\sigma_\beta\beta) < 1e-2 means that the
    # corrections to the estimation is less than 1% w.r.t. to its
    # standard deviation.
    my ($abort_criterion,$abort_p);
    if (!$only_observations) {
      # only look at those entries where \sigma_\beta\beta is
      # significantly larger than 0
      my ($db_s,$sigma_bb_s) = where(abs($delta_b->clump(2)),
				     sqrt($Sigma_bb->diagonal(0,1)),
				     $Sigma_bb->diagonal(0,1) > 1e-16);
      if ( $OUTPUT) {
	print LOG "|db| (selected)",$db_s,"\n"; 
	print LOG "sqrt(sigma_bb^2) (selected)",$sigma_bb_s,"\n"; 
      }
      $abort_criterion = max($db_s/$sigma_bb_s);
      $abort_p = ($abort_criterion < $eps);
    }
    else {
      $abort_p = 0; # needs to be defined properly
    }



    # that's all computation for this iteration, but maybe we want to
    # output some of the results....
    if ( $OUTPUT) {
      my $Omega1 = 0;  
      foreach (0..$#y) { 
	$Omega1 += sum(~$e[$_] x ~$B[$_] x $Sigma_ww_inv[$_] x $B[$_] x $e[$_]);
      };
      
      # proof if Omega is equal to ...
      print LOG "Omega e^T B \\lambda : $Omega\n";
      print LOG "Omega1 e^T B^T \\Sigma B e : $Omega1\n";
      print "Omega = ($Omega | $Omega1 )\n";

      if ($sigma_hat<0) {
	print "*** Warning: sigma0^2 = $sigma_hat < 0, taking absolute value\n";
#	print "$Omega,$Omega1\n";
#	die;
	print LOG "*** Warning: sigma0^2 = $sigma_hat < 0, taking absolute value\n";
      }
      print LOG "sigma0 : ",sqrt(abs($sigma_hat)),"\n";
      print "sigma0 : ",sqrt(abs($sigma_hat)),"\n";

      
      my $test=0;
      foreach (0..$#y) { 
	$test += sum(~$lambda[$_] x $B[$_] x $y[$_]); 
      };
      print LOG "e^T x Cyy^-1 x y^\\hat => $test\n";
      print "e^T x Cyy^-1 x y^\\hat => $test\n";


      @w = &$func_w([@y_observed],$b);	
      foreach (0..$#y) { 
	my $tmp = PDL::max(abs(($B[$_] x $e[$_]) - $w[$_]->clump(3))); 
	print LOG " max |B*e - w(b$i,y) | ($_)  : $tmp\n";
	print "max |B*e - w(b$i,y) | ($_)  : $tmp\n";
      };

      @w = &$func_w([@y],$b);	# new calculation of w for output purposes
      foreach (0..$#w) { 
	print LOG "w(b$i,y$i) part ($_) ".$w[$_]->clump(3)."\n";
	print LOG "max | w(b$i,y$i) | ($_) ".PDL::max(abs($w[$_]->clump(3)))."\n";
 	print "max | w(b$i,y$i) | ($_) ".PDL::max(abs($w[$_]->clump(3)))."\n";
	
      } 

      if (!$only_observations) {
	my $h = &$func_h($b);	# new calculation of h
	print LOG "h(b$i) ".$h->slice(",(0)")."\n";
 	print LOG "max | h(b$i) | ".PDL::max(abs($h->clump(2)))."\n";
 	print "max | h(b$i) |  ".PDL::max(abs($h->clump(2)))."\n";
      }

      if (!$only_observations) {
	print LOG "db($i) ".$delta_b->slice(",(0)")."\n";
	print LOG "b($i) ".$b->slice(",(0)")."\n";
	print LOG "b($i) (spherical)",($b->slice(",(0)")/sqrt(sum($b**2))),"\n";
	print "b($i) ".$b->slice(",(0)")."\n";
	print "b($i) (spherical)",($b->slice(",(0)")/sqrt(sum($b**2))),"\n";
      }

      foreach (0..$#e) { 
	print LOG "e($i) part $_ ".$e[$_]->clump(3)."\n";
      } 


      foreach (0..$#y) { 
	print LOG "y($i) part $_ ".$y[$_]->clump(3)."\n";
      }
      
      print LOG "max(db(i)/sigma_b(i)) $abort_criterion\n";
      print "max(db(i)/sigma_b(i)) $abort_criterion\n";

      print LOG "--------------------------------------------------\n";
      print "--------------------------------------------------\n";

    }
    
    # check if we can leave this estimation earlier
    last if ($abort_p); 
    $Omega_last_iteration=$Omega;
  }
  
  if ($OUTPUT) {
    close LOG;
  }
  # return the unknown and its covariance matrix, the estimated
  # observations.  Furthermore return an array with the unknown vector
  # for each iteration step and the number of iterations
  return { "estimated_unknown" => $b,
	   "sigma0_squared" => $sigma_hat,
	   "covariance_unknown" => $Sigma_bb, # without sigma_0 !
	   "estimated_observations" => [@y],
	   "estimated_unknowns_iterations" => [@b_iterations],
	   "number_of_iterations" => $i,
	 };
}
push @EXPORT,"GaussHelmertBlocks";


1;

__END__

